@@ -5,20 +5,35 @@ import os
 import sys
 import traceback
 from pathlib import Path
-from typing import List, Optional
+from typing import TYPE_CHECKING, List, Optional, cast
 
-from pydantic import ValidationError
+from dataclasses_json import DataClassJsonMixin
+
+from chalk._lsp.error_builder import LSPErrorBuilder
+
+if TYPE_CHECKING:
+    from pydantic import ValidationError
+else:
+    try:
+        from pydantic.v1 import ValidationError
+    except ImportError:
+        from pydantic import ValidationError
 
 from chalk.config.project_config import load_project_config
 from chalk.importer import import_all_files
-from chalk.parsed.user_types_to_json import get_registered_types_as_json
+from chalk.parsed.user_types_to_json import get_lsp_gql, get_registered_types_as_json
 from chalk.utils.stubgen import configure_stubgen_argparse, run_stubgen
 
 
-def _get_list_results(directory: Optional[str], file_allowlist: Optional[List[str]]):
+def get_list_results(directory: Optional[str], file_allowlist: Optional[List[str]]):
+    if directory is not None:
+        os.chdir(directory)
     scope_to = Path(directory or os.getcwd())
     try:
-        failed = import_all_files(file_allowlist=file_allowlist)
+        failed = import_all_files(
+            file_allowlist=file_allowlist,
+            project_root=None if directory is None else Path(directory),
+        )
         return get_registered_types_as_json(scope_to, failed)
     except Exception:
         ex_type, ex_value, ex_traceback = sys.exc_info()
@@ -26,7 +41,13 @@ def _get_list_results(directory: Optional[str], file_allowlist: Optional[List[st
         relevant_traceback = f"""{os.linesep.join(traceback.format_tb(ex_traceback))}
 \n{ex_type.__name__}: {str(ex_value)}
 """
-        return json.dumps(dict(failed=[dict(traceback=relevant_traceback)]), indent=2)
+        return json.dumps(
+            dict(
+                failed=[dict(traceback=relevant_traceback)],
+                lsp=cast(DataClassJsonMixin, get_lsp_gql()).to_dict(),
+            ),
+            indent=2,
+        )
 
 
 def dump_cmd(filename: str, directory: Optional[str], filter_file: Optional[str]):
@@ -37,17 +58,13 @@ def dump_cmd(filename: str, directory: Optional[str], filter_file: Optional[str]
 
     with open(filename, "w") as f:
         f.write(
-            _get_list_results(
+            get_list_results(
                 directory=directory,
                 file_allowlist=file_allowlist,
             )
         )
 
 
-def list_cmd(directory: str):
-    print(_get_list_results(directory=directory, file_allowlist=None))
-
-
 def config_cmd():
     json_response: str
     try:
@@ -73,20 +90,30 @@ def cli(args_override: Optional[List[str]] = None):
     )
     parser.add_argument("--log-level", help="Print debug info", nargs="?")
     subparsers = parser.add_subparsers(dest="command")
+
+    # CONFIG
     subparsers.add_parser("config", help="Print the config for the current project")
-    list_parser = subparsers.add_parser("list", help="List the resolvers for the current project and print as json")
-    list_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
-    dump_parser = subparsers.add_parser("dump", help="Write the resolvers for the current project to the given file")
-    dump_parser.add_argument("filename", help="Write to this file")
-    dump_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
+
+    # EXPORT
     export_parser = subparsers.add_parser(
         "export", help="Write the resolvers for the current project to the given file"
     )
     export_parser.add_argument("filename", help="Write to this file")
     export_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
     export_parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
+    # >= 3.9 version
+    # export_parser.add_argument(
+    #     "--lsp",
+    #     help="If set, will output the diagnostics in the format expected by the LSP instead of eagerly raising",
+    #     action=argparse.BooleanOptionalAction,
+    # )
+    # < 3.9 version
+    export_parser.add_argument("--lsp", action="store_true")
+    export_parser.add_argument("--no-lsp", dest="lsp", action="store_false")
+    export_parser.set_defaults(lsp=False)
+
+    # STUBGEN
     stubgen_parser = subparsers.add_parser("stubgen", help="Generate type stubs for feature set classes")
-    stubgen_parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
     configure_stubgen_argparse(stubgen_parser)
 
     # Parsing only known args for forwards compatibility.
@@ -102,20 +129,15 @@ def cli(args_override: Optional[List[str]] = None):
             level=level,
         )
 
-    if args.command == "list":
-        list_cmd(args.directory)
+    if args.command == "export":
+        LSPErrorBuilder.lsp = args.lsp
+        dump_cmd(filename=args.filename, directory=args.directory, filter_file=args.file_filter)
 
     elif args.command == "config":
         config_cmd()
 
     elif args.command == "stubgen":
-        run_stubgen(args, args.file_filter)
-
-    elif args.command == "export":
-        dump_cmd(args.filename, args.directory, args.file_filter)
-
-    elif args.command == "dump":
-        dump_cmd(args.filename, args.directory, None)
+        run_stubgen(args=args, file_filter=args.file_filter)
 
     else:
         parser.print_help(sys.stderr)
