@@ -4,21 +4,33 @@ import dataclasses
 import uuid
 from datetime import datetime
 from enum import Enum, IntEnum
-from typing import TYPE_CHECKING, Any, Dict, List, Literal, Mapping, Optional, Sequence, Union
+from typing import TYPE_CHECKING, Any, Dict, List, Literal, Mapping, Optional, Sequence, Tuple, Union
 
 import numpy as np
-from pydantic import BaseModel, Field, root_validator
 from typing_extensions import TypeAlias
 
 from chalk.byte_transmit.model import ByteBaseModel, ByteDict
-from chalk.features import Feature, ensure_feature
+from chalk.features import Feature
 from chalk.features._encoding.json import FeatureEncodingOptions
 from chalk.features.resolver import Resolver
 from chalk.features.tag import EnvironmentId
-from chalk.utils.missing_dependency import missing_dependency_exception
+from chalk.utils.df_utils import read_parquet
 
 if TYPE_CHECKING:
+    import pandas as pd
     import polars as pl
+    from pydantic import BaseModel, Field
+
+    root_validator = lambda _: (lambda x: x)
+else:
+    try:
+        from pydantic.v1 import BaseModel, Field, root_validator
+    except ImportError:
+        from pydantic import BaseModel, Field, root_validator
+
+MAX_STR_LENGTH = 10_000
+
+FeatureReference: TypeAlias = Union[str, Any]
 
 
 def _category_for_error_code(c: Union[ErrorCode, str]) -> ErrorCodeCategory:
@@ -58,6 +70,8 @@ class OnlineQueryContext(BaseModel):
     More information at https://docs.chalk.ai/docs/resolver-tags
     """
 
+    required_resolver_tags: Optional[List[str]] = None
+
 
 class OfflineQueryContext(BaseModel):
     environment: Optional[str] = None
@@ -175,6 +189,14 @@ class ChalkException(BaseModel, frozen=True):
     internal_stacktrace: Optional[str] = None
     """The stacktrace produced by the code, full detail."""
 
+    @root_validator
+    def _validate_category(cls, values: Dict[str, Any]):
+        values["message"] = values["message"][0:MAX_STR_LENGTH]
+        values["stacktrace"] = values["stacktrace"][0:MAX_STR_LENGTH]
+        if values.get("internal_stacktrace") is not None:
+            values["internal_stacktrace"] = values["internal_stacktrace"][0:MAX_STR_LENGTH]
+        return values
+
 
 class ChalkError(BaseModel, frozen=True):
     """
@@ -217,6 +239,12 @@ class ChalkError(BaseModel, frozen=True):
     The fully qualified name of the failing resolver, e.g. `my.project.get_fraud_score`.
     """
 
+    def is_resolver_runtime_error(self) -> bool:
+        """
+        Returns True if the error indicates an issue with user's resolver, rather than an internal Chalk failure.
+        """
+        return self.code in [ErrorCode.RESOLVER_FAILED, ErrorCode.RESOLVER_TIMED_OUT]
+
     def copy_for_feature(self, feature: str) -> "ChalkError":
         return self.copy(update={"feature": feature})
 
@@ -226,12 +254,13 @@ class ChalkError(BaseModel, frozen=True):
     @root_validator
     def _validate_category(cls, values: Dict[str, Any]):
         values["category"] = _category_for_error_code(values["code"])
+        values["message"] = values["message"][0:MAX_STR_LENGTH]
         return values
 
     if TYPE_CHECKING:
         # Defining __hash__ only when type checking
         # since pydantic provides a hash for frozen models
-        def __hash__(self) -> int:  # type: ignore
+        def __hash__(self) -> int:
             ...
 
 
@@ -324,6 +353,7 @@ class ExchangeCredentialsRequest(BaseModel):
     client_id: str
     client_secret: str
     grant_type: str
+    scope: Optional[str] = None
 
 
 class ExchangeCredentialsResponse(BaseModel):
@@ -355,12 +385,14 @@ class OnlineQueryRequest(BaseModel):
     skip_cache_lookups: Optional[bool] = False
     correlation_id: Optional[str] = None
     query_name: Optional[str] = None
+    query_name_version: Optional[str] = None
     deployment_id: Optional[str] = None
     branch_id: Optional[str] = None
     meta: Optional[Mapping[str, str]] = None
     store_plan_stages: Optional[bool] = False
     gcs_client: Optional[Any] = None
     encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
+    planner_options: Optional[Mapping[str, Union[str, int, bool]]] = None
 
 
 @dataclasses.dataclass
@@ -369,6 +401,7 @@ class OnlineQuery:
     output: List[str]
     staleness: Optional[Mapping[str, str]] = None
     tags: Optional[List[str]] = None
+    required_resolver_tags: Optional[List[str]] = None
 
 
 class OnlineQueryManyRequest(BaseModel):
@@ -385,6 +418,7 @@ class OnlineQueryManyRequest(BaseModel):
     skip_cache_lookups: Optional[bool] = False
     correlation_id: Optional[str] = None
     query_name: Optional[str] = None
+    query_name_version: Optional[str] = None
     deployment_id: Optional[str] = None
     branch_id: Optional[str] = None
     meta: Optional[Mapping[str, str]] = None
@@ -393,8 +427,21 @@ class OnlineQueryManyRequest(BaseModel):
     encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
 
 
+class MultiUploadFeaturesRequest(ByteBaseModel):
+    features: List[str]
+    table_compression: str
+    table_bytes: bytes
+
+
+class MultiUploadFeaturesResponse(BaseModel):
+    operation_id: str
+    errors: List[ChalkError]
+
+
 class TriggerResolverRunRequest(BaseModel):
     resolver_fqn: str
+    upper_bound: Optional[str] = None
+    lower_bound: Optional[str] = None
 
 
 class QueryMeta(BaseModel):
@@ -458,11 +505,37 @@ class OnlineQueryResponse(BaseModel):
 
 @dataclasses.dataclass
 class BulkOnlineQueryResult:
+    """
+    Represents the result of a single `OnlineQuery`, returned by
+    `query_bulk`.
+
+    The `scalars_df` member holds the primary results of the `OnlineQuery`.
+    Access this data using `to_polars()` or `to_pandas()`.
+    """
+
     scalars_df: Optional[pl.DataFrame]
     groups_dfs: Optional[Dict[str, pl.DataFrame]]  # change to chalk df whenever we figure out pydantic
     errors: Optional[List[ChalkError]]
     meta: Optional[QueryMeta]
 
+    def to_polars(self) -> pl.DataFrame:
+        """
+        Allows access to the results of an `OnlineQuery` submitted to `query_bulk` as a Polars dataframe.
+        """
+
+        return self.scalars_df if self.scalars_df is not None else pl.DataFrame()
+
+    def to_pandas(self) -> pd.DataFrame:
+        """
+        Allows access to the results of an `OnlineQuery` submitted to `query_bulk` as a Pandas dataframe.
+        """
+        if self.scalars_df is not None:
+            return self.scalars_df.to_pandas()
+        else:
+            import pandas as pd
+
+            return pd.DataFrame()
+
     # def get_feature_value(self, pkey: str | int, f: FeatureReference):
     #     f_casted = ensure_feature(f)
     #     if f_casted.is_has_many:
@@ -475,8 +548,11 @@ class BulkOnlineQueryResult:
 class BulkOnlineQueryResponse:
     results: List[BulkOnlineQueryResult]
 
-
-FeatureReference: TypeAlias = Union[str, Any]
+    def __getitem__(self, item: int):
+        """
+        Support `client.query_bulk(...)[0]` syntax.
+        """
+        return self.results[item]
 
 
 class CreateOfflineQueryJobRequest(BaseModel):
@@ -490,7 +566,7 @@ class CreateOfflineQueryJobRequest(BaseModel):
             "then jobs with the same ID will be rejected."
         ),
     )
-    input: Optional[OfflineQueryInput] = Field(default=None, description="Any givens")
+    input: Union[OfflineQueryInput, Tuple[OfflineQueryInput, ...], None] = Field(default=None, description="Any givens")
     max_samples: Optional[int] = Field(
         default=None,
         description="The maximum number of samples. If None, no limit",
@@ -515,9 +591,26 @@ class CreateOfflineQueryJobRequest(BaseModel):
     branch: Optional[str] = None
     recompute_features: Union[bool, List[FeatureReference]] = False
     store_plan_stages: bool = False
-    explain: Union[bool, Literal["onlyu"]] = False
+    explain: Union[bool, Literal["only"]] = False
     tags: Optional[List[str]] = None
+    required_resolver_tags: Optional[List[str]] = Field(
+        default=None,
+        description="""
+    If specified, all resolvers invoked as part of this query must be tagged with all of these tags.
+    Can be used to ensure that expensive resolvers are not executed.
+    """,
+    )
     correlation_id: Optional[str] = None
+    planner_options: Optional[Any] = None
+    use_multiple_computers: bool = False
+
+    spine_sql_query: Optional[str] = None
+
+    @root_validator
+    def _validate_multiple_computers(cls, values: Dict[str, Any]):
+        if values["input"] is not None and values["use_multiple_computers"] != isinstance(values["input"], tuple):
+            raise ValueError("input should be tuple exactly when use_multiple_computers is True")
+        return values
 
 
 class ComputeResolverOutputRequest(BaseModel):
@@ -527,6 +620,14 @@ class ComputeResolverOutputRequest(BaseModel):
     environment: Optional[str] = None
 
 
+class DatasetJobStatusRequest(BaseModel):
+    job_id: Optional[str]  # same as revision_id
+    dataset_id: Optional[str] = None
+    dataset_name: Optional[str] = None
+    ignore_errors: bool = False
+    query_inputs: bool = False
+
+
 class DatasetRecomputeRequest(BaseModel):
     dataset_name: Optional[str] = None
     branch: str
@@ -654,10 +755,13 @@ class DatasetRevisionResponse(BaseModel):
     output_uris: str
     output_version: int
     branch: Optional[str] = None
+    dashboard_url: Optional[str] = None
+    num_computers: int = 1
+    errors: Optional[List[ChalkError]] = None
 
 
 class DatasetRecomputeResponse(DatasetRevisionResponse):
-    errors: Optional[List[ChalkError]] = None
+    num_computers: Literal[1] = 1  # pyright: ignore[reportIncompatibleVariableOverride]
 
     @classmethod
     def from_revision_response(
@@ -665,7 +769,7 @@ class DatasetRecomputeResponse(DatasetRevisionResponse):
     ) -> "DatasetRecomputeResponse":
         return cls(
             revision_id=revision.revision_id,
-            environment=revision.environment_id,
+            environment_id=revision.environment_id,
             creator_id=revision.creator_id,
             outputs=revision.outputs,
             givens_uri=revision.givens_uri,
@@ -681,7 +785,9 @@ class DatasetRecomputeResponse(DatasetRevisionResponse):
             dataset_name=revision.dataset_name,
             dataset_id=revision.dataset_id,
             branch=revision.branch,
+            dashboard_url=revision.dashboard_url,
             errors=errors,
+            num_computers=1,
         )
 
 
@@ -886,7 +992,19 @@ class BranchDeployResponse(BaseModel):
     branch_deployment_id: str
 
 
-BranchIdParam: TypeAlias = Union[None, str, type(Ellipsis)]
+class BranchStartRequest(BaseModel):
+    branch_environment_id: str
+    """
+    The environment id of the branch to start.
+    """
+
+
+class BranchStartResponse(BaseModel):
+    status: Union[Literal["ok"], Literal["error"]]
+    message: str
+
+
+BranchIdParam: TypeAlias = Union[None, str, "ellipsis"]
 """
 Type used for the 'branch' paremeter in calls to the Chalk Client.
 The branch can either be:
@@ -925,14 +1043,12 @@ class StreamResolverTestResponse(BaseModel):
     def features(self) -> pl.DataFrame:
         if self.data_uri is None:
             raise ValueError(
-                "Features were not saved to storage. "
-                "Please inspect 'ResolverTestResponse.errors' and 'ResolverTestResponse.message'."
+                (
+                    "Features were not saved to storage. "
+                    "Please inspect 'ResolverTestResponse.errors' and 'ResolverTestResponse.message'."
+                )
             )
-        try:
-            import polars as pl
-        except ImportError:
-            raise missing_dependency_exception("chalkpy[runtime]")
-        return pl.read_parquet(self.data_uri)
+        return read_parquet(self.data_uri)
 
 
 class OnlineQueryResultFeather(ByteBaseModel):
@@ -945,3 +1061,52 @@ class OnlineQueryResultFeather(ByteBaseModel):
 
 class OnlineQueryResponseFeather(ByteBaseModel):
     query_results_bytes: ByteDict
+
+
+class ResolverReplayResponse(BaseModel):
+    urls: Optional[List[str]] = None
+    error: Optional[str] = None
+
+
+class UploadFeaturesRequest(BaseModel):
+    input: Mapping[str, List[Any]]  # Values should be of type List[TJSON]
+    preview_deployment_id: Optional[str] = None
+    correlation_id: Optional[str] = None
+    query_name: Optional[str] = None
+    meta: Optional[Mapping[str, str]] = None
+
+
+class UploadFeaturesResponse(BaseModel):
+    errors: List[ChalkError]
+
+
+##
+class PlanQueryRequest(BaseModel):
+    inputs: List[str]
+    outputs: List[str]
+    staleness: Optional[Mapping[str, str]] = None
+    context: Optional[OnlineQueryContext] = None
+    query_name: Optional[str] = None
+    query_name_version: Optional[str] = None
+    deployment_id: Optional[str] = None
+    branch_id: Optional[str] = None
+    meta: Optional[Mapping[str, str]] = None
+    gcs_client: Optional[Any] = None  # deprecated
+    encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
+
+
+class FeatureSchema(BaseModel):
+    fqn: str
+    primitive_type: Optional[str]
+    rich_type: Optional[str]
+    nullable: bool
+    pyarrow_dtype: str
+
+
+class PlanQueryResponse(BaseModel):
+    rendered_plan: Optional[str]
+    output_schema: List[FeatureSchema]
+    errors: List[ChalkError]
+
+
+##
