@@ -1,10 +1,17 @@
 import os
 from functools import lru_cache
 from pathlib import Path
-from typing import Mapping, Optional
+from typing import TYPE_CHECKING, Mapping, Optional
 
 import yaml
-from pydantic import BaseModel
+
+if TYPE_CHECKING:
+    from pydantic import BaseModel
+else:
+    try:
+        from pydantic.v1 import BaseModel
+    except ImportError:
+        from pydantic import BaseModel
 
 from chalk.config.project_config import load_project_config
 from chalk.utils.log_with_context import get_logger
