@@ -1,13 +1,20 @@
 import itertools
 import os
 from pathlib import Path
-from typing import List, Literal, Mapping, Optional, Union
+from typing import TYPE_CHECKING, List, Literal, Mapping, Optional, Union
 
 import yaml
-from pydantic import BaseModel
 
 from chalk._version import __version__
 
+if TYPE_CHECKING:
+    from pydantic import BaseModel
+else:
+    try:
+        from pydantic.v1 import BaseModel
+    except ImportError:
+        from pydantic import BaseModel
+
 
 class EnvironmentSettings(BaseModel):
     runtime: Optional[Union[Literal["python310"], Literal["python311"]]]
@@ -24,8 +31,13 @@ class FeatureSettings(BaseModel):
     metadata: Optional[List[MetadataSettings]]
 
 
+class ResolverSettings(BaseModel):
+    metadata: Optional[List[MetadataSettings]]
+
+
 class ValidationSettings(BaseModel):
     feature: Optional[FeatureSettings]
+    resolver: Optional[ResolverSettings]
 
 
 class ProjectSettings(BaseModel):
