@@ -1,10 +1,14 @@
+from __future__ import annotations
+
 from typing import Any, List, Optional
 
 from chalk.features._chalkop import Aggregation, op
+from chalk.features._embedding.embedding import embed
 from chalk.features._encoding.converter import FeatureConverter
 from chalk.features._encoding.missing_value import MissingValueStrategy
 from chalk.features._encoding.primitive import TPrimitive
 from chalk.features._encoding.serialized_dtype import deserialize_dtype, serialize_dtype, serialize_pyarrow_dtype
+from chalk.features._vector import Vector
 from chalk.features.dataframe import DataFrame
 from chalk.features.feature_field import Feature, FeatureNotFoundException, feature, has_many, has_one
 from chalk.features.feature_set import Features, FeatureSetBase, is_features_cls
@@ -15,6 +19,7 @@ from chalk.features.filter import Filter, TimeDelta, after, before
 from chalk.features.hooks import after_all, before_all
 from chalk.features.primary import Primary, is_primary
 from chalk.features.resolver import Cron, Resolver, ResolverProtocol, offline, online, sink
+from chalk.features.scheduled_query import ScheduledQuery
 from chalk.features.tag import Environments, Tags
 from chalk.features.underscore import Underscore, _, underscore
 from chalk.utils import MachineType
@@ -142,6 +147,42 @@ def tags(f: Any) -> Optional[List[str]]:
     raise TypeError(f"Could not determine the tags of '{f}' as it is neither a Feature, Feature Set, or Resolver")
 
 
+def version(f: Any) -> int | None:
+    """Get the version for a feature.
+
+    Parameters
+    ----------
+    f
+        A feature (`User.email`)
+
+    Returns
+    -------
+    int | None
+        The version for a feature, if it exists.
+
+    Examples
+    --------
+    Feature version
+    >>> @features
+    ... class Animal:
+    ... id: str
+    ... sound: str = feature(version=2)
+    >>> version(Animal.sound)
+    2
+
+    Raises
+    ------
+    TypeError
+        If the supplied variable is not a feature.
+    """
+    if isinstance(f, (Feature, FeatureWrapper)):
+        feature_version_info = unwrap_feature(f).version
+        if feature_version_info:
+            return feature_version_info.version
+        return feature_version_info
+    raise TypeError(f"Could not determine the version of '{f}' as it is not a Feature.")
+
+
 __all__ = [
     "Aggregation",
     "Cron",
@@ -182,6 +223,7 @@ __all__ = [
     "online",
     "op",
     "owner",
+    "embed",
     "serialize_dtype",
     "serialize_pyarrow_dtype",
     "sink",
@@ -189,4 +231,6 @@ __all__ = [
     "unwrap_feature",
     "_",
     "underscore",
+    "Vector",
+    "ScheduledQuery",
 ]
