@@ -6,12 +6,15 @@ if TYPE_CHECKING:
     import polars as pl
 
     from chalk.features import Filter
+    from chalk.features.feature_field import Feature
+    from chalk.features.feature_wrapper import FeatureWrapper
 
 
 class Aggregation:
     """A class for refining an aggregation defined by `op`."""
 
     def __init__(self, col: pl.Expr, fn: Callable[[pl.Expr], pl.Expr]):
+        super().__init__()
         self.col = col
         self.fn = fn
         self.filters: List[Filter] = []
@@ -55,10 +58,12 @@ class Aggregation:
         self.filters.extend(f)
         return self
 
-    def __add__(self, other):
+    def __add__(self, other: object):
+        # Needed to get things like sum(agg) to work
         return self
 
-    def __radd__(self, other):
+    def __radd__(self, other: object):
+        # Needed to get things like sum(agg) to work
         return self
 
 
@@ -70,7 +75,9 @@ class op:
     """
 
     @classmethod
-    def sum(cls, col, *cols) -> Aggregation:
+    def sum(
+        cls, col: Union[Feature, FeatureWrapper, str, Any], *cols: Union[Feature, FeatureWrapper, str, Any]
+    ) -> Aggregation:
         """Add together the values of `col` and `*cols` in a `DataFrame`.
 
         Parameters
@@ -102,14 +109,17 @@ class op:
         """
         import polars as pl
 
-        cols = [col, *[t for t in cols]]
+        cols = tuple((col, *(t for t in cols)))
         if len(cols) == 1:
             return Aggregation(pl.col(str(cols[0])), lambda x: x.sum())
         c = pl.col([str(c) for c in cols])
-        return Aggregation(c, lambda x: pl.sum(c))
+        if hasattr(pl, "sum_horizontal"):
+            return Aggregation(c, lambda x: pl.sum_horizontal(c))
+        agg_fn = lambda x: pl.sum(c)  # type: ignore - polars backcompat
+        return Aggregation(c, agg_fn)
 
     @classmethod
-    def product(cls, col) -> Aggregation:
+    def product(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Multiply together the values of `col` in a `DataFrame`.
 
         Parameters
@@ -146,7 +156,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.product())
 
     @classmethod
-    def max(cls, col) -> Aggregation:
+    def max(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the maximum of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -179,7 +189,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.max())
 
     @classmethod
-    def min(cls, col) -> Aggregation:
+    def min(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the minimum of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -212,14 +222,14 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.min())
 
     @classmethod
-    def quantile(cls, col, q: float) -> Aggregation:
+    def quantile(cls, col: Union[Feature, FeatureWrapper, str, Any], q: float) -> Aggregation:
         import polars as pl
 
         assert q >= 0 <= 1, f"Quantile must be between 0 and 1, but given {q}"
         return Aggregation(pl.col(str(col)), lambda x: x.quantile(q))
 
     @classmethod
-    def median(cls, col) -> Aggregation:
+    def median(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the median of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -254,7 +264,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.median())
 
     @classmethod
-    def mean(cls, col) -> Aggregation:
+    def mean(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the mean of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -287,7 +297,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.mean())
 
     @classmethod
-    def std(cls, col) -> Aggregation:
+    def std(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the standard deviation of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -313,7 +323,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.std())
 
     @classmethod
-    def variance(cls, col) -> Aggregation:
+    def variance(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the variance of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -339,7 +349,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.var())
 
     @classmethod
-    def count(cls, col) -> Aggregation:
+    def count(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the count of the values of `col` in a `DataFrame`.
 
         Parameters
@@ -372,7 +382,12 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.count())
 
     @classmethod
-    def concat(cls, col, col2, sep: str = "") -> Aggregation:
+    def concat(
+        cls,
+        col: Union[Feature, FeatureWrapper, str, Any],
+        col2: Union[Feature, FeatureWrapper, str, Any],
+        sep: str = "",
+    ) -> Aggregation:
         """Concatenate the string values of `col` and `col2` in a `DataFrame`.
 
         Parameters
@@ -412,7 +427,12 @@ class op:
         return Aggregation(c, lambda x: pl.concat_str(c, separator=sep))
 
     @classmethod
-    def concat_str(cls, col, col2, sep: str = "") -> Aggregation:
+    def concat_str(
+        cls,
+        col: Union[Feature, FeatureWrapper, str, Any],
+        col2: Union[Feature, FeatureWrapper, str, Any],
+        sep: str = "",
+    ) -> Aggregation:
         """Deprecated. Use `concat` instead."""
         import polars as pl
 
@@ -420,7 +440,7 @@ class op:
         return Aggregation(c, lambda x: pl.concat_str(c, separator=sep))
 
     @classmethod
-    def last(cls, col) -> Aggregation:
+    def last(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the last value of `col` in a `DataFrame`.
 
         Parameters
@@ -457,7 +477,7 @@ class op:
         return Aggregation(pl.col(str(col)), lambda x: x.last())
 
     @classmethod
-    def first(cls, col) -> Aggregation:
+    def first(cls, col: Union[Feature, FeatureWrapper, str, Any]) -> Aggregation:
         """Find the first value of `col` in a `DataFrame`.
 
         Parameters
