@@ -5,18 +5,25 @@ import collections.abc
 import ipaddress
 import uuid
 from datetime import date, datetime, time, timedelta
-from typing import Any, List, Optional, Type, Union, cast
+from typing import TYPE_CHECKING, Any, List, Optional, Type, Union, cast
 
 import cattrs
 import dateutil.parser
 import isodate
-import pydantic
-from typing_extensions import get_args, get_origin, is_typeddict
+from typing_extensions import get_args, get_origin
 
-from chalk.features._encoding.primitive import TPrimitive
-from chalk.utils.cached_type_hints import cached_get_type_hints
+from chalk.features._encoding.primitive import ChalkStructType, TPrimitive
+from chalk.features._encoding.rich import is_chalk_struct
 from chalk.utils.json import TJSON
 
+if TYPE_CHECKING:
+    import pydantic
+else:
+    try:
+        import pydantic.v1 as pydantic
+    except ImportError:
+        import pydantic
+
 __all__ = ["unstructure_primitive_to_json", "structure_json_to_primitive"]
 
 
@@ -30,6 +37,7 @@ def structure_json_to_primitive(val: Union[TJSON, TPrimitive], typ: Type[TPrimit
 
 class FeatureJsonConverter:
     def __init__(self, options: FeatureEncodingOptions):
+        super().__init__()
         self._json_converter = cattrs.Converter()
         self._options = options
         #######
@@ -51,8 +59,10 @@ class FeatureJsonConverter:
                 dict, lambda x: [self._json_converter.unstructure(y) for y in x.values()]
             )
 
-        def _structure_typeddict(obj: Any, typ: Type):
-            type_hints = cached_get_type_hints(typ)
+        def _structure_chalk_struct(obj: Any, typ: Type):
+            if not isinstance(typ, ChalkStructType):
+                raise TypeError(f"Expected ChalkStructType instead of {typ}")
+            type_hints = typ.__chalk_type_hints__
             if obj is None:
                 return {
                     field_name: self._json_converter.structure(None, type_hint)
@@ -78,10 +88,7 @@ class FeatureJsonConverter:
             }
             return kwargs
 
-        if self._options.encode_structs_as_objects:
-            ...
-        else:
-            self._json_converter.register_structure_hook_func(is_typeddict, _structure_typeddict)
+        self._json_converter.register_structure_hook_func(is_chalk_struct, _structure_chalk_struct)
 
         #######
         # Lists
@@ -101,7 +108,7 @@ class FeatureJsonConverter:
                 raise TypeError(f"{typ} should be parameterized with only one type")
             if obj is None:
                 return cast(List, None)
-            if not isinstance(obj, (list, tuple)):
+            if not isinstance(obj, (list, tuple)):  # pyright: ignore[reportUnnecessaryIsInstance]
                 raise TypeError(f"Expected a list, Object `{obj}` is not a list.")
 
             inner_typ = args[0]
@@ -125,6 +132,8 @@ class FeatureJsonConverter:
         )
 
         def _structure_date(obj: Any, typ: Type):
+            if obj is None:
+                return None
             if isinstance(obj, datetime):
                 return obj.date()
             if isinstance(obj, date):
@@ -314,6 +323,6 @@ class FeatureEncodingOptions(pydantic.BaseModel, frozen=True):
 
 
 _default_feature_json_converter = FeatureJsonConverter(options=FeatureEncodingOptions())
-_structs_as_objects_feature_json_converter = FeatureJsonConverter(
+structs_as_objects_feature_json_converter = FeatureJsonConverter(
     options=FeatureEncodingOptions(encode_structs_as_objects=True)
 )
