@@ -4,37 +4,41 @@ import dataclasses
 import decimal
 import enum
 import ipaddress
+import typing
 import uuid
 from datetime import date, datetime, time, timedelta
-from typing import TYPE_CHECKING, Any, Dict, FrozenSet, List, Set, Tuple, Type, TypedDict, Union, cast
+from typing import TYPE_CHECKING, Any, Dict, FrozenSet, List, Set, Tuple, Type, Union, cast
 
 import attrs
 import pyarrow as pa
-from pydantic import BaseModel
-from typing_extensions import Annotated, get_args, get_origin, is_typeddict
+from typing_extensions import Annotated, TypeGuard, get_args, get_origin, is_typeddict
 
-from chalk.features._encoding.primitive import TPrimitive
+from chalk.features._encoding.primitive import ChalkStructType, TPrimitive
+from chalk.features.feature_set import Features
 from chalk.utils.cached_type_hints import cached_get_type_hints
-from chalk.utils.collections import is_namedtuple, unwrap_optional_and_annotated_if_needed
+from chalk.utils.collections import is_namedtuple, is_optional, unwrap_optional_and_annotated_if_needed
 from chalk.utils.enum import get_enum_value_type
 from chalk.utils.missing_dependency import missing_dependency_exception
+from chalk.utils.pydanticutil.pydantic_compat import is_pydantic_basemodel
+
+if TYPE_CHECKING:
+    import polars as pl
+
+    from chalk.features.dataframe import DataFrame
 
 try:
     import numpy as np
 except:
     np = None
 
-if TYPE_CHECKING:
-    import polars as pl
-
 __all__ = ["pyarrow_to_primitive", "pyarrow_to_polars", "rich_to_pyarrow"]
 
 
-def _is_features_cls(typ: Any) -> bool:
+def _is_features_cls(typ: Any) -> TypeGuard[type[Features]]:
     return getattr(typ, "__chalk_feature_set__", False)
 
 
-def _is_chalk_dataframe(typ: Any):
+def _is_chalk_dataframe(typ: Any) -> TypeGuard[type[DataFrame]]:
     from chalk.features.dataframe import DataFrame
 
     unwrapped = unwrap_optional_and_annotated_if_needed(typ)
@@ -42,7 +46,7 @@ def _is_chalk_dataframe(typ: Any):
 
 
 def rich_to_pyarrow(
-    python_type: Type, name: str, in_struct: bool = False, parent_namespaces: Tuple[str] = None
+    python_type: Type, name: str, in_struct: bool = False, parent_namespaces: Tuple[str, ...] = ()
 ) -> pa.DataType:
     """
     Recursively convert a python type to a PyArrow dtype.
@@ -61,21 +65,21 @@ def rich_to_pyarrow(
         # Handling namedtuples above as structs before tuples, to ensure a namedtuple is not treated like
         # a list
         args = get_args(python_type)
-        if origin == Annotated:
+        if origin in (Annotated, getattr(typing, "Annotated", Annotated)):
             if len(args) < 1:
                 raise TypeError(
-                    "Annotated types must contain the underlying type as the first argument -- e.g. Annotated[int, 'annotation']"
+                    "Annotated types must contain the underlying type as the first argument -- e.g. Annotated[int, 'annotation']."
                 )
             return rich_to_pyarrow(args[0], name, in_struct=in_struct, parent_namespaces=parent_namespaces)
         if origin in (list, List, set, Set, frozenset, FrozenSet):
             typ_name = origin.__name__
             if len(args) == 0:
                 raise TypeError(
-                    f"{typ_name} features must be annotated with the type of the element -- e.g. {typ_name}[int]"
+                    f"{typ_name} features must be annotated with the type of the element -- e.g. {typ_name}[int]."
                 )
             if len(args) > 1:
                 raise TypeError(
-                    f"{typ_name} annotations should only take one argument -- e.g. {typ_name}[int]. Instead, got {typ_name}[{', '.join(args)}]"
+                    f"{typ_name} annotations should only take one argument -- e.g. {typ_name}[int]. Instead, got {typ_name}[{', '.join(args)}]."
                 )
             arg = args[0]
             return pa.large_list(
@@ -84,7 +88,7 @@ def rich_to_pyarrow(
         if origin in (tuple, Tuple):
             if len(args) == 0:
                 raise TypeError(
-                    "Tuple features must be annotated with the type of the tuple element -- e.g. `Tuple[int, ...]`"
+                    "Tuple features must be annotated with the type of the tuple element -- e.g. `Tuple[int, ...]`."
                 )
             if len(args) == 2 and args[1] is ...:
                 # Treat a variable-sized, homogenous tuple like a list
@@ -95,12 +99,34 @@ def rich_to_pyarrow(
             raise TypeError(
                 (
                     "Tuple features must have a fixed type and be variable-length tuples (e.g. `Tuple[int, ...]`). "
-                    " If you would like a fixed-length of potentially different types, used a NamedTuple."
+                    "If you would like a fixed-length of potentially different types, used a NamedTuple."
                 )
             )
+
+        if origin in (dict, Dict):
+            if len(args) != 2:
+                raise TypeError(
+                    "Dict features must be annotated with the type of the key and value -- e.g. `Dict[str, int]`."
+                )
+
+            key_type, value_type = args
+            if is_optional(key_type):
+                raise TypeError(f"Dict keys cannot be optional, found: `{key_type}`")
+            if not isinstance(key_type, type):
+                raise TypeError(f"Dict keys must be annotated with a `type`, found `{key_type}`")
+
+            try:
+                pa_value_type = rich_to_pyarrow(value_type, "value")
+            except TypeError as e:
+                raise TypeError(f"Failed to determine type of dict value from the annotation `{value_type}`: {e}")
+            else:
+                value_field = pa.field(name="value", type=pa_value_type, nullable=is_optional(value_type))
+
+            return pa.map_(key_type=rich_to_pyarrow(key_type, "key"), item_type=value_field)
+
         raise TypeError(f"Unsupported varardic type annotation: {origin}")
     else:
-        if python_type in [list]:
+        if python_type in (list, List, set, Set, frozenset, FrozenSet, tuple, Tuple):
             raise TypeError(
                 (
                     f"Unable to determine the PyArrow type for field '{name}' with type `{python_type}`. "
@@ -138,7 +164,7 @@ def rich_to_pyarrow(
         or is_namedtuple(python_type)
         or is_typeddict(python_type)
         or attrs.has(python_type)
-        or issubclass(python_type, BaseModel)
+        or is_pydantic_basemodel(python_type)
     ):
         annotations = cached_get_type_hints(python_type)
         fields: List[pa.Field] = []
@@ -148,21 +174,23 @@ def rich_to_pyarrow(
             )
             fields.append(pa.field(field_name, underlying_dtype))
         return pa.struct(fields)
-
     if _is_chalk_dataframe(python_type):
-        assert python_type.namespace is not None
-        parent_namespaces = (parent_namespaces or ()) + (python_type.namespace,)
-        column_fields = []
-        for col in python_type.columns:
-            if _is_infinitely_recursive_features_class(parent_namespaces, col.typ.parsed_annotation):
-                continue
-            else:
-                column_fields.append((col.fqn, col.typ.parsed_annotation))
-        pa_fields = [
-            pa.field(fqn, rich_to_pyarrow(typ, name=fqn, in_struct=True, parent_namespaces=parent_namespaces))
-            for fqn, typ in column_fields
-        ]
-        return pa.large_list(pa.struct(pa_fields))
+        if python_type.__pydantic_model__ is None:
+            assert python_type.namespace is not None
+            parent_namespaces = (parent_namespaces or ()) + (python_type.namespace,)
+            column_fields = []
+            for col in python_type.columns:
+                if _is_infinitely_recursive_features_class(parent_namespaces, col.typ.parsed_annotation):
+                    continue
+                else:
+                    column_fields.append((col.fqn, col.typ.parsed_annotation))
+            pa_fields = [
+                pa.field(fqn, rich_to_pyarrow(typ, name=fqn, in_struct=True, parent_namespaces=parent_namespaces))
+                for fqn, typ in column_fields
+            ]
+            return pa.large_list(pa.struct(pa_fields))
+        else:
+            return pa.large_list(rich_to_pyarrow(python_type=python_type.__pydantic_model__, name="item"))
     if issubclass(python_type, str):
         return pa.large_utf8()
     if issubclass(python_type, bool):
@@ -199,7 +227,7 @@ def rich_to_pyarrow(
     raise TypeError(
         (
             f"Unable to determine the PyArrow type for field '{name}' with type `{python_type}`. "
-            "Please set the `dtype` attribute when defining the feature"
+            "Please set the `dtype` attribute when defining the feature."
         )
     )
 
@@ -235,17 +263,33 @@ def pyarrow_to_primitive(pyarrow_typ: pa.DataType, name: str) -> Type[TPrimitive
         return cast(Type[TPrimitive], List[pyarrow_to_primitive(underlying, name=f"{name}[]")])
     if pa.types.is_struct(pyarrow_typ):
         pyarrow_typ = cast(pa.StructType, pyarrow_typ)
-        # Dynamically making a TypedDict so members will be parsed with the correct types by cattrs
+
         annotations: Dict[str, Type[TPrimitive]] = {}
         schema = pa.schema(pyarrow_typ)
         for sub_name, sub_typ in zip(schema.names, schema.types):
-            annotations[sub_name] = pyarrow_to_primitive(sub_typ, name=f"{name}.{sub_name}")
-        typed_dict_cls = TypedDict(f"__chalk_struct__{name}", annotations)  # type: ignore
-        return cast(Type[TPrimitive], typed_dict_cls)
-    raise TypeError(f"Unsupported PyArrow type {pyarrow_typ} for field {name}")
+            field_type = pyarrow_to_primitive(sub_typ, name=f"{name}.{sub_name}")
+            # TODO: CHA-2850
+            # if pyarrow_typ.field(sub_name).nullable:
+            #     field_type = typing.Optional[field_type]
+            annotations[sub_name] = field_type
+        struct_type = ChalkStructType(f"__chalk_struct__{name}", (object,), annotations)
+        return cast(Type[TPrimitive], struct_type)
+    if pa.types.is_map(pyarrow_typ):
+        pyarrow_typ = cast(pa.MapType, pyarrow_typ)
+        value_typ = pyarrow_to_primitive(pyarrow_typ.item_type, name=f"{name}::value")
+        if pyarrow_typ.item_field.nullable:
+            value_typ = typing.Optional[value_typ]
+        return cast(
+            Type[TPrimitive],
+            Dict[
+                pyarrow_to_primitive(pyarrow_typ.key_type, name=f"{name}::key"),
+                value_typ,
+            ],
+        )
+    raise TypeError(f"Unsupported PyArrow type '{pyarrow_typ}' for field '{name}'.")
 
 
-def _is_infinitely_recursive_features_class(parent_namespaces: Tuple[str, ...], typ: Type) -> bool:
+def _is_infinitely_recursive_features_class(parent_namespaces: Tuple[str, ...], typ: object) -> bool:
     """
     If a python type is a Features Class, make sure it's not an infinite loop of nested has-manys or has-ones.
     For instance, if there is a has-one relationship between Card and Customer, trying to serialize Card.customer.card
@@ -254,12 +298,23 @@ def _is_infinitely_recursive_features_class(parent_namespaces: Tuple[str, ...],
     :param typ: Same format as Feature.typ.parsed_annotation
     :return: True if `typ` is a nested has-one/many and its namespace is one of the parent_namespaces
     """
+    typ = unwrap_optional_and_annotated_if_needed(typ)
     if not (_is_chalk_dataframe(typ) or _is_features_cls(typ)):
         return False
     assert typ.namespace is not None
     return typ.namespace in parent_namespaces
 
 
+def is_map_in_dtype_tree(dtype: pa.DataType) -> bool:
+    if isinstance(dtype, pa.MapType):
+        return True
+    if isinstance(dtype, pa.StructType):
+        return any(is_map_in_dtype_tree(x.type) for x in dtype)
+    if isinstance(dtype, (pa.ListType, pa.LargeListType, pa.FixedSizeListType)):
+        return is_map_in_dtype_tree(dtype.value_type)
+    return False
+
+
 def pyarrow_to_polars(pa_type: pa.DataType, name: str | None = None) -> pl.PolarsDataType:
     """Convert a PyArrow data type into a Polars DataType
 
@@ -326,4 +381,13 @@ def pyarrow_to_polars(pa_type: pa.DataType, name: str | None = None) -> pl.Polar
             for sub_name, sub_typ in zip(schema.names, schema.types)
         ]
         return pl.Struct(fields)
-    raise TypeError(f"Unsupported PyArrow type {pa_type}" + (f" for field {name}" if name is not None else ""))
+    if pa.types.is_decimal(pa_type):
+        return pl.Decimal(precision=pa_type.precision, scale=pa_type.scale)
+    if pa.types.is_map(pa_type):
+        pa_type = cast(pa.MapType, pa_type)
+        fields = [
+            pl.Field("key", pyarrow_to_polars(pa_type.key_type, name=f"{name}::key")),
+            pl.Field("value", pyarrow_to_polars(pa_type.item_type, name=f"{name}::value")),
+        ]
+        return pl.List(pl.Struct(fields))
+    raise TypeError(f"Unsupported PyArrow type '{pa_type}'{name and f' for field {name}' or ''}.")
