@@ -11,6 +11,7 @@ import pathlib
 import warnings
 from datetime import date, datetime, timedelta, timezone
 from typing import (
+    IO,
     TYPE_CHECKING,
     Any,
     Callable,
@@ -24,39 +25,51 @@ from typing import (
     Sequence,
     Tuple,
     Type,
+    TypeVar,
     Union,
     cast,
     overload,
 )
 
+import packaging.version
 import pyarrow as pa
-from pydantic import BaseModel
-from typing_extensions import Self
 
 from chalk.features._chalkop import Aggregation
 from chalk.features._encoding.missing_value import MissingValueStrategy
-from chalk.features.dataframe._filters import convert_filters_to_pl_expr, filter_data_frame
+from chalk.features._encoding.pyarrow import is_map_in_dtype_tree
+from chalk.features.dataframe._filters import convert_filters_to_pl_expr, dataframe_missing_key_error, filter_data_frame
 from chalk.features.dataframe._validation import validate_df_schema, validate_nulls
 from chalk.features.feature_field import Feature
 from chalk.features.feature_wrapper import FeatureWrapper, ensure_feature, unwrap_feature
-from chalk.features.filter import Filter, _get_filter_now
+from chalk.features.filter import Filter, get_filter_now
+from chalk.features.pseudofeatures import PSEUDONAMESPACE, Distance
 from chalk.features.underscore import Underscore
 from chalk.features.underscore_features import parse_underscore_in_context
 from chalk.utils.collections import ensure_tuple, get_unique_item
+from chalk.utils.df_utils import (
+    pa_array_to_pl_series,
+    pa_cast,
+    pa_table_to_pl_df,
+    read_parquet,
+    recursive_convert_map_primitive,
+)
 from chalk.utils.duration import Duration, parse_chalk_duration
 from chalk.utils.missing_dependency import missing_dependency_exception
+from chalk.utils.pydanticutil.pydantic_compat import is_pydantic_basemodel
 
 if TYPE_CHECKING:
     import pandas as pd
     import polars as pl
+    from pydantic import BaseModel
 
     from chalk.features.feature_set import Features
-
 else:
     try:
-        import pandas as pd
+        from pydantic.v1 import BaseModel
     except ImportError:
-        pd = None
+        from pydantic import BaseModel
+
+TIntStr = TypeVar("TIntStr", int, str)
 
 
 class DataFrameMeta(type):
@@ -89,6 +102,7 @@ class DataFrameMeta(type):
         new_references_feature_set: Optional[Type[Features]] = None
         new_columns: List[Feature] = []
         pydantic_model = None
+        limit = None
         for a in item:
             if isinstance(a, Underscore):
                 if pydantic_model:
@@ -97,7 +111,7 @@ class DataFrameMeta(type):
                 else:
                     context = cls.references_feature_set
                     is_pydantic = False
-                a = parse_underscore_in_context(exp=x, context=context, is_pydantic=is_pydantic)
+                a = parse_underscore_in_context(exp=a, context=context, is_pydantic=is_pydantic)
             if isinstance(a, Filter):
                 new_filters.append(a)
             elif isinstance(a, type) and issubclass(a, Features):
@@ -109,14 +123,20 @@ class DataFrameMeta(type):
             elif isinstance(a, Feature):
                 new_columns.append(a)
             elif isinstance(a, FeatureWrapper):
-                new_columns.append(a._chalk_feature)
+                new_columns.append(unwrap_feature(a))
             elif isinstance(a, bool):
                 # If we encounter a bool, that means we are evaluating the type annotation before
                 # the ResolverAstParser had a chance to extract the source and rewrite the and/or/in operations
                 # into expressions that return filters instead of booleans
                 # This function will be called again for this annotation, so we can ignore it for now.
                 pass
-            elif inspect.isclass(a) and issubclass(a, BaseModel):
+            elif isinstance(a, slice):
+                if a.start is not None and a.start != 0:
+                    raise ValueError(f"The slice expression {a} with an explicit start is not supported.")
+                if a.step is not None and a.step != 1:
+                    raise ValueError(f"The slice expression {a} with an explicit step is not supported")
+                limit = a.stop
+            elif inspect.isclass(a) and is_pydantic_basemodel(a):
                 pydantic_model = a
             else:
                 raise TypeError(f"Invalid type for DataFrame[{a}]: {type(a)}")
@@ -136,6 +156,7 @@ class DataFrameMeta(type):
             __columns__ = tuple(new_columns)
             __references_feature_set__ = new_references_feature_set
             __pydantic_model__ = pydantic_model
+            __limit__ = limit
 
             def __eq__(self, other: object):
                 if not isinstance(other, SubclassedDataFrame):
@@ -204,11 +225,20 @@ class DataFrameMeta(type):
     @property
     def namespace(cls) -> str:
         cls = cast(Type[DataFrame], cls)
-        namespaces = [x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in cls.columns]
+        namespaces = (x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in cls.columns)
         # Remove the pseudo-columns
-        namespaces = [x for x in namespaces if not x.startswith("__chalk__")]
+        namespaces = (x for x in namespaces if not x.startswith("__chalk__"))
         return get_unique_item(namespaces, f"dataframe {cls.__name__} column namespaces")
 
+    def _parse_underscore(self, exp: Underscore, pydantic_model: Optional[BaseModel] = None) -> Any:
+        if pydantic_model:
+            context = self
+            is_pydantic = True
+        else:
+            context = self.references_feature_set
+            is_pydantic = False
+        parse_underscore_in_context(exp=exp, context=context, is_pydantic=is_pydantic)
+
 
 class DataFrame(metaclass=DataFrameMeta):
     filters: ClassVar[Tuple[Filter, ...]] = ()
@@ -217,12 +247,21 @@ class DataFrame(metaclass=DataFrameMeta):
     references_feature_set: Optional[Type[Features]]  # set via a @property on the metaclass
     __references_feature_set__: ClassVar[Optional[Type[Features]]] = None
     __pydantic_model__: ClassVar[Optional[Type[BaseModel]]] = None
+    __limit__: int | None = None
+    """The maximum number of rows to return"""
+
+    if TYPE_CHECKING:
+
+        def __class_getitem__(cls, cols: Any) -> Type[DataFrame]:
+            ...
 
     def __init__(
         self,
         data: Union[
             Dict[Union[str, Feature, Any], Sequence[Any]],
             Sequence[Union[Features, Any]],
+            pa.Table,
+            pa.RecordBatch,
             pl.DataFrame,
             pl.LazyFrame,
             pd.DataFrame,
@@ -294,6 +333,7 @@ class DataFrame(metaclass=DataFrameMeta):
         ...     "user.last": ["Wu", "Xi"]
         ... }))
         """
+        super().__init__()
         from chalk.features.feature_set import Features, FeatureSetBase
 
         try:
@@ -311,13 +351,23 @@ class DataFrame(metaclass=DataFrameMeta):
                 DeprecationWarning("ChalkDataFrameImpl(pandas_dataframe=...) has been renamed to DataFrame(data=...)")
             )
             data = pandas_dataframe
-        if pd is not None and isinstance(data, pd.DataFrame):
-            # Convert the columns to root fqn strings
-            # str(Feature) and str(FeatureWrapper) return the root fqns
-            data = data.rename(columns={k: str(k) for k in data.columns})
-            assert isinstance(data, pd.DataFrame)
-            data.columns = data.columns.astype("string")
-            data = pl.from_pandas(data)
+        try:
+            import pandas as pd
+        except ImportError:
+            pass
+        else:
+            if isinstance(data, pd.DataFrame):
+                # Convert the columns to root fqn strings
+                # str(Feature) and str(FeatureWrapper) return the root fqns
+                data = data.rename(columns={k: str(k) for k in data.columns})
+                assert isinstance(data, pd.DataFrame)
+                data.columns = data.columns.astype("string")
+                data = pl.from_pandas(data)
+        if isinstance(data, pa.RecordBatch):
+            data = pa.Table.from_batches([data])
+        if isinstance(data, pa.Table):
+            data = pa_table_to_pl_df(data)
+            assert isinstance(data, pl.DataFrame)
         if not isinstance(data, (pl.LazyFrame, pl.DataFrame)):
             if isinstance(data, (collections.abc.Sequence)) and not isinstance(data, str):
                 # If it is a sequence, it could be a sequence of feature classes instances
@@ -365,6 +415,9 @@ class DataFrame(metaclass=DataFrameMeta):
                 for k, v in data.items():
                     str_k = str(k)
                     feature = Feature.from_root_fqn(str_k)
+                    if feature.is_distance_pseudofeature:
+                        feature = Distance
+                        str_k = str(Distance)
 
                     if feature.is_has_one:
                         warnings.warn(
@@ -395,8 +448,7 @@ class DataFrame(metaclass=DataFrameMeta):
                                     f"of type `{feature.converter.pyarrow_dtype}`.{type_error}"
                                 )
                             ) from e
-                        series = pl.from_arrow(pa_array)
-                        assert isinstance(series, pl.Series)
+                        series = pa_array_to_pl_series(pa_array)
                     else:
                         series = v
                     new_data_dict[str_k] = series
@@ -410,11 +462,6 @@ class DataFrame(metaclass=DataFrameMeta):
         # Rename / validate that all column names are root fqns
         if self._pydantic_model is None:
             self.columns = tuple(Feature.from_root_fqn(str(c)) for c in underlying.columns)
-            # n.b. the goal here is to normalize "feature" and "str" columns to "str" columns
-            if any(not isinstance(c, str) for c in underlying.columns):
-                underlying = underlying.rename(
-                    {original_c: new_c.root_fqn for (original_c, new_c) in zip(underlying.columns, self.columns)}
-                )
         else:
             self.columns = ()
 
@@ -424,13 +471,12 @@ class DataFrame(metaclass=DataFrameMeta):
             if verify_validity:
                 underlying = validate_nulls(underlying, missing_value_strategy=missing_value_strategy)
 
-        if isinstance(underlying, pl.DataFrame):
-            underlying = underlying.lazy()
-        self._underlying = underlying
+        self._distance_feature_fqns = {x.root_fqn for x in self.columns if x.is_distance_pseudofeature}
+        self._underlying = pl.LazyFrame()
+        self._swap_underlying(underlying.lazy())
 
         # Remove the pseudo-features when determining the namespace
-        namespaces = (x.path[0].parent.namespace if len(x.path) > 0 else x.namespace for x in self.columns)
-        namespaces_set = set(x for x in namespaces if not x.startswith("__chalk__"))
+        namespaces_set = {x.root_namespace for x in self.columns if not x.root_namespace == "__chalk__"}
         if len(namespaces_set) == 1:
             self.namespace = get_unique_item(namespaces_set, "DataFrame column namespaces")
             self.references_feature_set = FeatureSetBase.registry[self.namespace]
@@ -439,10 +485,24 @@ class DataFrame(metaclass=DataFrameMeta):
             self.namespace = None
             self.references_feature_set = None
 
+    def _swap_underlying(self, underlying: pl.LazyFrame):
+        """Swap the underlying lazyframe, without validation. Equivalent to DataFrame(...) for an identical LazyFrame of the same shape
+
+        This is a private method used internally by Chalk.
+        """
+        import polars as pl
+
+        rename_map = {x: Distance.fqn for x in underlying.columns if x in self._distance_feature_fqns}
+
+        underlying = underlying.rename(rename_map)
+        if len(rename_map) > 0:
+            underlying = underlying.sort(pl.col(Distance.fqn))
+        self._underlying = underlying
+
     #######################
     # Filtering / Selecting
     #######################
-    def __getitem__(self, item: Any) -> DataFrame:
+    def __getitem__(self, item: Any) -> Union[Features, DataFrame]:
         """Filter the rows of a `DataFrame` or project out columns.
 
         You can select columns out of a `DataFrame` from
@@ -516,20 +576,7 @@ class DataFrame(metaclass=DataFrameMeta):
             item = parse_dataframe_getitem()
 
         # Resolve underscores to be within the dataframe's context
-        # Set columns based on existence of pydantic model
-        if self._pydantic_model:
-            context = self
-            is_pydantic = True
-        else:
-            context = self.references_feature_set
-            is_pydantic = False
-
-        item = tuple(
-            parse_underscore_in_context(exp=x, context=context, is_pydantic=is_pydantic)
-            if isinstance(x, Underscore)
-            else x
-            for x in ensure_tuple(item)
-        )
+        item = tuple(self._parse_underscore(exp=x) if isinstance(x, Underscore) else x for x in ensure_tuple(item))
         assert all(not isinstance(x, Underscore) for x in item)
         if len(item) == 1:
             item = item[0]
@@ -542,6 +589,9 @@ class DataFrame(metaclass=DataFrameMeta):
             )
         elif all(isinstance(col, str) for col in ensure_tuple(item)):
             # Select the columns with `.select()` since they're by name.
+            key_error_or_none = dataframe_missing_key_error(ensure_tuple(item), self._underlying.columns)
+            if key_error_or_none is not None:
+                raise key_error_or_none
             materialized = self._materialize()
             df = materialized.select(ensure_tuple(item))
             return DataFrame(df, pydantic_model=self._pydantic_model, verify_validity=False)
@@ -621,8 +671,17 @@ class DataFrame(metaclass=DataFrameMeta):
 
             cols.append(operation.fn(c).alias(str(alias)))
 
+        data = self._underlying.lazy()
+
+        if packaging.version.parse(pl.__version__) <= packaging.version.parse("0.19.0"):
+            data = data.groupby(groupby)
+        else:
+            data = data.group_by(groupby)
+
+        data = data.agg(cols).collect()
+
         return DataFrame(
-            self._underlying.lazy().groupby(groupby).agg(cols).collect(),
+            data,
             convert_dtypes=self._convert_dtypes,
             pydantic_model=self._pydantic_model,
         )
@@ -690,17 +749,19 @@ class DataFrame(metaclass=DataFrameMeta):
 
         if isinstance(base, (date, datetime)):
             epoch = datetime.utcfromtimestamp(0)
-            if base.tzinfo is not None:
-                epoch = epoch.replace(tzinfo=base.tzinfo)
+            if isinstance(base, datetime):
+                if base.tzinfo is not None:
+                    epoch = epoch.replace(tzinfo=base.tzinfo)
             base = (base - epoch).total_seconds() * 1_000_000.0
 
-        max_value = underlying.select(col_str).max().collect().item()
-        min_value = underlying.select(col_str).min().collect().item()
+        max_value: float = underlying.select(col_str).max().collect().item()
+        min_value: float = underlying.select(col_str).min().collect().item()
 
         if base is None:
             base = min_value - eps
 
         if bin_width is None:
+            assert nbins is not None
             bin_width = (max_value - base) / nbins
 
         if nbins is None:
@@ -728,7 +789,7 @@ class DataFrame(metaclass=DataFrameMeta):
         self,
         index: Union[Feature, Any],
         agg: Mapping[Any, Any],
-        every: Optional[Union[str, timedelta]],
+        every: Union[str, timedelta],
         group: Mapping[Union[Feature, Any], Union[Feature, Any]] | None = None,
         period: Optional[Union[str, timedelta]] = None,
         offset: Optional[Union[str, timedelta]] = None,
@@ -850,7 +911,7 @@ class DataFrame(metaclass=DataFrameMeta):
         namespace = namespaces.pop() if len(namespaces) == 1 else None
         timestamp_feature = None if namespace is None else FeatureSetBase.registry[namespace].__chalk_ts__
 
-        now = _get_filter_now()
+        now = get_filter_now()
         cols: List[pl.Expr] = []
         operation: Aggregation
         for alias, operation in agg.items():
@@ -1055,11 +1116,26 @@ class DataFrame(metaclass=DataFrameMeta):
 
         from chalk.features.feature_set import FeatureSetBase
 
+        new_c = {}
+        for k, v in c.items():
+            if isinstance(k, Underscore):
+                new_k = self._parse_underscore(exp=k)
+            else:
+                new_k = k
+            if isinstance(v, Underscore):
+                new_v = self._parse_underscore(exp=v)
+            else:
+                new_v = v
+            new_c[new_k] = new_v
+        c = new_c
+
+        assert all(not isinstance(k, Underscore) for k in c.keys())
+
         namespaces = {unwrap_feature(a).namespace for a in c.keys()}
         namespace = namespaces.pop() if len(namespaces) == 1 else None
         timestamp_feature = None if namespace is None else FeatureSetBase.registry[namespace].__chalk_ts__
 
-        now = _get_filter_now()
+        now = get_filter_now()
         cols: List[pl.Expr] = []
         for alias, operation in c.items():
             # if isinstance(operation, int):
@@ -1081,6 +1157,8 @@ class DataFrame(metaclass=DataFrameMeta):
                 cols.append(operation.fn(col).alias(str(alias)))
             elif isinstance(operation, FeatureWrapper):
                 cols.append(pl.col(str(operation)).alias(str(alias)))
+            elif isinstance(operation, pl.Expr):
+                cols.append(operation.alias(str(alias)))
             else:
                 cols.append(pl.lit(operation).alias(str(alias)))
 
@@ -1100,7 +1178,7 @@ class DataFrame(metaclass=DataFrameMeta):
     def from_dict(
         cls,
         data: Dict[Union[str, Feature, FeatureWrapper, Any], Sequence[Any]],
-    ) -> Self:
+    ) -> DataFrame:
         """Deprecated. Use DataFrame(...) instead."""
         warnings.warn(DeprecationWarning("DataFrame.from_dict(...) is deprecated. Instead, use DataFrame(...)"))
         df = cls(data)
@@ -1112,18 +1190,18 @@ class DataFrame(metaclass=DataFrameMeta):
         cls,
         data: Sequence[Features],
         /,
-    ) -> Self:
+    ) -> DataFrame:
         """Deprecated. Use DataFrame(...) instead."""
         ...
 
     @overload
     @classmethod
-    def from_list(cls, *data: Features) -> Self:
+    def from_list(cls, *data: Features) -> DataFrame:
         """Deprecated. Use DataFrame(...) instead."""
         ...
 
     @classmethod
-    def from_list(cls, *data: Union[Features, Sequence[Features]]) -> Self:
+    def from_list(cls, *data: Union[Features, Sequence[Features]]) -> DataFrame:
         """Deprecated. Use DataFrame(...) instead."""
         warnings.warn(DeprecationWarning("DataFrame.from_list(...) is deprecated. Instead, use DataFrame(...)"))
         if len(data) == 1 and isinstance(data[0], collections.abc.Sequence):
@@ -1160,14 +1238,11 @@ class DataFrame(metaclass=DataFrameMeta):
         table_uri: str,
         version: Optional[int] = None,
         columns: Optional[
-            Union[
-                Dict[str, Union[str, Feature, Any]],
-                Dict[int, Union[str, Feature, Any]],
-            ]
+            Dict[str, Union[str, Feature, Any]],
         ] = None,
         delta_table_options: Optional[Dict[str, Any]] = None,
         pyarrow_options: Optional[Dict[str, Any]] = None,
-    ) -> Self:
+    ) -> DataFrame:
         try:
             import polars as pl
         except ImportError:
@@ -1179,7 +1254,7 @@ class DataFrame(metaclass=DataFrameMeta):
             cols_to_select, _, new_columns = cls._parse_columns(columns)
 
         data = pl.read_delta(
-            table_uri=table_uri,
+            source=table_uri,
             version=version,
             columns=cols_to_select,
             storage_options=DataFrame._get_storage_options(),
@@ -1201,18 +1276,13 @@ class DataFrame(metaclass=DataFrameMeta):
             ]
         ] = None,
         use_pyarrow: bool = False,
-    ) -> Self:
-        try:
-            import polars as pl
-        except ImportError:
-            raise missing_dependency_exception("chalkpy[runtime]")
-
+    ) -> DataFrame:
         if columns is None:
             cols_to_select, _, new_columns = None, None, None
         else:
             cols_to_select, _, new_columns = cls._parse_columns(columns)
 
-        data = pl.read_parquet(
+        data = read_parquet(
             source=path,
             columns=cols_to_select,
             storage_options=DataFrame._get_storage_options(),
@@ -1223,14 +1293,27 @@ class DataFrame(metaclass=DataFrameMeta):
         return cls(data)
 
     @classmethod
+    @overload
     def _parse_columns(
         cls,
-        col_mapping: Union[
-            Dict[str, Union[str, Feature, FeatureWrapper, Any]],
-            Dict[int, Union[str, Feature, FeatureWrapper, Any]],
-        ],
-    ) -> Tuple[Union[List[str], List[int]], List[Type[pl.DataType]], List[str]]:
-        if not isinstance(col_mapping, dict):
+        col_mapping: dict[int, Union[str, Feature, FeatureWrapper, Any]],
+    ) -> Tuple[list[int], List[Type[pl.DataType]], List[str]]:
+        ...
+
+    @classmethod
+    @overload
+    def _parse_columns(
+        cls,
+        col_mapping: dict[str, Union[str, Feature, FeatureWrapper, Any]],
+    ) -> Tuple[list[str], List[Type[pl.DataType]], List[str]]:
+        ...
+
+    @classmethod
+    def _parse_columns(
+        cls,
+        col_mapping: dict[TIntStr, Union[str, Feature, FeatureWrapper, Any]],
+    ) -> Tuple[list[TIntStr], List[Type[pl.DataType]], List[str]]:
+        if not isinstance(col_mapping, dict):  # pyright: ignore[reportUnnecessaryIsInstance]
             raise ValueError(f"Invalid column mapping. Received '{type(col_mapping).__name__}'; expected dict")
         columns = []
         dtypes = []
@@ -1252,7 +1335,7 @@ class DataFrame(metaclass=DataFrameMeta):
                 Dict[int, Union[str, Feature, Any]],
             ]
         ] = None,
-    ) -> Self:
+    ) -> DataFrame:
         """Read a .csv file as a `DataFrame`.
 
         Parameters
@@ -1302,7 +1385,7 @@ class DataFrame(metaclass=DataFrameMeta):
     def read_avro(
         cls,
         path: Union[str, pathlib.Path],
-    ) -> Self:
+    ) -> DataFrame:
         """Read a `.avro` file as a `DataFrame`.
 
         Parameters
@@ -1336,22 +1419,34 @@ class DataFrame(metaclass=DataFrameMeta):
                 import s3fs
             except ImportError:
                 raise missing_dependency_exception("s3fs")
+            else:
+                del s3fs
             try:
                 import io
             except ImportError:
                 raise missing_dependency_exception("io")
             with fsspec.open(path, "rb") as f:
+                f = cast(IO[bytes], f)
                 buffer = io.BytesIO(initial_bytes=f.read())
                 data = pl.read_avro(buffer)
         else:
             data = pl.read_avro(source=path)
         return cls(data)
 
+    def _parse_underscore(self, exp: Underscore) -> Any:
+        if self._pydantic_model:
+            context = self
+            is_pydantic = True
+        else:
+            context = self.references_feature_set
+            is_pydantic = False
+        return parse_underscore_in_context(exp=exp, context=context, is_pydantic=is_pydantic)
+
     #############
     # Aggregation
     #############
 
-    def max(self) -> Self:
+    def max(self) -> DataFrame:
         """Compute the max value of each of the columns in the `DataFrame`.
         The resulting `DataFrame` will have a single row with the max value
         of each column.
@@ -1378,7 +1473,7 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         return DataFrame(self._underlying.max(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
-    def mean(self) -> Self:
+    def mean(self) -> DataFrame:
         """Compute the mean value of each of the columns in the `DataFrame`.
         The resulting `DataFrame` will have a single row with the mean value
         of each column.
@@ -1405,7 +1500,7 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         return DataFrame(self._underlying.mean(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
-    def median(self) -> Self:
+    def median(self) -> DataFrame:
         """Compute the median value of each of the columns in the `DataFrame`.
         The resulting `DataFrame` will have a single row with the median value
         of each column.
@@ -1432,7 +1527,7 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         return DataFrame(self._underlying.median(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
-    def min(self) -> Self:
+    def min(self) -> DataFrame:
         """Compute the min value of each of the columns in the `DataFrame`.
         The resulting `DataFrame` will have a single row with the min value
         of each column.
@@ -1459,7 +1554,7 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         return DataFrame(self._underlying.min(), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
-    def std(self, ddof: int = 1) -> Self:
+    def std(self, ddof: int = 1) -> DataFrame:
         """Compute the standard deviation of each of the columns in the `DataFrame`.
         The resulting `DataFrame` will have a single row with the standard deviation
         of each column.
@@ -1492,7 +1587,7 @@ class DataFrame(metaclass=DataFrameMeta):
         """
         return DataFrame(self._underlying.std(ddof), convert_dtypes=False, pydantic_model=self._pydantic_model)
 
-    def sum(self) -> Self:
+    def sum(self) -> DataFrame:
         """
         Compute the sum of each of the columns in the `DataFrame`.
         The resulting `DataFrame` will have a single row with the sum
@@ -1510,7 +1605,7 @@ class DataFrame(metaclass=DataFrameMeta):
             pydantic_model=self._pydantic_model,
         )
 
-    def var(self, ddof: int = 1) -> Self:
+    def var(self, ddof: int = 1) -> DataFrame:
         """Compute the variance of each of the columns in the `DataFrame`.
 
         Parameters
@@ -1559,8 +1654,12 @@ class DataFrame(metaclass=DataFrameMeta):
 
         if len(self) == 0:
             return True
-        if not all(isinstance(x, type) and issubclass(x, pl.Boolean) for x in self._underlying.dtypes):
-            raise TypeError("DataFrame.any() is not defined on a dataframe that contains non-boolean columns.")
+
+        if any(
+            (not isinstance(x, type) or not issubclass(x, pl.Boolean)) and x != pl.Boolean
+            for x in self._underlying.dtypes
+        ):
+            raise TypeError("DataFrame.any() is not defined on a DataFrame that contains non-boolean columns.")
         materialized = self._materialize()
         return all(col.all() for col in materialized.get_columns())
 
@@ -1642,7 +1741,7 @@ class DataFrame(metaclass=DataFrameMeta):
         *more_by: Union[str, Feature, Any],
         descending: Union[bool, Sequence[bool]] = False,
         nulls_last: bool = False,
-    ) -> Self:
+    ) -> DataFrame:
         """Sort the `DataFrame` by the given columns.
 
         Parameters
@@ -1675,10 +1774,10 @@ class DataFrame(metaclass=DataFrameMeta):
         1     2  2
         2     3  1
         """
-        by = tuple(str(x) for x in itertools.chain(ensure_tuple(by), more_by))
+        parsed_by = tuple(str(x) for x in itertools.chain(ensure_tuple(by), more_by))
         return DataFrame(
             self._underlying.sort(
-                by=by,
+                by=parsed_by,
                 descending=descending,
                 nulls_last=nulls_last,
             ),
@@ -1717,7 +1816,7 @@ class DataFrame(metaclass=DataFrameMeta):
         op: Union[str, Callable[[Any, Any], Any]],
         other: Union[DataFrame, pl.DataFrame, pd.DataFrame, Any],
         convert_dtypes: bool,
-    ) -> Self:
+    ) -> DataFrame:
         import polars as pl
 
         materialized = self._materialize()
@@ -1725,8 +1824,13 @@ class DataFrame(metaclass=DataFrameMeta):
             other = other.to_polars()
         if isinstance(other, pl.LazyFrame):
             other = other.collect()
-        if pd is not None and isinstance(other, pd.DataFrame):
-            other = pl.from_pandas(other)
+        try:
+            import pandas as pd
+        except ImportError:
+            pass
+        else:
+            if isinstance(other, pd.DataFrame):
+                other = pl.from_pandas(other)
         if op in ("eq", "ne", "lt", "le", "ge", "gt"):
             if isinstance(other, pl.DataFrame):
                 return self._perform_comp_df(op, other, convert_dtypes=convert_dtypes)
@@ -1749,7 +1853,7 @@ class DataFrame(metaclass=DataFrameMeta):
         op: str,
         other: pl.DataFrame,
         convert_dtypes: bool,
-    ) -> Self:
+    ) -> DataFrame:
         # There's a bug in the default polars implementation for comparisons -- see
         # https://github.com/pola-rs/polars/issues/5870
         import polars as pl
@@ -1815,10 +1919,14 @@ class DataFrame(metaclass=DataFrameMeta):
 
         return DataFrame(combined.select(expr), convert_dtypes=convert_dtypes, pydantic_model=self._pydantic_model)
 
-    def __eq__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
+    def __eq__(  # pyright: ignore[reportIncompatibleMethodOverride]
+        self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]
+    ):
         return self._perform_op("eq", other, convert_dtypes=False)
 
-    def __ne__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
+    def __ne__(  # pyright: ignore[reportIncompatibleMethodOverride]
+        self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]
+    ):
         return self._perform_op("ne", other, convert_dtypes=False)
 
     def __gt__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
@@ -1830,28 +1938,28 @@ class DataFrame(metaclass=DataFrameMeta):
     def __ge__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]):
         return self._perform_op("ge", other, convert_dtypes=False)
 
-    def __le__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> Self:
+    def __le__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> DataFrame:
         return self._perform_op("le", other, convert_dtypes=False)
 
-    def __add__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> Self:
+    def __add__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> DataFrame:
         return self._perform_op(operator.add, other, convert_dtypes=True)
 
-    def __sub__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> Self:
+    def __sub__(self, other: Union[DataFrame, pl.DataFrame, pl.LazyFrame, pd.DataFrame, Any]) -> DataFrame:
         return self._perform_op(operator.sub, other, convert_dtypes=True)
 
-    def __mul__(self, other: Union[int, float]) -> Self:
+    def __mul__(self, other: Union[int, float]) -> DataFrame:
         return self._perform_op(operator.mul, other, convert_dtypes=True)
 
-    def __truediv__(self, other: Union[int, float]) -> Self:
+    def __truediv__(self, other: Union[int, float]) -> DataFrame:
         return self._perform_op(operator.truediv, other, convert_dtypes=True)
 
-    def __floordiv__(self, other: Union[int, float]) -> Self:
+    def __floordiv__(self, other: Union[int, float]) -> DataFrame:
         return self._perform_op(operator.floordiv, other, convert_dtypes=True)
 
-    def __mod__(self, other: Union[int, float]) -> Self:
+    def __mod__(self, other: Union[int, float]) -> DataFrame:
         return self._perform_op(operator.mod, other, convert_dtypes=True)
 
-    def __pow__(self, other: Union[int, float]) -> Self:
+    def __pow__(self, other: Union[int, float]) -> DataFrame:
         return self._perform_op(operator.pow, other, convert_dtypes=True)
 
     ############
@@ -1897,7 +2005,7 @@ class DataFrame(metaclass=DataFrameMeta):
         pa_table = materialized.to_arrow()
         if self._convert_dtypes:
             pa_schema_fields: Dict[str, pa.DataType] = {}
-            new_names = []
+            new_names: List[str] = []
             for col_fqn in pa_table.column_names:
                 f = Feature.from_root_fqn(col_fqn)
                 output_col_name = col_fqn
@@ -1905,18 +2013,24 @@ class DataFrame(metaclass=DataFrameMeta):
                     output_col_name = f.name
                     new_names.append(output_col_name)
 
-                if f.typ.is_dataframe:
+                if f.is_has_many:
                     # Special case for DataFrames -- restrict the struct type to only the fields that are present
                     full_dtype = f.converter.pyarrow_dtype
+                    assert full_dtype is not None
                     present_field_names = [field.name for field in pa_table.schema.field(col_fqn).type.value_type]
                     new_fields = [full_dtype.value_type.field(name) for name in present_field_names]
-                    pa_schema_fields[output_col_name] = pa.large_list(pa.struct(new_fields))
+                    pa_type = pa.large_list(pa.struct(new_fields))
                 else:
-                    pa_schema_fields[output_col_name] = f.converter.pyarrow_dtype
+                    pa_type = f.converter.pyarrow_dtype
+                # For features within a has-many (user.transactions.amount), wrap them in a list[]
+                for path_obj in f.path:
+                    if path_obj.parent.is_has_many:
+                        pa_type = pa.large_list(pa_type)
+                pa_schema_fields[output_col_name] = pa_type
 
             if not prefixed:
                 pa_table = pa_table.rename_columns(new_names)
-            pa_table = pa_table.cast(pa.schema(pa_schema_fields))
+            pa_table = pa_cast(pa_table, pa.schema(pa_schema_fields))
         return pa_table
 
     def to_pandas(self, string_names: bool = False, prefixed: bool = True) -> pd.DataFrame:
@@ -1936,8 +2050,9 @@ class DataFrame(metaclass=DataFrameMeta):
         pandas.DataFrame
             The data formatted as a `pandas.DataFrame`.
         """
-
-        if pd is None:
+        try:
+            import pandas as pd
+        except:
             raise missing_dependency_exception("chalkpy[pandas]")
 
         def types_mapper(dtype: pa.DataType):
@@ -1981,16 +2096,16 @@ class DataFrame(metaclass=DataFrameMeta):
             raise ValueError("This method is not supported if the DataFrame spans multiple namespaces")
 
         if self._pydantic_model is not None:
-            ans = []
-            for row in self.to_pyarrow().to_pylist():
-                ans.append(self._pydantic_model(**row))
-            return ans
+            # This definitely violates typing -- we should not have a PydanticModel be part of the DataFrame api
+            return cast(Sequence["Features"], [self._pydantic_model(**row) for row in self.to_pyarrow().to_pylist()])
 
         ans: List[Features] = []
         for row in self.to_pyarrow().to_pylist():
             rooted_prefix_to_values: Dict[str, Dict[str, Any]] = {}
 
             for k, v in row.items():
+                if k.startswith(f"{PSEUDONAMESPACE}."):
+                    continue
                 rooted_prefix_split = k.split(".")[:-1]
                 for i in range(1, len(rooted_prefix_split) + 1):
                     rooted_prefix = ".".join(rooted_prefix_split[:i])
@@ -2001,24 +2116,43 @@ class DataFrame(metaclass=DataFrameMeta):
             sorted_sub_features = sorted(rooted_prefix_to_values.keys(), key=lambda x: len(x), reverse=True)
 
             for rooted_prefix in sorted_sub_features:
+                values_are_lists = False
+                if rooted_prefix != self.namespace:
+                    rooted_prefix_feature = Feature.from_root_fqn(rooted_prefix)
+                    values_are_lists = rooted_prefix_feature.is_has_many or rooted_prefix_feature.is_has_many_subfeature
+
                 values = rooted_prefix_to_values[rooted_prefix]
                 sub_kwargs: Dict[str, Any] = {}
                 for k, v in values.items():
                     feature = Feature.from_root_fqn(k)
-                    sub_kwargs[feature.attribute_name] = feature.converter.from_primitive_to_rich(v)
+                    if is_map_in_dtype_tree(feature.converter.pyarrow_dtype):
+                        v = recursive_convert_map_primitive(v, feature.converter.pyarrow_dtype)
+                    if values_are_lists:
+                        sub_kwargs[feature.attribute_name] = [feature.converter.from_primitive_to_rich(vv) for vv in v]
+                    else:
+                        sub_kwargs[feature.attribute_name] = feature.converter.from_primitive_to_rich(v)
                 if rooted_prefix == self.namespace:
+                    assert not values_are_lists, "Top level feature shouldn't be a has-many"
+                    assert self.namespace is not None
                     features_cls = FeatureSetBase.registry[self.namespace]
                     ans.append(features_cls(**sub_kwargs))
                 else:
                     rooted_prefix_split = rooted_prefix.split(".")
-                    features_cls = Feature.from_root_fqn(rooted_prefix).joined_class
-                    assert features_cls is not None
+                    feature_cls = Feature.from_root_fqn(rooted_prefix).joined_class
+                    assert feature_cls is not None
                     parent_rooted_prefix, feature_name = ".".join(rooted_prefix_split[:-1]), rooted_prefix_split[-1]
                     del feature_name  # unused
-                    rooted_prefix_to_values[parent_rooted_prefix][rooted_prefix] = features_cls(**sub_kwargs)
+                    if values_are_lists:
+                        feature_values = []
+                        ks, vs = zip(*sub_kwargs.items())  # [k1, k2], [v1[...], v2[...]]
+                        for row in zip(*vs):
+                            feature_values.append(feature_cls(**dict(zip(ks, row))))
+                        rooted_prefix_to_values[parent_rooted_prefix][rooted_prefix] = feature_values
+                    else:
+                        rooted_prefix_to_values[parent_rooted_prefix][rooted_prefix] = feature_cls(**sub_kwargs)
         return ans
 
-    def slice(self, offset: int = 0, length: Optional[int] = None) -> Self:
+    def slice(self, offset: int = 0, length: Optional[int] = None) -> DataFrame:
         """Slice the `DataFrame`.
 
         Parameters
