@@ -2,15 +2,18 @@ from __future__ import annotations
 
 import copy
 from collections.abc import Iterable
-from typing import Any, TypeVar, Union
+from typing import Any, Literal, TypeVar, Union, TYPE_CHECKING
 
 from chalk.features._chalkop import op
-from chalk.features.feature_field import Feature
 from chalk.features.filter import Filter
 from chalk.serialization.parsed_annotation import ParsedAnnotation
 from chalk.streams import get_name_with_duration
 from chalk.utils.collections import ensure_tuple
 
+if TYPE_CHECKING:
+    from chalk.features.feature_field import Feature
+
+
 T = TypeVar("T")
 
 
@@ -23,20 +26,20 @@ class FeatureWrapper:
 
     def __init__(self, feature: Feature) -> None:
         # Binding as a private variable as not to have naming conflicts user's use of __getattr__
+        super().__init__()
         self._chalk_feature = feature
 
-    def __add__(self, other):
+    def __add__(self, other: object):
+        if not isinstance(other, FeatureWrapper):
+            return NotImplemented
         f_self = unwrap_feature(self)
 
-        if isinstance(other, FeatureWrapper):
-            f_other = unwrap_feature(other)
-            if f_other.is_scalar and f_self.is_scalar:
-                if str == f_other.converter.rich_type == f_self.converter.rich_type:
-                    return op.concat_str(self, other)
-                else:
-                    return op.sum(self, other)
-
-        raise NotImplementedError(f"Addition not implemented for '{type(other).__name__}'")
+        f_other = unwrap_feature(other)
+        if f_other.is_scalar and f_self.is_scalar:
+            if str == f_other.converter.rich_type == f_self.converter.rich_type:
+                return op.concat_str(self, other)
+            else:
+                return op.sum(self, other)
 
     def __matmul__(self, other: int):
         return FeatureWrapper(self._chalk_feature.for_version(other))
@@ -76,6 +79,8 @@ class FeatureWrapper:
             if op == "!=" and not is_eq:
                 return False
             return NotImplemented  # GT / LT doesn't really make sense otherwise
+        if isinstance(other, type):
+            return False
         return Filter(self._chalk_feature, op, other)
 
     def __ne__(self, other: object):
@@ -113,7 +118,7 @@ class FeatureWrapper:
         raise TypeError(f"Feature {self} is not callable")
 
     def _chalk_get_windowed_feature(self, window: Union[str, int]):
-        if not isinstance(window, (str, int)):
+        if not isinstance(window, (str, int)):  # pyright: ignore[reportUnnecessaryIsInstance]
             raise TypeError("Window duration must be a string or an int")
 
         from chalk.features import FeatureSetBase
@@ -135,18 +140,46 @@ class FeatureWrapper:
         return getattr(parent, desired_attribute_name)
 
     def __getitem__(self, item: Any):
-        from chalk.features.dataframe import DataFrame
+        from chalk.features.feature_field import get_distance_feature_name, Feature
+        from chalk.features.pseudofeatures import Distance
+        from chalk.features.feature_set import FeatureSetBase
 
         if len(self._chalk_feature.window_durations) > 0:
             return self._chalk_get_windowed_feature(*ensure_tuple(item))
 
-        assert isinstance(self._chalk_feature.typ.parsed_annotation, type)
-        if self._chalk_feature.typ is not None and issubclass(self._chalk_feature.typ.parsed_annotation, DataFrame):
+        dataframe_typ = self._chalk_feature.typ.as_dataframe()
+        if dataframe_typ is not None:
+            item = list(ensure_tuple(item))
+            for i, x in enumerate(item):
+                if x is Distance:
+                    # If requesting a distance, we need to replace it with the distance pseudofeature
+                    filter = self._chalk_feature.join
+                    if filter is None:
+                        raise ValueError("The `Distance` feature can only be used with has-many relationship")
+                    if not filter.operation.startswith("is_near_"):
+                        raise ValueError("The `Distance` feature can only be used with a nearest neighbor join")
+                    assert isinstance(filter.lhs, Feature)
+                    assert isinstance(filter.rhs, Feature)
+                    local_feature = filter.lhs if filter.lhs.namespace == self._chalk_feature.namespace else filter.rhs
+                    foreign_feature = (
+                        filter.lhs if filter.lhs.namespace != self._chalk_feature.namespace else filter.rhs
+                    )
+                    assert local_feature != foreign_feature, "The local and foreign features must be different"
+                    key = get_distance_feature_name(
+                        local_namespace=local_feature.namespace,
+                        local_name=local_feature.name,
+                        local_hm_name=self._chalk_feature.name,
+                        op=filter.operation,
+                        foreign_name=foreign_feature.name,
+                    )
+                    x = next(f for f in FeatureSetBase.registry[foreign_feature.namespace].features if f.name == key)
+                    item[i] = x
+            item = tuple(item)
+
             f_copy = FeatureWrapper(copy.copy(self._chalk_feature))
-            f_copy._chalk_feature.typ = ParsedAnnotation(underlying=self._chalk_feature.typ.parsed_annotation[item])
+            f_copy._chalk_feature.typ = ParsedAnnotation(underlying=dataframe_typ[item])
             if (
                 hasattr(f_copy._chalk_feature, "path")
-                and f_copy._chalk_feature.path is not None
                 and len(f_copy._chalk_feature.path) > 0
             ):
                 f_copy._chalk_feature.path[-1].child = unwrap_feature(f_copy)
@@ -175,7 +208,44 @@ class FeatureWrapper:
                 assert isinstance(f, Feature), f"HasOne feature {f} does not inherit from FeaturesBase"
                 if f.attribute_name == item:
                     return FeatureWrapper(self._chalk_feature.copy_with_path(f))
-        raise AttributeError(f"'{self}' has no attribute '{item}'")
+
+            assert self._chalk_feature.features_cls is not None
+            self._chalk_feature.features_cls.__chalk_error_builder__.invalid_attribute(
+                self._chalk_feature.joined_class.namespace,
+                item=item,
+                candidates=[f.name for f in self._chalk_feature.joined_class.features],
+                back=1,
+            )
+
+        assert self._chalk_feature.features_cls is not None
+        self._chalk_feature.features_cls.__chalk_error_builder__.invalid_attribute(
+            self._chalk_feature.features_cls.namespace,
+            item=item,
+            candidates=[],
+            back=1,
+        )
+
+    def is_near(self, item: Any, metric: Literal["l2", "cos", "ip"] = "l2") -> Filter:
+        other = unwrap_feature(item)
+        self_vector = self._chalk_feature.typ.as_vector()
+        if self_vector is None:
+            raise TypeError(
+                f"Nearest neighbor relationships are only supported for vector features. Feature '{self._chalk_feature.root_fqn}' is not a vector."
+            )
+        other_vector = other.typ.as_vector()
+        if other_vector is None:
+            raise TypeError(
+                f"Nearest neighbor relationships are only supported for vector features. Feature '{other.root_fqn}' is not a vector."
+            )
+        if self._chalk_feature.converter.pyarrow_dtype != other.converter.pyarrow_dtype:
+            raise TypeError(
+                (
+                    f"Nearest neighbor relationships are only supported if both vectors have the same data type and dimensions. "
+                    f"Feature '{self._chalk_feature.root_fqn}' is of type `{self._chalk_feature.converter.pyarrow_dtype}` "
+                    f" while feature '{other.root_fqn}' is `{other.converter.pyarrow_dtype}`."
+                )
+            )
+        return Filter(self._chalk_feature, f"is_near_{metric}", other)
 
 
 def unwrap_feature(maybe_feature_wrapper: Any) -> Feature:
@@ -191,8 +261,10 @@ def unwrap_feature(maybe_feature_wrapper: Any) -> Feature:
         type(FooBar.foo) is FeatureWrapper
         type(unwrap_feature(FooBar.foo)) is Feature
     """
+    from chalk.features.feature_field import Feature
+
     if isinstance(maybe_feature_wrapper, FeatureWrapper):
-        maybe_feature_wrapper = maybe_feature_wrapper._chalk_feature
+        maybe_feature_wrapper = maybe_feature_wrapper._chalk_feature  # pyright: ignore[reportPrivateUsage]
     if isinstance(maybe_feature_wrapper, Feature):
         return maybe_feature_wrapper
     raise TypeError(
@@ -201,6 +273,8 @@ def unwrap_feature(maybe_feature_wrapper: Any) -> Feature:
 
 
 def ensure_feature(feature: Union[str, Feature, FeatureWrapper, Any]) -> Feature:
+    from chalk.features.feature_field import Feature
+
     if isinstance(feature, str):
         return Feature.from_root_fqn(feature)
     if isinstance(feature, FeatureWrapper):
