@@ -4,19 +4,35 @@ import collections.abc
 import contextlib
 import contextvars
 import dataclasses
-import threading
 from datetime import datetime, timedelta, timezone
 from types import TracebackType
-from typing import TYPE_CHECKING, Any, Hashable, List, Set, Type
+from typing import TYPE_CHECKING, Any, Hashable, Set, Type
 
 if TYPE_CHECKING:
     from chalk.features.feature_field import Feature
 
+_DISABLE_FILTER_BOOL: contextvars.ContextVar[bool] = contextvars.ContextVar("DISABLE_FILTER_BOOL", default=False)
+
+
+@contextlib.contextmanager
+def disable_filter_bool(value: bool = True):
+    token = _DISABLE_FILTER_BOOL.set(value)
+    try:
+        yield
+    finally:
+        _DISABLE_FILTER_BOOL.reset(token)
+
+
+class FilterBoolNotAllowedError(ValueError):
+    """Exception that is raised if Filter.__bool__ is called when DISABLE_FILTER_BOOL is True"""
+
 
 class Filter:
     def __init__(self, lhs: Any, operation: str, rhs: Any):
         from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
 
+        super().__init__()
+
         # Feature or other could be another feature, filter, featuretime, literal
         # Other could also be a sequence (in the case of operation = "in")
         self.operation = operation
@@ -61,6 +77,15 @@ class Filter:
     def __repr__(self):
         return f"Filter({self.lhs} {self.operation} {self.rhs})"
 
+    def __bool__(self):
+        if _DISABLE_FILTER_BOOL.get():
+            raise FilterBoolNotAllowedError(
+                "Joining filters together with `and` or `or` is not allowed in the current context"
+            )
+        else:
+            # By default custom class instances are always "True"
+            return True
+
 
 def _get_feature_time_index(explicit_index: Any):
     if explicit_index is None:
@@ -245,23 +270,65 @@ def after(
 _FILTER_NOW_CONTEXT_STACK: contextvars.ContextVar[datetime] = contextvars.ContextVar("_FILTER_NOW_CONTEXT_STACK")
 
 
-def _get_filter_now() -> datetime:
+def get_filter_now() -> datetime:
     return _FILTER_NOW_CONTEXT_STACK.get(datetime.now(tz=timezone.utc))
 
 
-class freeze_time(contextlib.AbstractContextManager):
+class freeze_time:
     def __init__(self, at: datetime):
+        """
+        Used to freeze the 'now' value used to execute filters like `after(days_ago=30)`.
+
+        Parameters
+        ----------
+        at
+            The time to freeze to. Must be timezone aware.
+
+        Examples
+        --------
+        >>> @online
+        ... def get_average_spend_30d(
+        ...     spend: User.cards[after(days_ago=30)],
+        ... ) -> User.average_spend_30d:
+        ...     return spend.mean()
+        >>> with freeze_time(datetime(2021, 1, 1, tzinfo=timezone.utc)):
+        >>>     now = datetime.now(tz=timezone.utc)
+        >>>     get_average_spend_30d(
+        ...         spend=DataFrame([
+        ...             Card(spend=10, ts=now - timedelta(days=31)),
+        ...             Card(spend=20, ts=now - timedelta(days=29)),
+        ...             Card(spend=30, ts=now - timedelta(days=28)),
+        ...         ])
+        ...     )
+        """
+        super().__init__()
         if at.tzinfo is None:
-            raise ValueError("Timestamps must have timezone information associated with them.")
+            raise ValueError("Timestamps used with 'freeze_time' must have timezone information associated with them.")
         self._time = at
         # Keep a list of previous contextvar tokens so that this context manager is re-entrant
         # (i.e. you can enter multiple times and then exit multiple times)
         self._token_stack = []
 
     def time(self) -> datetime:
+        """
+        Returns the current time that filters will use.
+
+        Returns
+        -------
+        datetime
+            The current time that filters will use.
+
+        Examples
+        --------
+        >>> with freeze_time(datetime(2021, 1, 1, tzinfo=timezone.utc)) as ft:
+        ...     assert ft.time() == datetime(2021, 1, 1, tzinfo=timezone.utc)
+        """
         return self._time
 
     def __enter__(self):
+        """The freeze_time class is a context manager, so it can be used with the `with` statement.
+        The `__enter__` method is called when entering the context manager.
+        """
         token = _FILTER_NOW_CONTEXT_STACK.set(self._time)
         self._token_stack.append(token)
         return self
@@ -272,6 +339,10 @@ class freeze_time(contextlib.AbstractContextManager):
         __exc_value: BaseException | None,
         __traceback: TracebackType | None,
     ) -> bool | None:
+        """
+        The freeze_time class is a context manager, so it can be used with the `with` statement.
+        The `__exit__` method is automatically called when exiting the context manager.
+        """
         previous_token = self._token_stack.pop(-1)
         _FILTER_NOW_CONTEXT_STACK.reset(previous_token)
         return None
