@@ -1,12 +1,27 @@
-from collections import OrderedDict
-from typing import Any, Dict, List, Optional, Tuple, Type, Union
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Any, Sequence
 
 import cloudpickle
 
-from chalk.features import Feature
 from chalk.utils import notebook
 from chalk.utils.log_with_context import get_logger
 
+if TYPE_CHECKING:
+    import IPython
+    import IPython.core.interactiveshell
+    import pydantic.fields as pydantic_fields
+
+    from chalk.client.models import UpdateGraphEntityResponse
+    from chalk.features.feature_set import Features
+    from chalk.features.resolver import OfflineResolver, OnlineResolver, Resolver
+
+else:
+    try:
+        import pydantic.v1.fields as pydantic_fields
+    except ImportError:
+        import pydantic.fields as pydantic_fields
+
 _logger = get_logger(__name__)
 
 NO_CLIENT_HELPTEXT = """A Chalk client has not yet been initialized in this notebook.
@@ -37,19 +52,19 @@ Then, in Python you can point a Chalk client at an existing branch with the foll
 """
 
 
-def _upload_object(obj: Any) -> "UpdatedGraphEntityResponse":
+def _upload_object(obj: Any) -> UpdateGraphEntityResponse:
     pickled_obj = cloudpickle.dumps(obj)
 
     from chalk.client.client_impl import ChalkAPIClientImpl
 
-    client = ChalkAPIClientImpl._latest_client
+    client = ChalkAPIClientImpl.latest_client
     if client is None:
         raise RuntimeError(NO_CLIENT_HELPTEXT)
-    if client._config.branch is None:
+    if client.config.branch is None:
         raise RuntimeError(NO_BRANCH_HELPTEXT)
     try:
-        resp = client._send_updated_entity(environment=None, pickled_entity=pickled_obj)
-    except Exception as err:
+        resp = client.send_updated_entity(environment=None, pickled_entity=pickled_obj)
+    except:
         _logger.error(
             f"Failed to upload features/resolvers to branch server on branch '{client.get_branch()}'.", exc_info=True
         )
@@ -57,7 +72,7 @@ def _upload_object(obj: Any) -> "UpdatedGraphEntityResponse":
     return resp
 
 
-def _print_responses(responses: List["UpdateGraphEntityResponse"]):
+def _print_responses(responses: list[UpdateGraphEntityResponse]):
     from chalk.client.models import SingleEntityUpdate
 
     all_errors = [e for r in responses for e in (r.errors or [])]
@@ -66,7 +81,7 @@ def _print_responses(responses: List["UpdateGraphEntityResponse"]):
             _logger.error(e.message)
         return
 
-    all_updated_objects: List[Tuple[str, SingleEntityUpdate]] = []
+    all_updated_objects: list[tuple[str, SingleEntityUpdate]] = []
     for resp in responses:
         all_updated_objects.extend(("+", o) for o in (resp.added or []))
         all_updated_objects.extend(("*", o) for o in (resp.modified or []))
@@ -78,47 +93,23 @@ def _print_responses(responses: List["UpdateGraphEntityResponse"]):
         print(f"{update_char}\t{update_resp.entity_type}: {update_resp.entity_fqn}")
 
 
-def _validate_current_graph() -> List["UpdateGraphError"]:
-    # TODO figure out circular imports
-    from chalk.features import unwrap_feature
-    from chalk.features.feature_set import FeatureSetBase
-    from chalk.features.resolver import Resolver
-    from chalk.parsed._graph_validation import validate_graph
-    from chalk.parsed.duplicate_input_gql import UpsertGraphGQL
-    from chalk.parsed.json_conversions import convert_type_to_gql
-
-    resolvers = [convert_type_to_gql(r) for r in Resolver.registry]
-    features = []
-    for fs in FeatureSetBase.registry.values():
-        all_features = [f for f in fs.features if not f.is_autogenerated]
-        deduped_features = list({unwrap_feature(f).fqn: f for f in all_features}.values())
-        features.extend(convert_type_to_gql(f) for f in deduped_features)
-    graph_gql = UpsertGraphGQL(
-        resolvers=resolvers,
-        features=features,
-        # TODO
-        config=None,  # load_project_config()
-    )
-    return validate_graph(graph_gql)
-
-
 def _remove_object_from_registries(obj: Any):
     """
     If an object fails validation and we keep it registered, all future resolvers/features will cause validation to fail as well.
     """
     from chalk.features.feature_set import FeatureSetBase, is_features_cls
-    from chalk.features.resolver import Resolver
+    from chalk.features.resolver import RESOLVER_REGISTRY, Resolver
 
     if isinstance(obj, Resolver):
-        obj.registry.remove(obj)
+        RESOLVER_REGISTRY.remove_resolver(obj.fqn)
     elif is_features_cls(obj):
         FeatureSetBase.registry.pop(obj.namespace)
     else:
         raise ValueError(f"Can't remove object from registry, unrecognized type: {obj}")
 
 
-def _add_object_to_cache(obj):
-    if notebook.is_defined_in_module(obj):
+def _add_object_to_cache(obj: OnlineResolver | OfflineResolver | type[Features]):
+    if notebook.is_defined_in_module(obj) and not notebook.is_defined_in_cell_magic(obj):
         # If resolver is defined in a module that's imported by a notebook, don't deploy it.
         # This is to avoid re-deploying every feature if customer imports their existing codebase into a notebook.
         return
@@ -134,26 +125,27 @@ def _add_object_to_cache(obj):
 
 
 # features.namespace => features
-_UPDATED_FEATURES_CACHE: Dict[str, "Features"] = OrderedDict()
+_UPDATED_FEATURES_CACHE: dict[str, type["Features"]] = {}
 # resolver.fqn => resolver
-_UPDATED_RESOLVERS_CACHE: Dict[str, Union["OnlineResolver", "OfflineResolver"]] = OrderedDict()
+_UPDATED_RESOLVERS_CACHE: dict[str, OnlineResolver | OfflineResolver] = {}
 
 
-def _clear_entity_update_cache(*_, **__):
+def _clear_entity_update_cache(*args: Any, **kwargs: Any):
     """
     :param _, __: IPython runtime might pass in some objects; ignored
     """
+    del args, kwargs
     _UPDATED_RESOLVERS_CACHE.clear()
     _UPDATED_FEATURES_CACHE.clear()
 
 
-def _deploy_objects(entities: List[Union["Features", "Resolver"]]):
-    try:
-        from chalk.features.feature_set import FeatureSetBase, is_features_cls
-        from chalk.features.resolver import Resolver
+def _deploy_objects(entities: Sequence[type[Features] | Resolver]):
+    from chalk.features.feature_set import is_features_cls
+    from chalk.features.resolver import Resolver
 
+    try:
         # Upload
-        resps = []
+        resps: list[UpdateGraphEntityResponse] = []
         num_resolvers = len([e for e in entities if isinstance(e, Resolver)])
         num_feature_classes = len([e for e in entities if is_features_cls(e)])
         log_messages = filter(
@@ -174,13 +166,13 @@ def _deploy_objects(entities: List[Union["Features", "Resolver"]]):
         raise
 
 
-def _pluralize(count: int, singular_word: str, plural_word: str) -> Optional[str]:
+def _pluralize(count: int, singular_word: str, plural_word: str) -> str | None:
     if count == 0:
         return None
     return f"{count} {singular_word if count == 1 else plural_word}"
 
 
-def _deploy_objects_from_cache(result: "IPython.ExecutionResult"):
+def _deploy_objects_from_cache(result: IPython.core.interactiveshell.ExecutionResult):
     """
     Runs after a Jupyter cell has finished execution.
     Validates and deploys any new entities defined in the cell to the branch server.
@@ -188,17 +180,18 @@ def _deploy_objects_from_cache(result: "IPython.ExecutionResult"):
     This works and should still avoid the forward reference/validation issues in the client.
     :param result: IPython result object.
     """
+    from chalk.client.client_impl import ChalkAPIClientImpl
+
     if result.error_in_exec is not None:
         return
     if len(_UPDATED_FEATURES_CACHE) == 0 and len(_UPDATED_RESOLVERS_CACHE) == 0:
         return
-    from chalk.client.client_impl import ChalkAPIClientImpl
 
-    if ChalkAPIClientImpl._latest_client and ChalkAPIClientImpl._latest_client.get_branch() is None:
-        from IPython.core.display import display_markdown, display_pretty
+    if ChalkAPIClientImpl.latest_client and ChalkAPIClientImpl.latest_client.get_branch() is None:
+        from IPython.core.display import display_markdown
 
         display_markdown(
-            f"Chalk client `<{ChalkAPIClientImpl._latest_client._config.name}>` has no branch set, not updating resolvers/features to server.",
+            f"Chalk client `<{ChalkAPIClientImpl.latest_client.config.name}>` has no branch set, not updating resolvers/features to server.",
             raw=True,
         )
         return
@@ -208,7 +201,7 @@ def _deploy_objects_from_cache(result: "IPython.ExecutionResult"):
 
 def register_cell_hooks():
     # noinspection PyUnresolvedReferences
-    ip = get_ipython()
+    ip = get_ipython()  # type: ignore -- this will be defined if in an interactive environment
     ip.events.register(notebook.IPythonEvents.PRE_RUN_CELL.value, _clear_entity_update_cache)
     ip.events.register(notebook.IPythonEvents.POST_RUN_CELL.value, _deploy_objects_from_cache)
 
@@ -225,12 +218,8 @@ def register_pydantic_serializer():
 
     Borrowed from: https://github.com/ray-project/ray/blob/master/python/ray/util/serialization_addons.py
     """
-    try:
-        import pydantic.fields
-    except ImportError:
-        return
 
-    def _custom_serializer(o):
+    def _custom_serializer(o: pydantic_fields.ModelField):
         return {
             "name": o.name,
             # outer_type_ is the original type for ModelFields,
@@ -246,24 +235,25 @@ def register_pydantic_serializer():
             "field_info": o.field_info,
         }
 
-    def _custom_deserializer(kwargs):
-        return (pydantic.fields.ModelField(**kwargs),)
+    def _custom_deserializer(kwargs: dict[str, Any]):
+        return (pydantic_fields.ModelField(**kwargs),)
 
-    def _cloud_pickler_reducer(obj):
+    def _cloud_pickler_reducer(obj: Any):
         return _custom_deserializer, (_custom_serializer(obj),)
 
-    import cloudpickle
-
-    cloudpickle.CloudPickler.dispatch[pydantic.fields.ModelField] = _cloud_pickler_reducer
+    # See https://github.com/ray-project/ray/blob/7fe451a8b83ea789d17fef3b8a954164008bf8a8/python/ray/_private/serialization.py#L168
+    cloudpickle.CloudPickler.dispatch[pydantic_fields.ModelField] = _cloud_pickler_reducer  # type: ignore
 
 
-def register_live_updates_if_in_notebook(cls: Type):
+def register_live_updates_if_in_notebook(cls: Any):
     """
     This is called manually by the modules containing FeatureSetBase and Resolver due to circular imports
     :param cls:
     :return:
     """
     if notebook.is_notebook():
+        if not hasattr(cls, "hook"):
+            raise TypeError(f"{cls} has no hook attribute")
         setattr(cls, "hook", _add_object_to_cache)
 
 
