@@ -1,6 +1,7 @@
 from dataclasses import dataclass
+from datetime import datetime
 from enum import Enum
-from typing import Any, Dict, List, Optional, Sequence, Union
+from typing import Any, List, Optional
 
 import dataclasses_json
 
@@ -166,7 +167,7 @@ class UpsertStreamResolverGQL:
     machineType: Optional[str] = None
     environment: Optional[List[str]] = None
     output: Optional[List[UpsertFeatureIdGQL]] = None
-    inputs: Optional[Sequence[UpsertStreamResolverParamGQL]] = None
+    inputs: Optional[List[UpsertStreamResolverParamGQL]] = None
     doc: Optional[str] = None
     owner: Optional[str] = None
     filename: Optional[str] = None
@@ -206,6 +207,7 @@ class UpsertResolverGQL:
     timeout: Optional[str] = None
     filename: Optional[str] = None
     sourceLine: Optional[int] = None
+    dataSources: Optional[List[Any]] = None
 
 
 @dataclasses_json.dataclass_json
@@ -227,6 +229,30 @@ class UpsertSinkResolverGQL:
     sourceLine: Optional[int] = None
 
 
+@dataclasses_json.dataclass_json
+@dataclass
+class RecomputeFeaturesGQL:
+    featureFqns: Optional[List[str]]
+    all: Optional[bool]
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class UpsertCronQueryGQL:
+    name: str
+    cron: str
+    filename: str
+    output: List[str]  # exploded engine-side
+    maxSamples: Optional[int]
+    recomputeFeatures: RecomputeFeaturesGQL
+    lowerBound: Optional[datetime]
+    upperBound: Optional[datetime]
+    tags: Optional[List[str]]
+    requiredResolverTags: Optional[List[str]]
+    storeOnline: Optional[bool] = True  # None = True
+    storeOffline: Optional[bool] = True  # None = True
+
+
 @dataclasses_json.dataclass_json
 @dataclass
 class MetadataSettings:
@@ -240,10 +266,17 @@ class FeatureSettings:
     metadata: Optional[List[MetadataSettings]] = None
 
 
+@dataclasses_json.dataclass_json
+@dataclass
+class ResolverSettings:
+    metadata: Optional[List[MetadataSettings]] = None
+
+
 @dataclasses_json.dataclass_json
 @dataclass
 class ValidationSettings:
     feature: Optional[FeatureSettings] = None
+    resolver: Optional[ResolverSettings] = None
 
 
 @dataclasses_json.dataclass_json
@@ -276,6 +309,7 @@ class FailedImport:
 @dataclass
 class ChalkPYInfo:
     version: str
+    python: Optional[str] = None
 
 
 class MetricKindGQL(str, Enum):
@@ -451,6 +485,7 @@ class CreateAlertTriggerGQL:
     thresholdValue: float
     seriesName: Optional[str] = None
     channelName: Optional[str] = None
+    description: Optional[str] = None
 
 
 @dataclasses_json.dataclass_json
@@ -510,6 +545,129 @@ class FeatureClassGQL:
     tags: List[str]
 
 
+@dataclasses_json.dataclass_json
+@dataclass
+class PositionGQL:
+    """Mirrors an LSP Position
+
+    https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#position
+    """
+
+    line: int
+    """Line position in a document (one-based)."""
+
+    character: int
+    """Character offset on a line in a document (zero-based)."""
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class RangeGQL:
+    """Mirrors an LSP Range
+
+    https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#range
+    """
+
+    start: PositionGQL
+    end: PositionGQL
+
+
+class DiagnosticSeverityGQL(str, Enum):
+    """Mirrors an LSP DiagnosticSeverity"""
+
+    Error = "Error"
+    Warning = "Warning"
+    Information = "Information"
+    Hint = "Hint"
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class CodeDescriptionGQL:
+    href: str
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class LocationGQL:
+    uri: str
+    range: RangeGQL
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class DiagnosticRelatedInformationGQL:
+    location: LocationGQL
+    """The location of this related diagnostic information."""
+
+    message: str
+    """The message of this related diagnostic information."""
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class DiagnosticGQL:
+    range: RangeGQL
+    message: str
+    severity: Optional[DiagnosticSeverityGQL]
+    code: Optional[str]
+    codeDescription: Optional[CodeDescriptionGQL]
+    relatedInformation: Optional[List[DiagnosticRelatedInformationGQL]] = None
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class PublishDiagnosticsParams:
+    """Mirrors an LSP PublishDiagnosticsParams
+
+    https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#publishDiagnosticsParams
+    """
+
+    uri: str
+    diagnostics: List[DiagnosticGQL]
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class TextDocumentIdentifierGQL:
+    uri: str
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class TextEditGQL:
+    range: RangeGQL
+    newText: str
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class TextDocumentEditGQL:
+    textDocument: TextDocumentIdentifierGQL
+    edits: List[TextEditGQL]
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class WorkspaceEditGQL:
+    documentChanges: List[TextDocumentEditGQL]
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class CodeActionGQL:
+    title: str
+    diagnostics: Optional[List[DiagnosticGQL]]
+    edit: WorkspaceEditGQL
+
+
+@dataclasses_json.dataclass_json
+@dataclass
+class LspGQL:
+    diagnostics: List[PublishDiagnosticsParams]
+    actions: List[CodeActionGQL]
+
+
 @dataclasses_json.dataclass_json
 @dataclass
 class UpsertGraphGQL:
@@ -517,6 +675,7 @@ class UpsertGraphGQL:
     features: Optional[List[UpsertFeatureGQL]] = None
     streams: Optional[List[UpsertStreamResolverGQL]] = None
     sinks: Optional[List[UpsertSinkResolverGQL]] = None
+    cronQueries: Optional[List[UpsertCronQueryGQL]] = None
     charts: Optional[List[CreateChartGQL]] = None
     config: Optional[ProjectSettingsGQL] = None
     failed: Optional[List[FailedImport]] = None
@@ -526,3 +685,4 @@ class UpsertGraphGQL:
     cdcSources: Optional[List[UpsertCDCSourceGQL]] = None
     sqlSources: Optional[List[UpsertSQLSourceGQL]] = None
     featureClasses: Optional[List[FeatureClassGQL]] = None
+    lsp: Optional[LspGQL] = None
