@@ -1,19 +1,24 @@
+from __future__ import annotations
+
 import collections
 import dataclasses
 import json
-from typing import Any, List, Optional, Type, TypeVar, Union, overload
+from datetime import timedelta
+from typing import TYPE_CHECKING, Any, List, Optional, Type, TypeVar, Union, overload
 
-import pydantic
-from pydantic.main import BaseModel
-
-from chalk.features import DataFrame, Feature, FeatureWrapper, Filter, serialize_dtype, unwrap_feature
+from chalk import Resolver, Windowed
+from chalk.features import DataFrame, Feature, Features, Filter, serialize_dtype
+from chalk.features._encoding.outputs import find_feature_or_resolver_by_fqn
 from chalk.features.feature_field import HasOnePathObj
 from chalk.features.feature_set import is_feature_set_class
 from chalk.features.pseudofeatures import PSEUDONAMESPACE
 from chalk.features.resolver import Cron, OfflineResolver, OnlineResolver, SinkResolver, StreamResolver
+from chalk.features.scheduled_query import ScheduledQuery
 from chalk.parsed.duplicate_input_gql import (
     FeatureClassGQL,
+    RecomputeFeaturesGQL,
     UpsertCDCSourceGQL,
+    UpsertCronQueryGQL,
     UpsertDataFrameGQL,
     UpsertFeatureGQL,
     UpsertFeatureIdGQL,
@@ -35,7 +40,6 @@ from chalk.parsed.duplicate_input_gql import (
     VersionInfoGQL,
 )
 from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
-from chalk.streams import Windowed
 from chalk.streams.types import (
     StreamResolverParam,
     StreamResolverParamKeyedState,
@@ -43,16 +47,21 @@ from chalk.streams.types import (
     StreamResolverParamMessageWindow,
 )
 from chalk.utils import paths
-from chalk.utils.duration import timedelta_to_duration
-from chalk.utils.string import to_snake_case
+from chalk.utils.collections import get_unique_item
+from chalk.utils.duration import CronTab, Duration, timedelta_to_duration
+from chalk.utils.pydanticutil.pydantic_compat import is_pydantic_basemodel, is_pydantic_basemodel_instance
+from chalk.utils.string import removeprefix, to_snake_case
+
+if TYPE_CHECKING:
+    import pydantic
+else:
+    try:
+        import pydantic.v1 as pydantic
+    except ImportError:
+        import pydantic
 
 T = TypeVar("T")
 
-try:
-    import attrs
-except ImportError:
-    attrs = None
-
 
 @dataclasses.dataclass
 class _ConvertedType:
@@ -65,39 +74,32 @@ def _get_qualified_class_name(cls: Type[Any]):
     return cls.__qualname__ if mod == "builtins" else f"{mod}.{cls.__qualname__}"
 
 
-_primitive_types = {bool, str, int, float}
-
-
-def _convert_type(underlying_type: Type) -> _ConvertedType:
+def _convert_type(underlying_type: Type, is_windowed: bool) -> _ConvertedType:
     if isinstance(underlying_type, Windowed):
-        if underlying_type._kind in _primitive_types:
-            return _ConvertedType(name=f"Windowed[{str(underlying_type._kind.__name__)}]", bases=["CHALK_WINDOWED"])
-        else:
-            kind = str(underlying_type._kind).replace("typing.", "")
-            return _ConvertedType(name=f"Windowed[{kind}]", bases=["CHALK_WINDOWED"])
-    if isinstance(underlying_type, FeatureWrapper):
-        return _convert_type(unwrap_feature(underlying_type).typ.underlying)
-    assert isinstance(underlying_type, type), f"Underlying type is not a type: '{str(underlying_type)}'"
-    base_classes = [x.__name__ for x in type.mro(underlying_type)]
-    if "Enum" in base_classes:
-        # add type(s) of Enum to base_classes
-        # Enums are already confirmed to be of one type by here
-        value_type = [type(t).__name__ for t in underlying_type._value2member_map_.keys()][0]
-        base_classes.insert(base_classes.index("Enum") + 1, value_type)
-
-    # Attrs and Dataclasses don't technically have base classes
-    # Pydantic calls their base class BaseModel which is way too generic for string comparison
-    # For simplicity on the server-side validation, we'll come up with our own "base class" names
-    if dataclasses.is_dataclass(underlying_type):
-        base_classes.append("__dataclass__")
-    if attrs is not None and isinstance(underlying_type, type) and attrs.has(underlying_type):
-        base_classes.append("__attrs__")
-    if issubclass(underlying_type, pydantic.BaseModel):
-        base_classes.append("__pydantic__")
+        is_windowed = True
+        underlying_type = underlying_type.kind
+    bases = []
+    pretty_name = repr(underlying_type)
+    if isinstance(underlying_type, type):
+        bases.extend(x.__name__ for x in underlying_type.mro())
+        pretty_name = underlying_type.__name__
+    else:
+        bases = ["object"]
+    pretty_name = pretty_name.replace("typing.", "")
+    if is_windowed:
+        bases.append("CHALK_WINDOWED")
+        return _ConvertedType(name=f"Windowed[{pretty_name}]", bases=bases)
+    if isinstance(underlying_type, Feature):
+        return _convert_type(underlying_type.converter.rich_type, underlying_type.is_windowed)
 
+    bases = []
+    if isinstance(underlying_type, type):
+        bases.extend(x.__name__ for x in underlying_type.mro())
+    else:
+        bases = ["object"]
     return _ConvertedType(
-        name=underlying_type.__name__,
-        bases=base_classes,
+        name=pretty_name,
+        bases=bases,
     )
 
 
@@ -144,6 +146,25 @@ def _get_path_component(pc: HasOnePathObj) -> UpsertReferencePathComponentGQL:
     )
 
 
+def _parse_generic_params_recursively(typ: Any) -> str:
+    """
+    Stringify a type with generic parameters like 'list[list[str]]',
+    making sure parameters are included recursively.
+    :return: the type name along with any parameters it may have
+    """
+    if not isinstance(typ, type):
+        type_name = repr(typ)
+    else:
+        type_name = typ.__name__
+
+    type_name = removeprefix(type_name, "typing.")
+    if not hasattr(typ, "__args__"):
+        return type_name
+    params = [_parse_generic_params_recursively(x) for x in typ.__args__]
+    param_string = ",".join(params)
+    return f"{type_name}[{param_string}]"
+
+
 @overload
 def convert_type_to_gql(t: Filter, path_prefix: Optional[str] = None) -> UpsertFilterGQL:
     ...
@@ -166,16 +187,33 @@ def convert_type_to_gql(
     ...
 
 
+@overload
+def convert_type_to_gql(
+    t: Resolver, path_prefix: Optional[str] = None
+) -> UpsertStreamResolverGQL | UpsertSinkResolverGQL | UpsertResolverGQL:
+    ...
+
+
 @overload
 def convert_type_to_gql(t: StreamResolverParam, path_prefix: Optional[str] = None) -> UpsertStreamResolverParamGQL:
     ...
 
 
+@overload
+def convert_type_to_gql(t: ScheduledQuery, path_prefix: Optional[str] = None) -> UpsertCronQueryGQL:
+    ...
+
+
 @overload
 def convert_type_to_gql(t: Feature, path_prefix: Optional[str] = None) -> UpsertFeatureGQL:
     ...
 
 
+@overload
+def convert_type_to_gql(t: Type[Features], path_prefix: Optional[str] = None) -> FeatureClassGQL:
+    ...
+
+
 def convert_type_to_gql(
     t: Any,
     # Resolvers have their filenames attached to them. But for UI presentation, we don't
@@ -189,7 +227,7 @@ def convert_type_to_gql(
             fqn=t.fqn,
             kind="stream",
             sourceClassName=_get_qualified_class_name(t.source.__class__),
-            sourceConfig=t.source._config_to_json(),
+            sourceConfig=t.source.config_to_json(),
             functionDefinition=t.function_definition,
             environment=[t.environment] if isinstance(t.environment, str) else t.environment,
             doc=t.fn.__doc__,
@@ -219,7 +257,7 @@ def convert_type_to_gql(
                     path=[_get_path_component(p) for p in f.path or []],
                 )
                 for f in t.inputs
-                if not f.is_autogenerated
+                if isinstance(f, Feature) and not f.is_autogenerated
             ],
             environment=t.environment,
             tags=t.tags,
@@ -260,6 +298,10 @@ def convert_type_to_gql(
                 all_inputs.append(UpsertResolverInputUnionGQL(dataframe=_convert_df(f)))
 
         assert t.output is not None
+        if t.data_sources is not None:
+            data_sources = [source.to_json() for source in t.data_sources]
+        else:
+            data_sources = None
         return UpsertResolverGQL(
             fqn=t.fqn,
             kind="offline" if isinstance(t, OfflineResolver) else "online",
@@ -279,19 +321,20 @@ def convert_type_to_gql(
             environment=t.environment,
             tags=t.tags,
             doc=t.doc,
-            cron=cron,
+            cron=convert_type_crontab(cron) if cron is not None else cron,
             machineType=t.machine_type,
             owner=t.owner,
             timeout=None if t.timeout is None else timedelta_to_duration(t.timeout),
             filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
             sourceLine=t.source_line,
+            dataSources=data_sources,
         )
 
     if isinstance(t, StreamResolverParam):
         if isinstance(t, StreamResolverParamMessage):
-            converted_type = _convert_type(t.typ)
+            converted_type = _convert_type(t.typ, is_windowed=False)
             schema = None
-            if issubclass(t.typ, BaseModel):
+            if is_pydantic_basemodel(t.typ):
                 schema = json.loads(t.typ.schema_json())
             return UpsertStreamResolverParamGQL(
                 message=UpsertStreamResolverParamMessageGQL(
@@ -303,14 +346,14 @@ def convert_type_to_gql(
                 state=None,
             )
         elif isinstance(t, StreamResolverParamKeyedState):
-            converted_type = _convert_type(t.typ)
+            converted_type = _convert_type(t.typ, is_windowed=False)
             default_value = None
             schema = None
             if dataclasses.is_dataclass(t.default_value):
                 default_value = json.loads(json.dumps(dataclasses.asdict(t.default_value)))
                 schema = pydantic.dataclasses.dataclass(t.typ).__pydantic_model__.schema()
-            elif isinstance(t.default_value, BaseModel):
-                assert issubclass(t.typ, BaseModel)
+            elif is_pydantic_basemodel_instance(t.default_value):
+                assert is_pydantic_basemodel(t.typ)
                 default_value = json.loads(t.default_value.json())
                 schema = json.loads(t.typ.schema_json())
             return UpsertStreamResolverParamGQL(
@@ -324,6 +367,32 @@ def convert_type_to_gql(
                 message=None,
             )
 
+    if isinstance(t, ScheduledQuery):
+        for f in t.output:
+            validate_feature_reference(f)
+        get_unique_item((f.split(".")[0] for f in t.output), " of root output feature namespaces")
+        if not isinstance(t.recompute_features, bool):
+            for f in t.recompute_features:
+                validate_feature_reference(f)
+
+        return UpsertCronQueryGQL(
+            name=t.name,
+            cron=convert_type_crontab(t.cron),
+            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
+            output=list(t.output),
+            maxSamples=t.max_samples,
+            recomputeFeatures=RecomputeFeaturesGQL(
+                featureFqns=list(t.recompute_features) if not isinstance(t.recompute_features, bool) else None,
+                all=t.recompute_features if isinstance(t.recompute_features, bool) else None,
+            ),
+            lowerBound=t.lower_bound,
+            upperBound=t.upper_bound,
+            tags=list(t.tags) if t.tags is not None else None,
+            requiredResolverTags=list(t.required_resolver_tags) if t.required_resolver_tags is not None else None,
+            storeOnline=t.store_online,
+            storeOffline=t.store_offline,
+        )
+
     if isinstance(t, Feature):
         assert t.name is not None, "Feature has no name"
         assert t.namespace is not None, "Feature has no namespace"
@@ -334,29 +403,87 @@ def convert_type_to_gql(
         feature_time_kind_gql = None
         if t.is_autogenerated:
             raise RuntimeError("Autogenerated features should not be converted")
+
         if t.is_has_one:
+            if t.join is None:
+                builder = t.features_cls.__chalk_error_builder__
+                builder.add_diagnostic(
+                    range=builder.annotation_range(t.attribute_name),
+                    message=(
+                        f"The attribute '{t.attribute_name}' on the feature class '{t.features_cls.__name__}' "
+                        f"references {t.typ}. However, the has-one relationship is not defined."
+                    ),
+                    raise_error=TypeError,
+                    code_href="https://docs.chalk.ai/docs/has-one",
+                    label="Missing has-one join",
+                    code="35",
+                )
             assert t.join is not None
             has_one_kind_gql = UpsertHasOneKindGQL(join=convert_type_to_gql(t.join))
+
         elif t.is_has_many:
-            assert t.join is not None
+            if t.join is None:
+                builder = t.features_cls.__chalk_error_builder__
+                builder.add_diagnostic(
+                    range=builder.annotation_range(t.attribute_name),
+                    message=(
+                        f"The attribute '{t.attribute_name}' on the feature class '{t.features_cls.__name__}' "
+                        f"references a DataFrame. However, the has-many relationship is not defined."
+                    ),
+                    label="Missing has-many join",
+                    raise_error=TypeError,
+                    code="34",
+                    code_href="https://docs.chalk.ai/docs/has-many",
+                )
             has_many_kind_gql = UpsertHasManyKindGQL(
                 join=convert_type_to_gql(t.join),
                 columns=None,
                 filters=None,
             )
+
+        elif t.join is not None:
+            # If a has_one/has_many has an incorrect type annotation
+            builder = t.features_cls.__chalk_error_builder__
+            builder.add_diagnostic(
+                range=builder.annotation_range(t.attribute_name),
+                message=(
+                    f"The attribute '{t.attribute_name}' on the feature class '{t.features_cls.__name__}' "
+                    f"has a join filter ({t.join}) but its type annotation is not a feature class or "
+                    f"DataFrame ({t.typ})."
+                ),
+                label="Incorrect join type annotation",
+                raise_error=TypeError,
+                code="34",
+                code_href="https://docs.chalk.ai/docs/has-many",
+            )
+
         elif t.is_feature_time:
             feature_time_kind_gql = UpsertFeatureTimeKindGQL()
+
         else:
             assert t.typ is not None
-            parsed_type = t.typ.parsed_annotation
-            if hasattr(parsed_type, "__chalk_feature_set__") and parsed_type.__chalk_feature_set__:
-                raise TypeError(
-                    f"Feature '{t.fqn}' is annotated with feature set '{parsed_type}' but no join was found. "
-                    f"Please ensure that join between keys is specified between '{t.namespace}' and '{parsed_type.namespace}'."
-                )
-            converted_type = _convert_type(t.typ.underlying)
+            try:
+                converted_type = _convert_type(t.converter.rich_type, is_windowed=t.is_windowed)
+            except Exception as e:
+                builder = t.features_cls.__chalk_error_builder__
+                if builder:
+                    builder.add_diagnostic(
+                        message=e.args[0] if len(e.args) > 0 else str(e),
+                        range=builder.annotation_range(t.attribute_name) or builder.property_range(t.attribute_name),
+                        code="39",
+                        label="invalid type",
+                        code_href="https://docs.chalk.ai/docs/feature-types",
+                    )
+                raise e
+            pretty_name = repr(t.converter.rich_type)
+            if isinstance(t.converter.rich_type, type):
+                pretty_name = _parse_generic_params_recursively(t.converter.rich_type)
+            pretty_name = pretty_name.replace("typing.", "")
+            if t.is_windowed:
+                pretty_name = f"Windowed[{pretty_name}]"
+
             scalar_kind_gql = UpsertScalarKindGQL(
-                scalarKind=converted_type.name,
+                scalarKind=pretty_name,
                 primary=t.primary,
                 baseClasses=converted_type.bases,
                 version=t.version and t.version.version,
@@ -369,6 +496,7 @@ def convert_type_to_gql(
                 ),
                 dtype=serialize_dtype(t.converter.pyarrow_dtype),
             )
+
         return UpsertFeatureGQL(
             id=UpsertFeatureIdGQL(
                 fqn=t.fqn,
@@ -384,12 +512,12 @@ def convert_type_to_gql(
             owner=t.owner,
             windowBuckets=None if t.window_buckets is None else sorted(list(t.window_buckets)),
             etlOfflineToOnline=t.etl_offline_to_online,
-            tags=t.tags,
+            tags=None if t.tags is None else list(t.tags),
             hasOneKind=has_one_kind_gql,
             hasManyKind=has_many_kind_gql,
             scalarKind=scalar_kind_gql,
             featureTimeKind=feature_time_kind_gql,
-            namespacePath=str(paths.get_classpath_or_name(t.features_cls)),
+            namespacePath=paths.get_classpath_or_name(t.features_cls),
             isSingleton=t.is_singleton,
         )
 
@@ -422,6 +550,19 @@ def convert_type_to_gql(
     raise ValueError(f"Unable to convert {t} to GQL")
 
 
+def convert_type_crontab(cron: Union[CronTab, Duration]) -> str:
+    return f"{cron.total_seconds()}s" if isinstance(cron, timedelta) else cron
+
+
+def validate_feature_reference(f: str):
+    path = f.split(".")
+    fqn = ".".join(path[: min(2, len(path))])
+    if (feature := find_feature_or_resolver_by_fqn(fqn)) is None:
+        raise ValueError(f"Feature or resolver '{f}' not found.")
+    if isinstance(feature, Resolver):
+        raise ValueError(f"'{f}' is a resolver, not a feature.")
+
+
 def gather_cdc_sources() -> List[UpsertCDCSourceGQL]:
     integration_name_to_tables = collections.defaultdict(list)
     for source in BaseSQLSource.registry:
