@@ -4,6 +4,7 @@ from typing import Any, Dict, Optional, Union, overload
 from chalk.sql._internal.incremental import IncrementalSettings
 from chalk.sql._internal.integrations.bigquery import BigQuerySourceImpl
 from chalk.sql._internal.integrations.cloudsql import CloudSQLSourceImpl
+from chalk.sql._internal.integrations.databricks import DatabricksSourceImpl
 from chalk.sql._internal.integrations.mysql import MySQLSourceImpl
 from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
 from chalk.sql._internal.integrations.redshift import RedshiftSourceImpl
@@ -182,6 +183,7 @@ def PostgreSQLSource(
     *,
     name: str,
     engine_args: Optional[Dict[str, Any]] = ...,
+    async_engine_args: Optional[Dict[str, Any]] = ...,
 ) -> SQLSourceWithTableIngestProtocol:
     """If you have only one PostgreSQL integration, there's no need to provide
     a distinguishing name.
@@ -198,6 +200,8 @@ def PostgreSQLSource(
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine. These arguments will be
         merged with any default arguments from the named integration.
+    async_engine_args
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
@@ -220,6 +224,7 @@ def PostgreSQLSource(
     user: str = ...,
     password: str = ...,
     engine_args: Optional[Dict[str, Any]] = ...,
+    async_engine_args: Optional[Dict[str, Any]] = ...,
 ) -> SQLSourceWithTableIngestProtocol:
     """You can also configure the integration directly using environment
     variables on your local machine or from those added through the
@@ -239,6 +244,8 @@ def PostgreSQLSource(
         The password to be used if the server demands password authentication.
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine.
+    async_engine_args
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
@@ -272,6 +279,7 @@ def PostgreSQLSource(
     password: Optional[str] = None,
     name: Optional[str] = None,
     engine_args: Optional[Dict[str, Any]] = None,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> TableIngestProtocol:
     """Create a PostgreSQL data source. SQL-based data sources
     created without arguments assume a configuration in your
@@ -282,7 +290,9 @@ def PostgreSQLSource(
     the data source. See the overloaded signatures for more
     details.
     """
-    return PostgreSQLSourceImpl(host, port, db, user, password, name, engine_args=engine_args)
+    return PostgreSQLSourceImpl(
+        host, port, db, user, password, name, engine_args=engine_args, async_engine_args=async_engine_args
+    )
 
 
 @overload
@@ -308,6 +318,7 @@ def MySQLSource(
     *,
     name: str,
     engine_args: Optional[Dict[str, Any]] = ...,
+    async_engine_args: Optional[Dict[str, Any]] = ...,
 ) -> SQLSourceWithTableIngestProtocol:
     """If you have only one MySQL integration, there's no need to provide
     a distinguishing name.
@@ -340,12 +351,13 @@ def MySQLSource(
 @overload
 def MySQLSource(
     *,
-    host: str = ...,
+    host: str,
     port: Union[int, str] = ...,
     db: str = ...,
     user: str = ...,
     password: str = ...,
     engine_args: Optional[Dict[str, Any]] = ...,
+    async_engine_args: Optional[Dict[str, Any]] = ...,
 ) -> SQLSourceWithTableIngestProtocol:
     """
     You can also configure the integration directly using environment
@@ -366,6 +378,8 @@ def MySQLSource(
         The password to be used if the server demands password authentication.
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine.
+    async_engine_args:
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
@@ -399,6 +413,7 @@ def MySQLSource(
     password: Optional[str] = None,
     name: Optional[str] = None,
     engine_args: Optional[Dict[str, Any]] = None,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> SQLSourceWithTableIngestProtocol:
     """Create a MySQL data source. SQL-based data sources
     created without arguments assume a configuration in your
@@ -409,11 +424,15 @@ def MySQLSource(
     the data source. See the overloaded signatures for more
     details.
     """
-    return MySQLSourceImpl(host, port, db, user, password, name, engine_args=engine_args)
+    return MySQLSourceImpl(
+        host, port, db, user, password, name, engine_args=engine_args, async_engine_args=async_engine_args
+    )
 
 
 def SQLiteInMemorySource(
-    name: Optional[str] = None, engine_args: Optional[Dict[str, Any]] = None
+    name: Optional[str] = None,
+    engine_args: Optional[Dict[str, Any]] = None,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> SQLSourceWithTableIngestProtocol:
     """Testing SQL source.
 
@@ -426,6 +445,8 @@ def SQLiteInMemorySource(
         The name of the integration.
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine.
+    async_engine_args
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
@@ -436,11 +457,14 @@ def SQLiteInMemorySource(
     --------
     >>> source = SQLiteInMemorySource(name="RISK")
     """
-    return SQLiteSourceImpl(name=name, engine_args=engine_args)
+    return SQLiteSourceImpl(name=name, engine_args=engine_args, async_engine_args=async_engine_args)
 
 
 def SQLiteFileSource(
-    filename: Union[str, PathLike], name: Optional[str] = None, engine_args: Optional[Dict[str, Any]] = None
+    filename: Union[str, PathLike],
+    name: Optional[str] = None,
+    engine_args: Optional[Dict[str, Any]] = None,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> SQLSourceWithTableIngestProtocol:
     """Create a SQLite source for a file.
 
@@ -452,13 +476,15 @@ def SQLiteFileSource(
         The name to use in testing
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine.
+    async_engine_args
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
     SQLSourceWithTableIngestProtocol
         The SQL source for use in Chalk resolvers.
     """
-    return SQLiteSourceImpl(filename=filename, name=name, engine_args=engine_args)
+    return SQLiteSourceImpl(filename=filename, name=name, engine_args=engine_args, async_engine_args=async_engine_args)
 
 
 @overload
@@ -520,6 +546,7 @@ def RedshiftSource(
     db: str = ...,
     user: str = ...,
     password: str = ...,
+    port: str = ...,
     engine_args: Optional[Dict[str, Any]] = ...,
 ) -> BaseSQLSourceProtocol:
     """You can also configure the integration directly using environment
@@ -536,6 +563,8 @@ def RedshiftSource(
         Redshify username to connect as.
     password
         The password for the Redshift database.
+    port
+        Port number to use.
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine.
 
@@ -568,6 +597,7 @@ def RedshiftSource(
     user: Optional[str] = None,
     password: Optional[str] = None,
     name: Optional[str] = None,
+    port: Optional[Union[str, int]] = None,
     engine_args: Optional[Dict[str, Any]] = None,
 ) -> BaseSQLSourceProtocol:
     """Create a Redshift data source. SQL-based data sources
@@ -579,7 +609,7 @@ def RedshiftSource(
     the data source. See the overloaded signatures for more
     details.
     """
-    return RedshiftSourceImpl(host, db, user, password, name, engine_args=engine_args)
+    return RedshiftSourceImpl(host, db, user, password, name, port, engine_args=engine_args)
 
 
 @overload
@@ -734,6 +764,7 @@ def CloudSQLSource(
     *,
     name: str,
     engine_args: Optional[Dict[str, Any]] = ...,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> BaseSQLSourceProtocol:
     """If you have only one CloudSQL integration, there's no need to provide
     a distinguishing name.
@@ -750,6 +781,8 @@ def CloudSQLSource(
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine. These arguments will be
         merged with any default arguments from the named integration.
+    async_engine_args
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
@@ -771,6 +804,7 @@ def CloudSQLSource(
     user: Optional[str] = ...,
     password: Optional[str] = ...,
     engine_args: Optional[Dict[str, Any]] = ...,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> BaseSQLSourceProtocol:
     """You can also configure the integration directly using environment
     variables on your local machine or from those added through the
@@ -788,6 +822,8 @@ def CloudSQLSource(
         The password to use.
     engine_args
         Additional arguments to use when constructing the SQLAlchemy engine.
+    async_engine_args
+        Additional arguments to use when constructing an async SQLAlchemy engine.
 
     Returns
     -------
@@ -814,6 +850,7 @@ def CloudSQLSource(
     user: Optional[str] = None,
     password: Optional[str] = None,
     engine_args: Optional[Dict[str, Any]] = None,
+    async_engine_args: Optional[Dict[str, Any]] = None,
 ) -> BaseSQLSourceProtocol:
     """Create a CloudSQL data source. SQL-based data sources
     created without arguments assume a configuration in your
@@ -831,6 +868,133 @@ def CloudSQLSource(
         user=user,
         password=password,
         engine_args=engine_args,
+        async_engine_args=async_engine_args,
+    )
+
+
+@overload
+def DatabricksSource() -> BaseSQLSourceProtocol:
+    """Connect to the only configured Databricks database.
+
+    If you have only one Databricks connection that you'd like
+    to add to Chalk, you do not need to specify any arguments
+    to construct the source in your code.
+
+    Returns
+    -------
+    BaseSQLSourceProtocol
+        The SQL source for use in Chalk resolvers.
+
+    Examples
+    --------
+    >>> source = DatabricksSource()
+    """
+
+
+@overload
+def DatabricksSource(*, name: str, engine_args: Optional[Dict[str, Any]] = ...) -> BaseSQLSourceProtocol:
+    """Chalk's injects environment variables to support data integrations.
+
+    But what happens when you have two data sources of the same kind?
+    When you create a new data source from your dashboard,
+    you have an option to provide a name for the integration.
+    You can then reference this name in the code directly.
+
+    Parameters
+    ----------
+    name
+        Name of the integration, as configured in your dashboard.
+    engine_args
+        Additional arguments to use when constructing the SQLAlchemy engine. These arguments will be
+        merged with any default arguments from the named integration.
+
+    Returns
+    -------
+    BaseSQLSourceProtocol
+        The SQL source for use in Chalk resolvers.
+
+    Examples
+    --------
+    >>> source = DatabricksSource(name="RISK")
+    """
+    ...
+
+
+@overload
+def DatabricksSource(
+    *,
+    host: str = ...,
+    http_path: str = ...,
+    access_token: str = ...,
+    db: str = ...,
+    port: str = ...,
+    engine_args: Optional[Dict[str, Any]] = ...,
+) -> BaseSQLSourceProtocol:
+    """You can also configure the integration directly using environment
+    variables on your local machine or from those added through the
+    generic environment variable support (https://docs.chalk.ai/docs/env-vars).
+
+    Parameters
+    ----------
+    host
+        Your Databricks host.
+    http_path
+        Databricks HTTP path to use.
+    access_token
+        Access token to connect to Databricks.
+    db
+        Database to use.
+    port
+        Port number to use.
+    engine_args
+        Additional arguments to use when constructing the SQLAlchemy engine.
+
+    Returns
+    -------
+    BaseSQLSourceProtocol
+        The SQL source for use in Chalk resolvers.
+
+    Examples
+    --------
+    >>> import os
+    >>> databricks = DatabricksSource(
+    ...     host=os.getenv("DATABRICKS_HOST"),
+    ...     http_path=os.getenv("DATABRICKS_HTTP_PATH"),
+    ...     access_token=os.getenv("DATABRICKS_TOKEN"),
+    ...     db=os.getenv("DATABRICKS_DATABASE"),
+    ...     port=os.getenv("DATABRICKS_PORT"),
+    ... )
+    """
+    ...
+
+
+def DatabricksSource(
+    *,
+    name: Optional[str] = None,
+    host: Optional[str] = None,
+    http_path: Optional[str] = None,
+    access_token: Optional[str] = None,
+    db: Optional[str] = None,
+    port: Optional[Union[str, int]] = None,
+    engine_args: Optional[Dict[str, Any]] = None,
+) -> BaseSQLSourceProtocol:
+    """Create a Databricks data source. SQL-based data sources
+    created without arguments assume a configuration in your
+    Chalk Dashboard. Those created with the `name=` keyword
+    argument will use the configuration for the integration
+    with the given name. And finally, those created with
+    explicit arguments will use those arguments to configure
+    the data source. See the overloaded signatures for more
+    details.
+    """
+    return DatabricksSourceImpl(
+        host=host,
+        http_path=http_path,
+        access_token=access_token,
+        db=db,
+        port=port,
+        name=name,
+        engine_args=engine_args,
     )
 
 
@@ -838,6 +1002,7 @@ __all__ = [
     "BaseSQLSourceProtocol",
     "BigQuerySource",
     "CloudSQLSource",
+    "DatabricksSource",
     "MySQLSource",
     "PostgreSQLSource",
     "RedshiftSource",
