@@ -11,7 +11,7 @@ if TYPE_CHECKING:
 
 
 class CloudSQLSourceImpl(BaseSQLSource):
-    kind = SQLSourceKind.postgres
+    kind = SQLSourceKind.cloudsql
 
     def __init__(
         self,
@@ -27,17 +27,17 @@ class CloudSQLSourceImpl(BaseSQLSource):
         try:
             import psycopg
             import psycopg2
-            import sqlalchemy.dialects
+            from sqlalchemy.dialects import registry  # pyright: ignore
         except ImportError:
             raise missing_dependency_exception("chalkpy[postgresql]")
         del psycopg2  # unused
         del psycopg
-        if "postgresql.psycopg" not in sqlalchemy.dialects.registry.impls:
-            sqlalchemy.dialects.registry.register(
+        if "postgresql.psycopg" not in registry.impls:
+            registry.register(
                 "postgresql.psycopg", "chalk.sql._internal.integrations.psycopg3.psycopg_dialect", "dialect"
             )
-        if "postgresql.psycopg_async" not in sqlalchemy.dialects.registry.impls:
-            sqlalchemy.dialects.registry.register(
+        if "postgresql.psycopg_async" not in registry.impls:
+            registry.register(
                 "postgresql.psycopg_async", "chalk.sql._internal.integrations.psycopg3.psycopg_dialect", "dialect_async"
             )
         prefix = name + "_" if name else ""
