@@ -9,9 +9,10 @@ from pathlib import Path
 from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Literal, Mapping, Optional, Sequence, Type, Union, cast
 
 import yaml
-from pydantic import BaseModel, ValidationError, validator
+from yaml.scanner import ScannerError
 
 from chalk import OfflineResolver, OnlineResolver
+from chalk._lsp.error_builder import SQLFileResolverErrorBuilder
 from chalk.features import DataFrame, Feature, Features, FeatureSetBase
 from chalk.features.resolver import ResolverArgErrorHandler, StreamResolver
 from chalk.sql._internal.integrations.bigquery import BigQuerySourceImpl
@@ -22,15 +23,25 @@ from chalk.sql._internal.integrations.redshift import RedshiftSourceImpl
 from chalk.sql._internal.integrations.snowflake import SnowflakeSourceImpl
 from chalk.sql._internal.integrations.sqlite import SQLiteSourceImpl
 from chalk.sql._internal.sql_source import BaseSQLSource
-from chalk.streams import KafkaSource
+from chalk.streams import KafkaSource, get_resolver_error_builder
 from chalk.streams.base import StreamSource
 from chalk.streams.types import StreamResolverSignature
+from chalk.utils.collections import get_unique_item, get_unique_item_if_exists
 from chalk.utils.duration import parse_chalk_duration
 from chalk.utils.missing_dependency import missing_dependency_exception
 from chalk.utils.string import to_snake_case
 
 if TYPE_CHECKING:
-    from sqlglot.expressions import Expression
+    import sqlglot.expressions
+    from pydantic import BaseModel, ValidationError
+
+    validator = lambda *args, pre=True: (lambda x: x)
+
+else:
+    try:
+        from pydantic.v1 import BaseModel, ValidationError, validator
+    except ImportError:
+        from pydantic import BaseModel, ValidationError, validator
 
 _SOURCES: Mapping[str, Union[Type[BaseSQLSource], Type[StreamSource]]] = {
     "snowflake": SnowflakeSourceImpl,
@@ -44,7 +55,6 @@ _SOURCES: Mapping[str, Union[Type[BaseSQLSource], Type[StreamSource]]] = {
     "kafka": KafkaSource,
 }
 
-
 _SQLGLOT_DIALECTS = frozenset(
     (
         "snowflake",
@@ -107,6 +117,7 @@ class CommentDict(BaseModel):
     owner: Optional[str]
     type: Optional[str]
     timeout: Optional[str]
+    fields: Optional[Dict[str, str]]
 
     @validator("tags", "environment", pre=True)
     @classmethod
@@ -179,7 +190,7 @@ class GlotResult:
     """Class for editing the sql string, and using sqlglot on sql string"""
 
     sql_string: str
-    glot: Optional[Expression]
+    glot: Optional[Union[sqlglot.expressions.Select, sqlglot.expressions.Union]]
     args: Dict[str, str]
     default_args: List[Union[Optional[str], ellipsis]]
     comment_dict: Optional[CommentDict]
@@ -200,9 +211,17 @@ class ParseResult:
     errors: List[ResolverError]
 
 
+_filepath_to_sql_string: dict[str, str] = {}
+"""Mapping from filepath to sql string. Used to skip reimporting the same sql file if the content is identical to what was already imported
+If the content is different, then we'll import it again, but we may error later when attempting to add the resolver to the registry if we don't
+allow overrides and we're not in a notebook.
+"""
+
+
 def get_sql_file_resolvers(
     sql_file_resolve_location: Path,
     sources: Sequence[BaseSQLSource],
+    has_import_errors: bool,
 ) -> Iterable[ResolverResult]:
     """Iterate through all `.chalk.sql` filepaths, gather the sql strings, and get a resolver hopefully for each."""
     for dp, dn, fn in os.walk(os.path.expanduser(sql_file_resolve_location)):
@@ -211,24 +230,39 @@ def get_sql_file_resolvers(
             filepath = os.path.join(dp, f)
             if not filepath.endswith(".chalk.sql"):
                 continue
+                # Already imported this file. Skipping it, assuming it did not change since the last time we imported it
             sql_string_result = _get_sql_string(filepath)
-            yield get_sql_file_resolver(sources, sql_string_result)
+            existing_sql_string = _filepath_to_sql_string.get(filepath)
+            if sql_string_result.sql_string is not None:
+                if existing_sql_string is not None and existing_sql_string == sql_string_result.sql_string:
+                    # The sql file is identical to what was already imported. skip it
+                    continue
+                _filepath_to_sql_string[filepath] = sql_string_result.sql_string
+            yield get_sql_file_resolver(sources, sql_string_result, has_import_errors)
 
 
 def get_sql_file_resolvers_from_paths(
-    sources: Sequence[BaseSQLSource],
-    paths: List[str],
-) -> List[ResolverResult]:
-    return [
-        get_sql_file_resolver(
-            sources=sources,
-            sql_string_result=_get_sql_string(p),
+    sources: Sequence[BaseSQLSource], paths: List[str], has_import_errors: bool
+) -> Iterable[ResolverResult]:
+    for p in paths:
+        sql_string_result = _get_sql_string(p)
+        existing_sql_string = _filepath_to_sql_string.get(p)
+        if sql_string_result.sql_string is not None:
+            if existing_sql_string is not None and existing_sql_string == sql_string_result.sql_string:
+                # The sql file is identical to what was already imported. skip it
+                continue
+            _filepath_to_sql_string[p] = sql_string_result.sql_string
+        yield get_sql_file_resolver(
+            sources=sources, sql_string_result=sql_string_result, has_import_errors=has_import_errors
         )
-        for p in paths
-    ]
 
 
-def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: SQLStringResult) -> ResolverResult:
+def get_sql_file_resolver(
+    sources: Iterable[BaseSQLSource], sql_string_result: SQLStringResult, has_import_errors: bool = False
+) -> ResolverResult:
+    assert sql_string_result.sql_string is not None, f"SQL string from path {sql_string_result.path} should not be None"
+    error_builder = SQLFileResolverErrorBuilder(uri=sql_string_result.path, sql_string=sql_string_result.sql_string)
+
     """Parse the sql strings and get a ResolverResult from each"""
     if sql_string_result.error:
         return ResolverResult(
@@ -238,14 +272,11 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
             fields=None,
             args=None,
         )
-    assert sql_string_result.sql_string is not None
     path = sql_string_result.path
 
     errors: List[ResolverError] = []
     glot_result = _get_sql_glot(
-        sql_string=sql_string_result.sql_string,
-        path=path,
-        sources=sources,
+        sql_string=sql_string_result.sql_string, path=path, sources=sources, error_builder=error_builder
     )
     if glot_result.errors:
         return ResolverResult(
@@ -256,7 +287,13 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
             args=None,
         )
 
-    parsed = _parse_glot(glot_result, path, sources)
+    parsed = _parse_glot(
+        glot_result=glot_result,
+        path=path,
+        sources=sources,
+        error_builder=error_builder,
+        has_import_errors=has_import_errors,
+    )
     if parsed.errors:
         return ResolverResult(
             resolver=None,
@@ -272,45 +309,111 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
         try:
             inputs.append(Feature.from_root_fqn(arg))
         except Exception:
-            errors.append(
-                ResolverError(
-                    display=f"The file '{path}' references an input feature '{arg}' which does not exist",
-                    path=path,
-                    parameter=arg,
+            message = f"SQL file resolver references an input feature '{arg}' which does not exist."
+            if arg in FeatureSetBase.registry:
+                message += f" It appears '{arg}' is a feature class, not a feature."
+
+            if not has_import_errors:
+                error_builder.add_diagnostic(
+                    message=message,
+                    code="152",
+                    label="input feature not recognized",
+                    range=error_builder.variable_range_by_name(arg),
+                )
+                errors.append(
+                    ResolverError(
+                        display=message,
+                        path=path,
+                        parameter=arg,
+                    )
                 )
-            )
     outputs: List[Feature] = []
     query_fields: Dict[str, str] = {}
     for variable, output in parsed.fields.items():
         if output.endswith("*"):
             features = FeatureSetBase.registry[parsed.namespace].features
             for feature in features:
-                if feature.is_scalar:
+                if (
+                    feature.is_scalar
+                    and not feature.is_autogenerated
+                    and not feature.is_pseudofeature
+                    and not feature.is_windowed
+                ):
                     outputs.append(feature)
             continue
-        query_fields[variable] = output
 
+        unrecognized_output = False
         try:
-            outputs.append(Feature.from_root_fqn(output))
+            feature = Feature.from_root_fqn(output)
+            outputs.append(feature)
+            query_fields[variable] = output
         except Exception:
-            errors.append(
-                ResolverError(
-                    display=f"The file '{path}' references an output feature '{output}' which does not exist",
-                    path=path,
-                    parameter=output,
+            if parsed.comment_dict.fields is not None:
+                split = output.split(".", 1)
+                namespace = split[0]
+                output_name = split[1]
+                if output_name in parsed.comment_dict.fields:
+                    field = parsed.comment_dict.fields.get(output_name)
+                    output = f"{namespace}.{field}"
+                    try:
+                        feature = Feature.from_root_fqn(output)
+                        outputs.append(feature)
+                        query_fields[output_name] = output
+                    except:
+                        unrecognized_output = True
+                else:
+                    unrecognized_output = True
+            else:
+                unrecognized_output = True
+        if unrecognized_output:
+            message = f"SQL file resolver references an output feature '{output}' which does not exist. "
+            if not has_import_errors:
+                value = output.split(".", 1)[1]
+                assert glot_result.glot is not None, f"Failed to parse {glot_result.sql_string}"
+                error_builder.add_diagnostic(
+                    message=message,
+                    code="153",
+                    label="output feature not recognized",
+                    range=error_builder.value_range_by_name(glot_result.glot, value),
                 )
+                errors.append(
+                    ResolverError(
+                        display=message,
+                        path=path,
+                        parameter=output,
+                    )
+                )
+    if len(outputs) == 0:
+        message = (
+            "SQL file resolver has no detected outputs. Make sure that all select outputs are aliased to features "
+            "defined within the feature set referenced by the `resolves` parameter."
+        )
+        if has_import_errors:
+            return ResolverResult(
+                resolver=None,
+                errors=[],
+                db=parsed.source,
+                fields=None,
+                args=None,
             )
+        error_builder.add_diagnostic(
+            message=message,
+            code="154",
+            label="no outputs detected",
+            range=error_builder.sql_range(),
+        )
+        errors.append(ResolverError(display=message, path=path, parameter=None))
 
     resolver_type_str = parsed.comment_dict.type if parsed.comment_dict.type else "online"
     resolver_type = _RESOLVER_TYPES[resolver_type_str]
 
     if resolver_type == StreamResolver:
-        return _get_stream_resolver(path, glot_result, parsed, outputs)
+        return _get_stream_resolver(path, glot_result, parsed, outputs, error_builder)
 
     incremental_dict = parsed.comment_dict.incremental.dict() if parsed.comment_dict.incremental else None
     return_one = parsed.comment_dict.count
     source = parsed.source
-    assert isinstance(source, BaseSQLSource)
+    assert isinstance(source, BaseSQLSource), f"source {source} is not configured. Is the driver installed?"
 
     # function for online resolver to process
     def fn(
@@ -339,6 +442,12 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
         return ResolverResult(
             resolver=None, errors=errors, db=parsed.source, fields=parsed.fields, args=glot_result.args
         )
+    if return_one is None:
+        # If the root ns of the inputs is the same as what is being resolved, then it is always return one
+        input_root_namespace = get_unique_item_if_exists(x.root_namespace for x in inputs)
+        output_root_namespace = get_unique_item(x.root_namespace for x in outputs)
+        if input_root_namespace is not None and input_root_namespace == output_root_namespace:
+            return_one = "one"
 
     if return_one:
         output = Features[tuple(outputs)]
@@ -354,6 +463,7 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
     # attempt to instantiate the resolver
     try:
         assert resolver_type in (OnlineResolver, OfflineResolver)
+        assert isinstance(parsed.source, BaseSQLSource)
         resolver = resolver_type(
             filename=path,
             function_definition=sql_string_result.sql_string,
@@ -364,7 +474,6 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
             fn=fn,
             environment=parsed.comment_dict.environment,
             tags=parsed.comment_dict.tags,
-            max_staleness=parsed.comment_dict.max_staleness,
             cron=parsed.comment_dict.cron,
             machine_type=parsed.comment_dict.machine_type,
             when=None,
@@ -373,11 +482,23 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
             owner=parsed.comment_dict.owner,
             timeout=parsed.comment_dict.timeout,
             is_sql_file_resolver=True,
+            data_sources=[parsed.source],
+            name=None,
+            source_line=None,
+            lsp_builder=get_resolver_error_builder(fn),
+            underscore=None,
         )
     except Exception as e:
+        message = f"SQL file resolver '{resolver_type_str.capitalize()}'  could not be instantiated, {e}"
+        error_builder.add_diagnostic(
+            message=message,
+            code="155",
+            label="resolver instantiation failed",
+            range=error_builder.full_range(),
+        )
         errors.append(
             ResolverError(
-                display=f"{resolver_type_str.capitalize()} resolver could not be instantiated, {e}",
+                display=message,
                 path=path,
                 parameter=None,
             )
@@ -401,7 +522,12 @@ def _get_sql_string(path: str) -> SQLStringResult:
         with open(path) as f:
             sql_string = f.read()
     else:
-        caller_filename = inspect.stack()[1].filename
+        frame = inspect.currentframe()
+        assert frame is not None, "could not inspect current frame"
+        caller_frame = frame.f_back
+        assert caller_frame is not None, "could not inspect caller frame"
+        caller_filename = inspect.getsourcefile(caller_frame)
+        assert caller_filename is not None, "could not find caller filename"
         dir_path = os.path.dirname(os.path.realpath(caller_filename))
         if isinstance(path, bytes):
             path = path.decode("utf-8")
@@ -414,10 +540,13 @@ def _get_sql_string(path: str) -> SQLStringResult:
     return SQLStringResult(path=path, sql_string=sql_string, error=None)
 
 
-def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource]) -> GlotResult:
+def _get_sql_glot(
+    sql_string: str, path: str, sources: Iterable[BaseSQLSource], error_builder: SQLFileResolverErrorBuilder
+) -> GlotResult:
     """Get sqlglot from sql string and gracefully exit if unable to"""
     try:
         import sqlglot
+        import sqlglot.errors
         import sqlglot.expressions
     except ImportError:
         raise missing_dependency_exception("chalkpy[runtime]")
@@ -434,12 +563,19 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
             if split_var in variable:
                 split = variable.split(split_var)
                 if len(split) != 2:
+                    message = (
+                        f"If character '|' is used, both variable name and default value must be "
+                        f"specified in '({variable})' like '?{{variable_name | \"default_value\"}}"
+                    )
+                    error_builder.add_diagnostic(
+                        message=message,
+                        code="140",
+                        label="invalid variable",
+                        range=error_builder.variable_range_by_name(variable),
+                    )
                     errors.append(
                         ResolverError(
-                            display=(
-                                f"If character '|' is used, both variable name and default value must be "
-                                f'specified in ({variable}) like \'?{{variable_name | "default_value"}}'
-                            ),
+                            display=message,
                             path=path,
                             parameter=None,
                         )
@@ -460,25 +596,46 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
         args[sql_safe_str] = variable
     comments = ""
     docstring = ""
-    for comment in sql_string.splitlines():
+    comment_row_to_file_row: Dict[int, int] = {}
+    comment_row_offset: Dict[int, int] = {}
+    comment_line_counter = 0
+    """
+    Comments and docstrings are required to be at the beginning of the file resolver.
+    Thus, we need to make sure that for ranges, comments are mapped to the right line number
+    as docstrings are not included in the comments variable above.
+
+    We also need to keep track of the offset of the comment in order to properly detect the range.
+    """
+    for line_no, comment in enumerate(sql_string.splitlines()):
         if comment.strip().startswith("--"):
-            comment = comment.strip().replace("--", "")
-            if comment.strip().startswith("-"):
-                comments += f"{comment}\n"
+            stripped_comment = comment.strip().replace("--", "")
+            comment_offset = len(comment) - len(stripped_comment)
+            if stripped_comment.strip().startswith("-"):
+                comments += f"{stripped_comment}\n"
             else:
-                split = comment.split(":")
+                split = stripped_comment.split(":")
                 if len(split) != 2:
-                    docstring += f"{comment.strip()}\n"
+                    docstring += f"{stripped_comment.strip()}\n"
                 else:
-                    comments += f"{comment}\n"
+                    comments += f"{stripped_comment}\n"
+                    comment_row_to_file_row[comment_line_counter] = line_no
+                    comment_row_offset[comment_line_counter] = comment_offset
+                    comment_line_counter += 1
         else:
             break
 
     if len(comments) == 0:
+        message = "SQL file resolvers require comments that describe key-value pairs in YAML form."
+        error_builder.add_diagnostic(
+            message=message,
+            code="141",
+            label="missing comments",
+            range=error_builder.full_range(),
+            code_href="https://docs.chalk.ai/docs/sql#sql-file-resolvers",
+        )
         errors.append(
             ResolverError(
-                display="SQL file resolvers require comments that describe key-value pairs in YAML form. "
-                "See https://docs.chalk.ai/docs/sql#sql-file-resolvers for reference",
+                display=message,
                 path=path,
                 parameter=None,
             )
@@ -495,9 +652,22 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
     try:
         comment_dict: Dict[str, Any] = yaml.safe_load(comments)
     except Exception as e:
+        message = f"SQL File resolver comments must have key-values in YAML form: {e}"
+        if isinstance(e, ScannerError) and e.problem_mark is not None:
+            comment_line_no = e.problem_mark.line
+            comment_col_no = e.problem_mark.column
+            file_line_no = comment_row_to_file_row[comment_line_no]
+            file_comment_no = comment_row_offset[comment_line_no] + comment_col_no
+            error_builder.add_diagnostic(
+                message=message,
+                code="142",
+                label="Could not parse comments as YAML",
+                range=error_builder.custom_range(line_no=file_line_no + 1, col=file_comment_no + 1),
+            )
+
         errors.append(
             ResolverError(
-                display=f"Comments key-values '{comments}' must be of YAML form, {e}",
+                display=message,
                 path=path,
                 parameter=comments,
             )
@@ -511,13 +681,83 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
             docstring=docstring,
             errors=errors,
         )
+    if "source" not in comment_dict:
+        """This would be caught by the following BaseModel.parse_obj() but the error message is bad"""
+        message = (
+            f"The datasource is a required field for SQL file resolvers. "
+            f"Please add a comment ' -- source: my_name'  where 'my_name' refers to a named integration. "
+        )
+        error_builder.add_diagnostic(
+            message=message,
+            code="143",
+            label="missing source",
+            range=error_builder.full_comment_range(),
+            code_href="https://docs.chalk.ai/docs/integrations",
+        )
+        errors.append(
+            ResolverError(
+                display=message,
+                path=path,
+                parameter=json.dumps(comment_dict),
+            )
+        )
+        return GlotResult(
+            sql_string=sql_string,
+            glot=None,
+            args=args,
+            default_args=default_args,
+            comment_dict=None,
+            docstring=docstring,
+            errors=errors,
+        )
+    if "resolves" not in comment_dict:
+        """This would be caught by the following BaseModel.parse_obj() but the error message is bad"""
+        message = (
+            f"A feature class must be specified for SQL file resolvers. "
+            f"Please add a comment ' -- resolves: my_name'  where 'my_name' refers to a feature class."
+        )
+        error_builder.add_diagnostic(
+            message=message,
+            code="144",
+            label="missing feature class",
+            range=error_builder.full_comment_range(),
+            code_href="https://docs.chalk.ai/docs/integrations",
+        )
+        errors.append(
+            ResolverError(
+                display=message,
+                path=path,
+                parameter=json.dumps(comment_dict),
+            )
+        )
+        return GlotResult(
+            sql_string=sql_string,
+            glot=None,
+            args=args,
+            default_args=default_args,
+            comment_dict=None,
+            docstring=docstring,
+            errors=errors,
+        )
+
     try:
         comment_dict_object = CommentDict.parse_obj(comment_dict)
     except ValidationError as e:
         for error in e.errors():
+            location = error["loc"][-1]  # the innermost error
+            message = f"SQL file resolver could not validate comment '{location}': {error['msg']}"
+            range = error_builder.comment_range_by_key(str(location))
+            if range is None and location in IncrementalSettings.__fields__:
+                range = error_builder.comment_range_by_key("incremental")
+            error_builder.add_diagnostic(
+                message=message,
+                code="145",
+                label="Could not validate comment",
+                range=range,
+            )
             errors.append(
                 ResolverError(
-                    display=f"Could not parse comment(s) for SQL file resolver at '{path}': '{'.'.join(str(x) for x in error['loc'])}': {error['msg']}",
+                    display=message,
                     path=path,
                     parameter=json.dumps(comment_dict),
                 )
@@ -547,9 +787,20 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
         if source_comment in _SQLGLOT_DIALECTS:  # source type, if supported
             source_type = source_comment
     try:
-        glot = sqlglot.parse_one(sql=sql_string, read=source_type)
+        glots = sqlglot.parse(sql=sql_string, read=source_type)
     except Exception as e:
-        errors.append(ResolverError(display=f"Cannot SQL parse {sql_string}, {e}", path=path, parameter=None))
+        message = f"SQL file resolver could not SQL parse string: {e}"
+        if isinstance(e, sqlglot.errors.ParseError):
+            for error in e.errors:
+                error_builder.add_diagnostic(
+                    message=message,
+                    code="146",
+                    label="could not parse SQL",
+                    range=error_builder.custom_range(
+                        line_no=error["line"], col=error["col"] - 2, length=len(error["highlight"])  # ????
+                    ),
+                )
+        errors.append(ResolverError(display=message, path=path, parameter=None))
         return GlotResult(
             sql_string=sql_string,
             glot=None,
@@ -559,10 +810,70 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
             docstring=docstring,
             errors=errors,
         )
-    if not isinstance(glot, sqlglot.expressions.Select):
-        errors.append(
-            ResolverError(display=f"SQL query {sql_string} should be of 'SELECT' type", path=path, parameter=None)
+    if len(glots) > 1:
+        message = f"SQL file resolver query {sql_string} has more than one 'SELECT' statements. Only one is permitted."
+        error_builder.add_diagnostic(
+            message=message,
+            code="147",
+            label="SQL query must be a single SELECT statement",
+            range=error_builder.sql_range(),
         )
+        errors.append(ResolverError(display=message, path=path, parameter=None))
+        return GlotResult(
+            sql_string=sql_string,
+            glot=None,
+            args=args,
+            default_args=default_args,
+            comment_dict=comment_dict_object,
+            docstring=docstring,
+            errors=errors,
+        )
+    glot = glots[0]
+    if not isinstance(glot, (sqlglot.expressions.Select, sqlglot.expressions.Union)):
+        message = f"SQL file resolver query {sql_string} should be of 'SELECT' type"
+        error_builder.add_diagnostic(
+            message=message,
+            code="147",
+            label="SQL query must be a SELECT statement",
+            range=error_builder.sql_range(),
+        )
+        errors.append(ResolverError(display=message, path=path, parameter=None))
+        return GlotResult(
+            sql_string=sql_string,
+            glot=None,
+            args=args,
+            default_args=default_args,
+            comment_dict=comment_dict_object,
+            docstring=docstring,
+            errors=errors,
+        )
+    if len(glot.selects) > len(glot.named_selects):
+        for select in glot.selects:
+            matched = False
+            for named_select in glot.named_selects:
+                if select.alias_or_name == named_select:
+                    matched = True
+            if not matched:
+                message = (
+                    f"SQL file resolver query with unnamed select '{str(select)}'. "
+                    f"All selects must either match a feature name, e.g. 'id', or must be aliased, e.g. "
+                    f"'SELECT COUNT (DISTINCT merchant_id) AS num_unique_merchant_ids. "
+                    f"All names/aliases must match to features on the feature set defined "
+                    f"by the 'resolves' comment parameter. "
+                )
+                error_builder.add_diagnostic(
+                    message=message,
+                    code="148",
+                    label="all selects must map to features",
+                    range=error_builder.value_range_by_name(glot=glot, name=str(select).lower()),
+                )
+                errors.append(
+                    ResolverError(
+                        display=message,
+                        path=path,
+                        parameter=None,
+                    )
+                )
     return GlotResult(
         sql_string=sql_string,
         glot=glot,
@@ -574,7 +885,13 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
     )
 
 
-def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSource]) -> ParseResult:
+def _parse_glot(
+    glot_result: GlotResult,
+    path: str,
+    sources: Iterable[BaseSQLSource],
+    error_builder: SQLFileResolverErrorBuilder,
+    has_import_errors: bool,
+) -> ParseResult:
     """Parse useful info from sqlglot and gracefully exit if unable to"""
     try:
         import sqlglot
@@ -583,7 +900,7 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
         raise missing_dependency_exception("chalkpy[runtime]")
     # define a source SQL database. Can either specify name or kind if only one of the kind is present.
     comment_dict = glot_result.comment_dict
-    assert comment_dict is not None
+    assert comment_dict is not None, f"comment dict failed to parse"
     docstring = glot_result.docstring
     source_name = comment_dict.source
     source = None
@@ -594,21 +911,48 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
                 source = possible_source
     else:
         for possible_source in sources:
-            if isinstance(possible_source, _SOURCES[source_name]):
+            source_type = _SOURCES.get(source_name)
+            if source_type is not None and isinstance(possible_source, source_type):
                 if source:
+                    message = (
+                        f"SQL file resolver refers to '{source_name}' when more than one {source_name}-type source "
+                        f"exists. Instead, refer to the integration by name among "
+                        f"({[source.name for source in sources]})."
+                    )
+                    error_builder.add_diagnostic(
+                        message=message,
+                        code="149",
+                        label="Refer to source via name instead",
+                        range=error_builder.comment_range_by_key("source"),
+                    )
                     errors.append(
                         ResolverError(
-                            display=(
-                                f"More than one {source_name} source exists. Instead, refer to the integration by "
-                                f"name among ({[source.name for source in sources]})."
-                            ),
+                            display=message,
                             path=path,
                             parameter=source_name,
                         )
                     )
                 source = possible_source
     if not source:
-        errors.append(ResolverError(display=f"Source '{source_name}' not found", path=path, parameter=source_name))
+        message = (
+            f"SQL file resolver refers to unrecognized source '{source_name}'. "
+            f"Please refer to your source via its name on your Chalk dashboard, "
+            f"and make sure the driver, e.g. chalkpy[snowflake], is installed"
+        )
+        if not has_import_errors:
+            error_builder.add_diagnostic(
+                message=message,
+                code="150",
+                label="Source not found",
+                range=error_builder.comment_range_by_key("source"),
+            )
+        errors.append(
+            ResolverError(
+                display=message,
+                path=path,
+                parameter=source_name,
+            )
+        )
     # get resolver fields: which columns selected will match to which chalk feature?
     namespace = comment_dict.resolves
     if namespace not in FeatureSetBase.registry:
@@ -616,28 +960,14 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
         if snaked_namespace in FeatureSetBase.registry:
             namespace = snaked_namespace
         else:
-            errors.append(
-                ResolverError(
-                    display=f"No @features class with the name '{namespace}'", path=path, parameter=source_name
-                )
+            message = f"No @features class with the name '{namespace}'"
+            error_builder.add_diagnostic(
+                message=message,
+                code="151",
+                label="Unrecognized namespace ",
+                range=error_builder.comment_range_by_key("resolves"),
             )
-
-    if glot_result.glot is None:
-        errors.append(
-            ResolverError(
-                display=f"Could not parse SQL query {glot_result.sql_string}.",
-                path=path,
-                parameter=source_name,
-            )
-        )
-    elif not isinstance(glot_result.glot, sqlglot.expressions.Select):
-        errors.append(
-            ResolverError(
-                display=f"SQL query {glot_result.sql_string} should be a 'SELECT'",
-                path=path,
-                parameter=None,
-            )
-        )
+            errors.append(ResolverError(display=message, path=path, parameter=source_name))
 
     if len(errors) > 0:
         return ParseResult(
@@ -650,7 +980,9 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
             errors=errors,
         )
 
-    assert isinstance(glot_result.glot, sqlglot.expressions.Select)
+    assert isinstance(
+        glot_result.glot, (sqlglot.expressions.Select, sqlglot.expressions.Union)
+    ), f"glot {glot_result.glot} is not a select or union statement"
     # sql target -> output feature string
     fields: Dict[str, str] = {
         column_name: f"{namespace}.{column_name}" for column_name in glot_result.glot.named_selects
@@ -668,7 +1000,11 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
 
 
 def _get_stream_resolver(
-    path: str, glot_result: GlotResult, parsed: ParseResult, outputs: List[Feature]
+    path: str,
+    glot_result: GlotResult,
+    parsed: ParseResult,
+    outputs: List[Feature],
+    error_builder: SQLFileResolverErrorBuilder,
 ) -> ResolverResult:
     errors = []
     output_features = Features[DataFrame[tuple(outputs)]]
@@ -685,8 +1021,6 @@ def _get_stream_resolver(
 
     sql_query: str = _remove_comments(parsed.sql_string)
     filename = os.path.basename(path)
-    assert isinstance(parsed.source, StreamSource)
-
     try:
 
         def fn():
@@ -697,14 +1031,13 @@ def _get_stream_resolver(
             function_definition=sql_query,
             fqn=filename.replace(".chalk.sql", ""),
             filename=path,
-            source=parsed.source,
+            source=cast(StreamSource, parsed.source),
             fn=fn,
             environment=parsed.comment_dict.environment,
             doc=parsed.docstring,
-            module=filename.replace(".chalk.sql", ""),
             mode=None,
             machine_type=parsed.comment_dict.machine_type,
-            message=parsed.comment_dict.message,
+            message=None,
             output=output_features,
             signature=signature,
             state=None,
@@ -713,11 +1046,22 @@ def _get_stream_resolver(
             parse=None,
             keys=None,  # TODO implement parse and keys for sql file resolvers?
             timestamp=None,
+            source_line=None,
+            name=None,
+            tags=None,
+            lsp_builder=get_resolver_error_builder(fn),
         )
     except Exception as e:
+        message = f"Streaming SQL file resolver could not be instantiated, {e}"
+        error_builder.add_diagnostic(
+            message=message,
+            code="141",
+            label="resolver instantiation failed",
+            range=error_builder.full_range(),
+        )
         errors.append(
             ResolverError(
-                display=f"Streaming resolver could not be instantiated, {e}",
+                display=message,
                 path=path,
                 parameter=None,
             )
@@ -725,7 +1069,11 @@ def _get_stream_resolver(
         return ResolverResult(resolver=None, errors=errors, db=parsed.source, fields=None, args=None)
 
     return ResolverResult(
-        resolver=resolver, errors=errors, db=parsed.source, fields=parsed.fields, args=glot_result.args
+        resolver=resolver,
+        errors=errors,
+        db=parsed.source,
+        fields=parsed.fields,
+        args=glot_result.args,
     )
 
 
