@@ -14,9 +14,11 @@ from os import PathLike
 from typing import (
     TYPE_CHECKING,
     Any,
+    AsyncIterable,
     Callable,
     ClassVar,
     Dict,
+    Iterable,
     List,
     Mapping,
     Optional,
@@ -30,14 +32,19 @@ from typing import (
 
 import pyarrow as pa
 
+from chalk.clogging import chalk_logger
 from chalk.features import Feature, FeatureConverter, Features, FeatureWrapper, unwrap_feature
 from chalk.integrations.named import load_integration_variable
 from chalk.sql._internal.chalk_query import ChalkQuery
 from chalk.sql._internal.incremental import IncrementalSettings
+from chalk.sql._internal.query_execution_parameters import (
+    QueryExecutionParameters,
+    query_execution_parameters_from_env_vars,
+)
 from chalk.sql._internal.string_chalk_query import StringChalkQuery
 from chalk.sql.finalized_query import FinalizedChalkQuery
 from chalk.sql.protocols import BaseSQLSourceProtocol, ChalkQueryProtocol, StringChalkQueryProtocol, TableIngestProtocol
-from chalk.utils.async_helpers import async_null_context
+from chalk.utils.async_helpers import async_null_context, to_async_iterable
 from chalk.utils.environment_parsing import env_var_bool
 from chalk.utils.log_with_context import get_logger, get_logging_context
 from chalk.utils.missing_dependency import missing_dependency_exception
@@ -47,10 +54,13 @@ if TYPE_CHECKING:
     from sqlalchemy.engine import URL, Connection, Dialect, Engine
     from sqlalchemy.orm import Session
     from sqlalchemy.sql import Select
-    from sqlalchemy.sql.compiler import Compiled
+    from sqlalchemy.sql.compiler import SQLCompiler
+    from sqlalchemy.sql.ddl import CreateTable, DropTable
     from sqlalchemy.sql.elements import Label
+    from sqlalchemy.sql.schema import Table
     from sqlalchemy.types import TypeEngine
 
+
 TTableIngestMixIn = TypeVar("TTableIngestMixIn", bound="TableIngestMixIn")
 CHALK_QUERY_LOGGING = env_var_bool("CHALK_QUERY_LOGGING")
 
@@ -139,6 +149,7 @@ class TableIngestMixIn(TableIngestProtocol):
 
 class SQLSourceKind(str, Enum):
     bigquery = "bigquery"
+    cloudsql = "cloudsql"
     databricks = "databricks"
     mysql = "mysql"
     postgres = "postgres"
@@ -158,6 +169,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         engine_args: Optional[Dict[str, Any]],
         async_engine_args: Optional[Dict[str, Any]],
     ):
+        super().__init__()
         try:
             import sqlalchemy
         except ImportError:
@@ -190,10 +202,6 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         This allows for use of dialect-specific syntax while parsing and modifying queries."""
         return None
 
-    def get_sqglot_dialect(self) -> Union[str, None]:
-        """(DEPRECATED: `get_sqlglot_dialect` INSTEAD)"""
-        return self.get_sqlglot_dialect()
-
     def query_sql_file(
         self,
         path: Union[str, bytes, PathLike],
@@ -201,6 +209,8 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         args: Optional[Mapping[str, object]] = None,
     ) -> StringChalkQueryProtocol:
         sql_string = None
+        if isinstance(path, bytes):
+            path = path.decode("utf-8")
 
         resolver_fqn = get_logging_context().get("labels", {}).get("resolver_fqn")
         if resolver_fqn is not None:
@@ -218,8 +228,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                 caller_frame = frame.f_back
                 assert caller_frame is not None
                 dir_path = os.path.dirname(os.path.realpath(inspect.getframeinfo(caller_frame).filename))
-                if isinstance(path, bytes):
-                    path = path.decode("utf-8")
+                path = str(path)
                 relative_path = os.path.join(dir_path, path)
                 if os.path.isfile(relative_path):
                     with open(relative_path) as f:
@@ -279,14 +288,24 @@ class BaseSQLSource(BaseSQLSourceProtocol):
     def execute_query(
         self,
         finalized_query: FinalizedChalkQuery,
-        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
         connection: Optional[Connection] = None,
-        attempt_efficient_execution: bool = True,
-        polars_read_csv: bool = False,
-    ) -> pa.Table:
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
+    ) -> Iterable[pa.RecordBatch]:
+        if query_execution_parameters is None:
+            """
+            Sometimes, in tests or in development or in user resolver code,
+            we may not have the query_execution_parameters argument
+            passed in. Thus, we need to hack the env vars in.
+            """
+            query_execution_parameters = query_execution_parameters_from_env_vars()
+        attempt_efficient_execution = query_execution_parameters.attempt_efficient_execution
         if attempt_efficient_execution:
             try:
-                return self.execute_query_efficient(finalized_query, columns_to_converters, connection, polars_read_csv)
+                yield from self._execute_query_efficient(
+                    finalized_query, columns_to_features, connection, query_execution_parameters
+                )
+                return
             except NotImplementedError:
                 _logger.debug(
                     (
@@ -298,30 +317,103 @@ class BaseSQLSource(BaseSQLSourceProtocol):
             except UnsupportedEfficientExecutionError as e:
                 log_level = e.log_level
                 _logger.log(log_level, str(e))
+            except Exception:
+                if query_execution_parameters.fallback_to_inefficient_execution:
+                    _logger.error(
+                        (
+                            f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
+                            f"with parameters {finalized_query.params} due to an unexpected error. Falling back to inefficient execution."
+                        ),
+                        exc_info=True,
+                    )
+                else:
+                    raise
 
-        return self.execute_query_inefficient(finalized_query, columns_to_converters, connection)
+        batch = self._execute_query_inefficient(
+            finalized_query, columns_to_features, connection, query_execution_parameters
+        )
+        if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
+            yield batch
+
+    @contextlib.contextmanager
+    def _create_temp_table(
+        self,
+        create_temp_table: CreateTable,
+        temp_table: Table,
+        drop_temp_table: DropTable,
+        connection: sqlalchemy.engine.Connection,
+        temp_value: pa.Table,
+    ):
+        chalk_logger.info(f"Creating temporary table {temp_table.name}.")
+        connection.execute(create_temp_table)
+        try:
+            batch_size = insert_size_limit // temp_value.num_columns
+            assert temp_value.num_columns <= insert_size_limit, "temp temp_value has too many columns to insert rowwise"
+            for batch_start in range(0, len(temp_value), batch_size):
+                batch_end = min(batch_start + batch_size, len(temp_value))
+                connection.execute(
+                    temp_table.insert(), temp_value.slice(batch_start, batch_end - batch_start).to_pylist()
+                )
+            yield
+        finally:
+            # "temp table", to snowflake, means that it belongs to the session. However, we keep using the same Snowflake session
+            chalk_logger.info(f"Dropping temporary table {temp_table.name}.")
+            connection.execute(drop_temp_table)
+
+    @contextlib.asynccontextmanager
+    async def _async_create_temp_table(
+        self,
+        create_temp_table: CreateTable,
+        temp_table: Table,
+        drop_temp_table: DropTable,
+        connection: sqlalchemy.ext.asyncio.AsyncConnection,
+        temp_value: pa.Table,
+    ):
+        chalk_logger.info(f"Creating temporary table {temp_table.name}.")
+        await connection.execute(create_temp_table)
+        try:
+            batch_size = insert_size_limit // temp_value.num_columns
+            assert temp_value.num_columns <= insert_size_limit, "temp temp_value has too many columns to insert rowwise"
+            for batch_start in range(0, len(temp_value), batch_size):
+                batch_end = min(batch_start + batch_size, len(temp_value))
+                await connection.execute(
+                    temp_table.insert(), temp_value.slice(batch_start, batch_end - batch_start).to_pylist()
+                )
+            yield
+        finally:
+            # "temp table", to snowflake, means that it belongs to the session. However, we keep using the same Snowflake session
+            chalk_logger.info(f"Dropping temporary table {temp_table.name}.")
+            await connection.execute(drop_temp_table)
 
     async def async_execute_query(
         self,
         finalized_query: FinalizedChalkQuery,
-        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
         connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
-        attempt_efficient_execution: bool = True,
-        polars_read_csv: bool = False,
-    ) -> pa.Table:
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
+    ) -> AsyncIterable[pa.RecordBatch]:
+        if query_execution_parameters is None:
+            query_execution_parameters = query_execution_parameters_from_env_vars()
+        attempt_efficient_execution = query_execution_parameters.attempt_efficient_execution
         if attempt_efficient_execution:
             try:
-                return await self.async_execute_query_efficient(finalized_query, columns_to_converters, connection)
+                async for res in self._async_execute_query_efficient(
+                    finalized_query, columns_to_features, connection, query_execution_parameters
+                ):
+                    yield res
+                return
             except NotImplementedError:
                 try:
-                    return await asyncio.get_running_loop().run_in_executor(
-                        None,
-                        self.execute_query_efficient,
-                        finalized_query,
-                        columns_to_converters,
-                        connection,
-                        polars_read_csv,
-                    )
+                    async for batch in to_async_iterable(
+                        self._execute_query_efficient(
+                            finalized_query,
+                            columns_to_features,
+                            None,  # async connection is not compatible with a non-async connection
+                            query_execution_parameters,
+                        )
+                    ):
+                        yield batch
+                    return
                 except NotImplementedError:
                     _logger.debug(
                         (
@@ -333,113 +425,151 @@ class BaseSQLSource(BaseSQLSourceProtocol):
                 except UnsupportedEfficientExecutionError as e:
                     log_level = e.log_level
                     _logger.log(log_level, str(e))
+                except Exception:
+                    if query_execution_parameters.fallback_to_inefficient_execution:
+                        _logger.error(
+                            (
+                                f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
+                                f"with parameters {finalized_query.params} due to an unexpected error. Falling back to inefficient execution."
+                            ),
+                            exc_info=True,
+                        )
+                    else:
+                        raise
             except UnsupportedEfficientExecutionError as e:
                 log_level = e.log_level
                 _logger.log(log_level, str(e))
+            except Exception:
+                if query_execution_parameters.fallback_to_inefficient_execution:
+                    _logger.error(
+                        (
+                            f"Failed to efficiently execute query {finalized_query.query.compile(compile_kwargs={'literal_binds': True})} "
+                            f"with parameters {finalized_query.params} due to an unexpected error. Falling back to inefficient execution."
+                        ),
+                        exc_info=True,
+                    )
+                else:
+                    raise
 
-        return await self.async_execute_query_inefficient(finalized_query, columns_to_converters, connection)
+        batch = await self._async_execute_query_inefficient(
+            finalized_query, columns_to_features, connection, query_execution_parameters
+        )
+        if len(batch) > 0 or query_execution_parameters.yield_empty_batches:
+            yield batch
 
-    async def async_execute_query_inefficient(
+    async def _async_execute_query_inefficient(
         self,
         finalized_query: FinalizedChalkQuery,
-        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        columns_to_feature: Callable[[Sequence[str]], Mapping[str, Feature]],
         connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection],
-    ) -> pa.Table:
+        query_execution_parameters: QueryExecutionParameters,
+    ) -> pa.RecordBatch:
         if connection is None:
             try:
                 eng = self.get_async_engine()
             except NotImplementedError:
                 return await asyncio.get_running_loop().run_in_executor(
-                    None, self.execute_query_inefficient, finalized_query, columns_to_converters, connection
+                    None,
+                    self._execute_query_inefficient,
+                    finalized_query,
+                    columns_to_feature,
+                    connection,
+                    query_execution_parameters,
                 )
             cnx_ctx = eng.connect()
         else:
             cnx_ctx = async_null_context(connection)
         async with cnx_ctx as cnx:
-            for _, temp_table_values, create_temp_table, temp_table in finalized_query.temp_tables.values():
-                columns = temp_table_values.column_names
-                await cnx.execute(create_temp_table, None)
-                batch_size = insert_size_limit // temp_table_values.num_columns
-                assert (
-                    temp_table_values.num_columns <= insert_size_limit
-                ), "temp table has too many columns to insert rowwise"
-                for batch_start in range(0, len(temp_table_values), batch_size):
-                    batch_end = min(batch_start + batch_size, len(temp_table_values))
-                    await cnx.execute(
-                        temp_table.insert(), temp_table_values.slice(batch_start, batch_end - batch_start).to_pylist()
+            async with contextlib.AsyncExitStack() as exit_stack:
+                for (
+                    _,
+                    temp_value,
+                    create_temp_table,
+                    temp_table,
+                    drop_temp_table,
+                ) in finalized_query.temp_tables.values():
+                    exit_stack.enter_context(
+                        self._async_create_temp_table(  # pyright: ignore -- type stubs are wrong
+                            create_temp_table, temp_table, drop_temp_table, cnx, temp_value
+                        )
                     )
             res = await cnx.execute(finalized_query.query, finalized_query.params)
             desc = res.cursor.description  # type: ignore
             result_columns: list[str] = [col[0] for col in desc]
-            converters = columns_to_converters(result_columns)
+            features = columns_to_feature(result_columns)
             data: Dict[str, List[Any]] = {}
-            for v in converters.keys():
+            for v in features.values():
                 # Create an entry for the columns, so the pyarrow table will have the correct columns even
                 # if there is no data
-                data[v] = []
+                data[v.root_fqn] = []
             for row in res.all():
                 for k, v in zip(result_columns, row):
-                    if k not in data:
+                    if k not in features:
                         # We are not interested in this column
                         continue
-                    data[k].append(self.convert_db_types(v, converters[k]))
+                    data[features[k].root_fqn].append(self.convert_db_types(v, features[k].converter))
             # Only keep the columns provided by the schema
-            schema = pa.schema([pa.field(k, v.pyarrow_dtype) for (k, v) in converters.items()])
-            return pa.Table.from_pydict(data, schema=schema)
+            schema = pa.schema([pa.field(v.root_fqn, v.converter.pyarrow_dtype) for v in features.values()])
+            return pa.RecordBatch.from_pydict(data, schema=schema)
 
-    def execute_query_inefficient(
+    def _execute_query_inefficient(
         self,
         finalized_query: FinalizedChalkQuery,
-        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
         connection: Optional[Connection],
-    ) -> pa.Table:
-        with self.get_engine().connect() if connection is None else contextlib.nullcontext(connection) as cnx:
-            for _, temp_table_values, create_temp_table, temp_table in finalized_query.temp_tables.values():
-                cnx.execute(create_temp_table, None)
-                batch_size = insert_size_limit // temp_table_values.num_columns
-                assert (
-                    temp_table_values.num_columns <= insert_size_limit
-                ), "temp table has too many columns to insert rowwise"
-                for batch_start in range(0, len(temp_table_values), batch_size):
-                    batch_end = min(batch_start + batch_size, len(temp_table_values))
-                    cnx.execute(
-                        temp_table.insert(), temp_table_values.slice(batch_start, batch_end - batch_start).to_pylist()
+        query_execution_parameters: QueryExecutionParameters,
+    ) -> pa.RecordBatch:
+        with (
+            self.get_engine().connect() if connection is None else contextlib.nullcontext(connection)
+        ) as cnx, cnx.begin():
+            with contextlib.ExitStack() as exit_stack:
+                for (
+                    _,
+                    temp_value,
+                    create_temp_table,
+                    temp_table,
+                    drop_temp_table,
+                ) in finalized_query.temp_tables.values():
+                    exit_stack.enter_context(
+                        self._create_temp_table(create_temp_table, temp_table, drop_temp_table, cnx, temp_value)
                     )
-            res = cnx.execute(finalized_query.query, finalized_query.params)
-            desc = res.cursor.description
-            result_columns: list[str] = [col[0] for col in desc]
-            converters = columns_to_converters(result_columns)
-            data: Dict[str, List[Any]] = {}
-            for v in converters.keys():
-                # Create an entry for the columns, so the pyarrow table will have the correct columns even
-                # if there is no data
-                data[v] = []
-            for row in res.all():
-                for k, v in zip(result_columns, row):
-                    if k not in data:
-                        # We are not interested in this column
-                        continue
-                    data[k].append(self.convert_db_types(v, converters[k]))
-            # Only keep the columns provided by the schema
-            schema = pa.schema([pa.field(k, v.pyarrow_dtype) for (k, v) in converters.items()])
-            return pa.Table.from_pydict(data, schema=schema)
+                # TODO: postgres may have timestamp casting errors due to timezones here
+                res = cnx.execute(finalized_query.query, finalized_query.params)
+                desc = res.cursor.description
+                result_columns: list[str] = [col[0] for col in desc]
+                features = columns_to_features(result_columns)
+                data: Dict[str, List[Any]] = {}
+                for v in features.values():
+                    # Create an entry for the columns, so the pyarrow table will have the correct columns even
+                    # if there is no data
+                    data[v.root_fqn] = []
+                for row in res.all():
+                    for k, v in zip(result_columns, row):
+                        if k not in features:
+                            # We are not interested in this column
+                            continue
+                        data[features[k].root_fqn].append(self.convert_db_types(v, features[k].converter))
+                # Only keep the columns provided by the schema
+                schema = pa.schema([pa.field(v.root_fqn, v.converter.pyarrow_dtype) for v in features.values()])
+                return pa.RecordBatch.from_pydict(data, schema=schema)
 
-    def execute_query_efficient(
+    def _execute_query_efficient(
         self,
         finalized_query: FinalizedChalkQuery,
-        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
         connection: Optional[Connection],
-        polars_read_csv: bool = False,
-    ) -> pa.Table:
-        raise NotImplementedError()
+        query_execution_parameters: QueryExecutionParameters,
+    ) -> Iterable[pa.RecordBatch]:
+        raise NotImplementedError
 
-    async def async_execute_query_efficient(
+    def _async_execute_query_efficient(
         self,
         finalized_query: FinalizedChalkQuery,
-        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        columns_to_features: Callable[[Sequence[str]], Mapping[str, Feature]],
         connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection],
-    ):
-        raise NotImplementedError()
+        query_execution_parameters: QueryExecutionParameters,
+    ) -> AsyncIterable[pa.RecordBatch]:
+        raise NotImplementedError
 
     def get_sqlalchemy_dialect(self, paramstyle: Optional[str] = None) -> Dialect:
         return self.local_engine_url().get_dialect()(paramstyle=paramstyle)
@@ -452,13 +582,13 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         """Compile a query into a string and the bindparams"""
         compiled_query = self._get_compiled_query(finalized_query, paramstyle)
         query_string = compiled_query.string
-        return query_string, compiled_query.positiontup or [], compiled_query.params
+        return query_string, compiled_query.positiontup or [], compiled_query.params or {}
 
     def _get_compiled_query(
         self,
         finalized_query: FinalizedChalkQuery,
         paramstyle: Optional[str] = None,
-    ) -> Compiled:
+    ) -> SQLCompiler:
         """Compile a query into a string and the bindparams"""
         dialect = self.get_sqlalchemy_dialect(paramstyle=paramstyle)
         query = finalized_query.query.params(finalized_query.params)
@@ -498,6 +628,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         from sqlalchemy.engine import create_engine
 
         if self._engine is None:
+            self.register_sqlalchemy_compiler_overrides()
             self._check_engine_isolation_level()
             self._engine = create_engine(url=self.local_engine_url(), **self._engine_args)
         return self._engine
@@ -506,6 +637,7 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         from sqlalchemy.ext.asyncio import create_async_engine
 
         if self._async_engine is None:
+            self.register_sqlalchemy_compiler_overrides()
             self._check_engine_isolation_level()
             self._async_engine = create_async_engine(url=self.async_local_engine_url(), **self._async_engine_args)
         return self._async_engine
@@ -530,6 +662,14 @@ class BaseSQLSource(BaseSQLSourceProtocol):
         """
         return converter.from_rich_to_primitive(v, missing_value_strategy="default_or_allow")
 
+    def to_json(self) -> Dict[str, str]:
+        return {"name": self.name or "", "kind": self.kind.value}
+
+    @classmethod
+    def register_sqlalchemy_compiler_overrides(cls):
+        """Hook to register SQLAlchemy Compiler overrides. These hooks are registered by default in tests and in the engine"""
+        return
+
 
 def _extract_features(feature_set: Features, prefix: str, targets: List[Label], features: Dict[str, Feature]):
     import sqlalchemy.sql.functions
