@@ -3,8 +3,9 @@ from __future__ import annotations
 from typing import TYPE_CHECKING, Any, Literal, Mapping, Optional, Union
 
 from chalk.features import Feature
-from chalk.sql.finalized_query import FinalizedChalkQuery, Finalizer
-from chalk.sql.protocols import BaseSQLSourceProtocol, IncrementalSettings, StringChalkQueryProtocol
+from chalk.sql import IncrementalSettings
+from chalk.sql.finalized_query import DataframeFinalizedChalkQuery, Finalizer, SingletonFinalizedChalkQuery
+from chalk.sql.protocols import BaseSQLSourceProtocol, StringChalkQueryProtocol
 from chalk.utils.duration import Duration, parse_chalk_duration
 from chalk.utils.missing_dependency import missing_dependency_exception
 
@@ -20,6 +21,7 @@ class StringChalkQuery(StringChalkQueryProtocol):
         query: Union[str, TextClause],
         fields: Mapping[str, Feature],
     ):
+        super().__init__()
         try:
             from sqlalchemy import text
         except ImportError:
@@ -32,8 +34,8 @@ class StringChalkQuery(StringChalkQueryProtocol):
     def __repr__(self):
         return f"StringChalkQuery(query='{self._query}')"
 
-    def one_or_none(self):
-        return FinalizedChalkQuery(
+    def one_or_none(self) -> SingletonFinalizedChalkQuery:
+        return SingletonFinalizedChalkQuery(
             query=self._query,
             params=self._params,
             finalizer=Finalizer.ONE_OR_NONE,
@@ -42,8 +44,8 @@ class StringChalkQuery(StringChalkQueryProtocol):
             fields=self._fields,
         )
 
-    def one(self):
-        return FinalizedChalkQuery(
+    def one(self) -> SingletonFinalizedChalkQuery:
+        return SingletonFinalizedChalkQuery(
             query=self._query,
             params=self._params,
             finalizer=Finalizer.ONE,
@@ -52,8 +54,8 @@ class StringChalkQuery(StringChalkQueryProtocol):
             fields=self._fields,
         )
 
-    def first(self):
-        return FinalizedChalkQuery(
+    def first(self) -> SingletonFinalizedChalkQuery:
+        return SingletonFinalizedChalkQuery(
             query=self._query,
             params=self._params,
             finalizer=Finalizer.FIRST,
@@ -62,8 +64,8 @@ class StringChalkQuery(StringChalkQueryProtocol):
             fields=self._fields,
         )
 
-    def all(self):
-        return FinalizedChalkQuery(
+    def all(self) -> DataframeFinalizedChalkQuery:
+        return DataframeFinalizedChalkQuery(
             query=self._query,
             params=self._params,
             finalizer=Finalizer.ALL,
@@ -76,10 +78,10 @@ class StringChalkQuery(StringChalkQueryProtocol):
         self,
         *,
         incremental_column: Optional[str] = None,
-        lookback_period: Duration = "0s",
+        lookback_period: Duration | None = "0s",
         mode: Literal["row", "group", "parameter"] = "row",
         incremental_timestamp: Literal["feature_time", "resolver_execution_time"] = "feature_time",
-    ):
+    ) -> DataframeFinalizedChalkQuery:
         if mode in {"row", "group"} and incremental_column is None:
             raise ValueError(f"incremental mode set to '{mode}' but no 'incremental_column' argument was passed.")
 
@@ -89,7 +91,7 @@ class StringChalkQuery(StringChalkQueryProtocol):
                 + " Please view documentation for proper usage."
             )
 
-        return FinalizedChalkQuery(
+        return DataframeFinalizedChalkQuery(
             query=self._query,
             params=self._params,
             finalizer=Finalizer.ALL,
