@@ -1,22 +1,44 @@
 from __future__ import annotations
 
+import dataclasses
 import warnings
 from enum import Enum
-from typing import TYPE_CHECKING, Any, Collection, Dict, FrozenSet, List, Mapping, Optional, Sequence, Tuple, Union
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    AsyncIterable,
+    Collection,
+    Dict,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+    overload,
+)
 
 import pyarrow as pa
 
 from chalk.features import DataFrame, Feature, FeatureConverter, Features
 from chalk.sql._internal.incremental import IncrementalSettings
+from chalk.sql._internal.query_execution_parameters import (
+    QueryExecutionParameters,
+    query_execution_parameters_from_env_vars,
+)
 from chalk.sql.protocols import BaseSQLSourceProtocol
+from chalk.utils.async_helpers import async_enumerate
+from chalk.utils.df_utils import pa_table_to_pl_df
 from chalk.utils.missing_dependency import missing_dependency_exception
 from chalk.utils.string import normalize_string_for_matching
 
 if TYPE_CHECKING:
     import sqlalchemy.ext.asyncio
+    from sqlalchemy import Table
     from sqlalchemy.engine import Connection
-    from sqlalchemy.sql import Insert, Select
-    from sqlalchemy.sql.ddl import CreateTable
+    from sqlalchemy.sql import Select
+    from sqlalchemy.sql.ddl import CreateTable, DropTable
     from sqlalchemy.sql.elements import TextClause
 
 
@@ -59,13 +81,16 @@ class FinalizedChalkQuery:
         incremental_settings: Optional[IncrementalSettings],
         source: BaseSQLSourceProtocol,
         fields: Mapping[str, Feature],
-        temp_tables: Mapping[str, Tuple[Mapping[str, sqlalchemy.types.TypeEngine], pa.Table, CreateTable, Insert]]
-        | None = None,
+        temp_tables: (
+            Mapping[str, Tuple[Mapping[str, sqlalchemy.types.TypeEngine], pa.Table, CreateTable, Table, DropTable]]
+            | None
+        ) = None,
         is_empty: bool = False,
     ) -> None:
         """
         :param fields: may be set to None, in which case it can be inferred from the output type of the resolver which returned this query using infer_fields_from_resolver_output
         """
+        super().__init__()
         self._query = query
         self._params = dict(params)
         self._finalizer = finalizer
@@ -128,7 +153,7 @@ class FinalizedChalkQuery:
         self,
         expected_features: Optional[Sequence[Feature]] = None,
         connection: Optional[Any] = None,
-        attempt_efficient_execution: bool = True,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
     ) -> Union[Features, DataFrame, None]:
         """Actually execute the query to a `DataFrame` or set of features.
 
@@ -144,9 +169,8 @@ class FinalizedChalkQuery:
         connection
             Execute the query using the supplied connection. If `None` (default), a new connection will be acquired
             from the underlying source for the query
-        attempt_efficient_execution
-            The executor will attempt to avoid the use of SQLAlchemy and use more efficient methods to
-            retrieve data.
+        query_execution_parameters
+            Parameters to decide method of efficient/inefficient execution
 
         Returns
         -------
@@ -155,16 +179,16 @@ class FinalizedChalkQuery:
             If the finalizer is ONE_OR_NONE or FIRST, then the result may be
             `None` if no row was found
         """
-
-        pa_table = self.execute_to_pyarrow(expected_features, connection, attempt_efficient_execution)
-        return self._pa_table_to_res(pa_table)
+        return self._pa_table_to_res(
+            self.execute_to_pyarrow(expected_features, connection, query_execution_parameters),
+            expected_features,
+        )
 
     async def async_execute(
         self,
         expected_features: Optional[Sequence[Feature]] = None,
         connection: Optional[Any] = None,
-        attempt_efficient_execution: bool = True,
-        polars_read_csv: bool = False,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
     ) -> Union[Features, DataFrame, None]:
         """Actually execute the query to a `DataFrame` or set of features.
 
@@ -180,6 +204,8 @@ class FinalizedChalkQuery:
         connection
             Execute the query using the supplied connection. If `None` (default), a new connection will be acquired
             from the underlying source for the query
+        query_execution_parameters
+            Parameters to decide method of efficient/inefficient execution
 
         Returns
         -------
@@ -188,30 +214,29 @@ class FinalizedChalkQuery:
             If the finalizer is ONE_OR_NONE or FIRST, then the result may be
             `None` if no row was found
         """
-
-        pa_table = await self.async_execute_to_pyarrow(
+        return self._pa_table_to_res(
+            await self.async_execute_to_pyarrow(expected_features, connection, query_execution_parameters),
             expected_features,
-            connection,
-            attempt_efficient_execution,
-            polars_read_csv,
         )
-        return self._pa_table_to_res(pa_table)
 
-    def _pa_table_to_res(self, pa_table: pa.Table):
+    def _pa_table_to_res(self, pa_table: pa.RecordBatch | pa.Table, expected_features: Optional[Sequence[Feature]]):
         try:
             import polars as pl
         except ImportError:
             raise missing_dependency_exception("chalkpy[runtime]")
-        df = pl.from_arrow(pa_table)
+        pa_table = self._postprocess_table(0, pa_table)
+        if isinstance(pa_table, pa.RecordBatch):
+            pa_table = pa.Table.from_batches([pa_table])
+        df = pa_table_to_pl_df(pa_table)
         assert isinstance(df, pl.DataFrame)
-        res = DataFrame(df)
-        if self._finalizer in (Finalizer.ONE_OR_NONE, Finalizer.FIRST) and len(res) == 0:
+        if self._finalizer in (Finalizer.ONE_OR_NONE, Finalizer.FIRST) and len(df) == 0:
             return None
+        res = DataFrame(df)
         if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
             return res.slice(0, 1).to_features()[0]
         return res
 
-    def _get_col_to_converter(
+    def _get_col_to_feature(
         self,
         expected_features: Optional[Sequence[Feature]] = None,
     ):
@@ -219,12 +244,12 @@ class FinalizedChalkQuery:
             None if expected_features is None else frozenset(x.root_fqn for x in expected_features)
         )
 
-        def col_to_converters(column_names: List[str]) -> Dict[str, FeatureConverter]:
+        def col_to_features(column_names: Sequence[str]) -> Mapping[str, Feature]:
             # First map the column names to determine the feature fqns
             col_name_mapping = self.get_col_name_mapping(tuple(column_names), expected_feature_root_fqns)
-            return {k: Feature.from_root_fqn(v).converter for (k, v) in col_name_mapping.items()}
+            return {k: Feature.from_root_fqn(v) for (k, v) in col_name_mapping.items()}
 
-        return col_to_converters
+        return col_to_features
 
     def _check_incremental_settings(self):
         if self.incremental_settings is not None:
@@ -245,8 +270,7 @@ class FinalizedChalkQuery:
         self,
         expected_features: Optional[Sequence[Feature]] = None,
         connection: Optional[Connection] = None,
-        attempt_efficient_execution: bool = True,
-        polars_read_csv: bool = False,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
     ) -> pa.Table:
         """Actually execute the query, and return a PyArrow table. Unlike :meth:`.execute`, this method will always keep the
         results as a PyArrow table, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).
@@ -259,30 +283,90 @@ class FinalizedChalkQuery:
         connection
             Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
             from the underlying source for the query
-        attempt_efficient_execution
-            The executor will attempt to avoid the use of SQLAlchemy and use more efficient methods to
-            retrieve data.
+        query_execution_parameters
+            Optional, but should be set if called from the engine. Contains options that decide the query execution
+            codepath.
 
         Returns
         -------
         Table
             A `pa.Table`, even if the result contains 0 or 1 rows.
         """
+        if query_execution_parameters is None:
+            query_execution_parameters = query_execution_parameters_from_env_vars()
+        query_execution_parameters = dataclasses.replace(query_execution_parameters, yield_empty_batches=True)
+        batches = list(self.execute_to_pyarrow_batches(expected_features, connection, query_execution_parameters))
+        if len(batches) == 0:
+            # This should be impossible to reach because we specify yield_empty_batches=True, so batches should generally contain at least one element
+            if expected_features is None:
+                # We have no idea what the expected schema is. This should
+                return pa.Table.from_pydict({})
+            else:
+                return pa.Table.from_pydict(
+                    {k.root_fqn: [] for k in expected_features},
+                    pa.schema({k.root_fqn: k.converter.pyarrow_dtype for k in expected_features}),
+                )
+        return pa.Table.from_batches(batches)
+
+    def execute_to_pyarrow_batches(
+        self,
+        expected_features: Optional[Sequence[Feature]] = None,
+        connection: Optional[Connection] = None,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
+    ) -> Iterable[pa.RecordBatch]:
+        """Actually execute the query, and return a stream of `pa.RecordBatch`.
+
+        Parameters
+        ----------
+        expected_features
+            The list of expected features for the output, as provided by the resolver in which this query is executed.
+            If not specified, the column names as returned by the query will be used to determine the output features.
+        connection
+            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
+            from the underlying source for the query
+        query_execution_parameters
+            Optional, but should be set if called from the engine. Contains options that decide the query execution
+            codepath.
+
+        Yields
+        ------
+        A stream of `pa.RecordBatch`
+        """
+        from chalk.sql._internal.sql_source import BaseSQLSource
+
         self._check_incremental_settings()
-        col_to_converters = self._get_col_to_converter(expected_features)
+        col_to_features = self._get_col_to_feature(expected_features)
+        assert isinstance(self.source, BaseSQLSource)
 
-        pa_table = self.source.execute_query(
-            self, col_to_converters, connection, attempt_efficient_execution, polars_read_csv
-        )
+        if query_execution_parameters is None:
+            query_execution_parameters = query_execution_parameters_from_env_vars()
 
-        return self._postprocess_table(pa_table, expected_features)
+        total_row_count = 0
+
+        for i, pa_table in enumerate(
+            self.source.execute_query(
+                finalized_query=self,
+                columns_to_features=col_to_features,
+                connection=connection,
+                query_execution_parameters=query_execution_parameters,
+            )
+        ):
+            assert (
+                len(pa_table) > 0 or query_execution_parameters.yield_empty_batches
+            ), "Should never yield empty batches"
+            pa_table = self._postprocess_table(i, pa_table)
+            total_row_count += len(pa_table)
+            yield pa_table
+            if self.finalizer == Finalizer.FIRST and total_row_count > 0:
+                break
+        if total_row_count == 0 and self.finalizer == Finalizer.ONE:
+            raise ValueError("Expected exactly one row; got 0 rows")
 
     async def async_execute_to_pyarrow(
         self,
         expected_features: Optional[Sequence[Feature]] = None,
         connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
-        attempt_efficient_execution: bool = True,
-        polars_read_csv: bool = False,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
     ) -> pa.Table:
         """Actually execute the query, and return a PyArrow table. Unlike :meth:`.execute`, this method will always keep the
         results as a PyArrow table, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).
@@ -290,38 +374,107 @@ class FinalizedChalkQuery:
         Parameters
         ----------
         expected_features
-            The list of expected features for the output, as provided by the resovler in which this query is executed.
+            The list of expected features for the output, as provided by the resolver in which this query is executed.
             If not specified, the column names as returned by the query will be used to determine the output features.
         connection
             Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
             from the underlying source for the query
+        query_execution_parameters
+            Optional, but should be set if called from the engine. Contains options that decide the query execution
+            codepath.
 
         Returns
         -------
         Table
             A `pa.Table`, even if the result contains 0 or 1 rows.
         """
+        if query_execution_parameters is None:
+            query_execution_parameters = query_execution_parameters_from_env_vars()
+        query_execution_parameters = dataclasses.replace(query_execution_parameters, yield_empty_batches=True)
+        batches: list[pa.RecordBatch] = []
+        async for batch in self.async_execute_to_pyarrow_batches(
+            expected_features, connection, query_execution_parameters
+        ):
+            batches.append(batch)
+        if len(batches) == 0:
+            # This should be impossible to reach because we specify yield_empty_batches=True, so batches should generally contain at least one element
+            if expected_features is None:
+                # We have no idea what the expected schema is
+                return pa.Table.from_pydict({})
+            else:
+                return pa.Table.from_pydict(
+                    {k.root_fqn: [] for k in expected_features},
+                    pa.schema({k.root_fqn: k.converter.pyarrow_dtype for k in expected_features}),
+                )
+        return pa.Table.from_batches(batches)
+
+    async def async_execute_to_pyarrow_batches(
+        self,
+        expected_features: Optional[Sequence[Feature]] = None,
+        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
+    ) -> AsyncIterable[pa.RecordBatch]:
+        """Actually execute the query, and return a stream of pa.RecordBatches.
+
+        Parameters
+        ----------
+        expected_features
+            The list of expected features for the output, as provided by the resolver in which this query is executed.
+            If not specified, the column names as returned by the query will be used to determine the output features.
+        connection
+            Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
+            from the underlying source for the query
+        query_execution_parameters
+            Optional, but should be set if called from the engine. Contains options that decide the query execution
+            codepath.
+
+        Yields
+        ------
+        A stream of `pa.RecordBatch`
+        """
+        from chalk.sql._internal.sql_source import BaseSQLSource
+
+        if query_execution_parameters is None:
+            query_execution_parameters = query_execution_parameters_from_env_vars()
+
         self._check_incremental_settings()
-        col_to_converters = self._get_col_to_converter(expected_features)
+        col_to_features = self._get_col_to_feature(expected_features)
+        assert isinstance(self.source, BaseSQLSource)
 
-        pa_table = await self.source.async_execute_query(
-            self,
-            col_to_converters,
-            connection,
-            attempt_efficient_execution,
-            polars_read_csv,
-        )
+        total_row_count = 0
 
-        return self._postprocess_table(pa_table, expected_features)
+        async for i, pa_table in async_enumerate(
+            self.source.async_execute_query(self, col_to_features, connection, query_execution_parameters)
+        ):
+            assert (
+                len(pa_table) > 0 or query_execution_parameters.yield_empty_batches
+            ), "Should never yield empty batches"
+            pa_table = self._postprocess_table(i, pa_table)
+            total_row_count += len(pa_table)
+            yield pa_table
+            if self.finalizer == Finalizer.FIRST and total_row_count > 0:
+                break
 
-    def _postprocess_table(self, pa_table: pa.Table, expected_features: Optional[Sequence[Feature]]):
-        expected_feature_root_fqns = (
-            None if expected_features is None else frozenset(x.root_fqn for x in expected_features)
-        )
-        col_name_mapping = self.get_col_name_mapping(tuple(pa_table.column_names), expected_feature_root_fqns)
-        pa_table = pa_table.select(list(col_name_mapping.keys()))
-        pa_table = pa_table.rename_columns([col_name_mapping[x] for x in pa_table.column_names])
+        if total_row_count == 0 and self.finalizer == Finalizer.ONE:
+            raise ValueError("Expected exactly one row; got 0 rows")
 
+    @overload
+    def _postprocess_table(self, i: int, pa_table: pa.RecordBatch) -> pa.RecordBatch:
+        ...
+
+    @overload
+    def _postprocess_table(self, i: int, pa_table: pa.Table) -> pa.Table:  # pyright: ignore[reportOverlappingOverload]
+        ...
+
+    def _postprocess_table(self, i: int, pa_table: pa.RecordBatch | pa.Table) -> pa.RecordBatch | pa.Table:
+        if len(pa_table) == 0:
+            return pa_table
+        if i > 0:
+            if self.finalizer == Finalizer.ONE_OR_NONE:
+                raise ValueError(f"Expected zero or one rows, but got multiple batches")
+            if self.finalizer == Finalizer.ONE:
+                raise ValueError(f"Expected one row, but got multiple batches")
+            assert self.finalizer != Finalizer.FIRST
         if self._finalizer == Finalizer.ONE:
             if len(pa_table) != 1:
                 raise ValueError(f"Expected exactly one row; got {len(pa_table)} rows")
@@ -334,56 +487,11 @@ class FinalizedChalkQuery:
             pa_table = pa_table.slice(0, 1)
         return pa_table
 
-    def execute_to_pyarrow_batches(
-        self,
-        expected_features: Optional[Sequence[Feature]] = None,
-        connection: Optional[Connection] = None,
-    ):
-        if not hasattr(self.source, "execute_to_batches"):
-            yield self.execute_to_pyarrow(expected_features, connection)
-
-        if self.incremental_settings is not None:
-            # FIXME: Move the incrementalization logic here, so then the `execute` and `execute_to_dataframe`
-            # methods can take the hwm timestamp as a paramater, to allow for direct execution
-            warnings.warn(
-                (
-                    "This query specified an incremental configuration, which has not been applied. "
-                    "This is likely because the resolver is being executed directly. "
-                    "The query will be attempted without any high-water-mark timestamp. "
-                    "This will attempt to select all data, "
-                    "or if the filters depend on the incremental timestamp, "
-                    "will result in a query execution error. "
-                )
-            )
-        expected_feature_root_fqns = (
-            None if expected_features is None else frozenset(x.root_fqn for x in expected_features)
-        )
-
-        def converter_getters(column_names: List[str]) -> Dict[str, FeatureConverter]:
-            # First map the column names to determine the feature fqns
-            col_name_mapping = self.get_col_name_mapping(tuple(column_names), expected_feature_root_fqns)
-            return {k: Feature.from_root_fqn(v).converter for (k, v) in col_name_mapping.items()}
-
-        for pa_table in self.source.execute_to_batches(self, converter_getters, connection):
-            col_name_mapping = self.get_col_name_mapping(tuple(pa_table.column_names), expected_feature_root_fqns)
-            pa_table = pa_table.select(list(col_name_mapping.keys()))
-            pa_table = pa_table.rename_columns([col_name_mapping[x] for x in pa_table.column_names])
-            # if self._finalizer == Finalizer.ONE:
-            #     if len(pa_table) != 1:
-            #         raise ValueError(f"Expected exactly one row; got {len(pa_table)} rows")
-            #
-            # if self._finalizer == Finalizer.ONE_OR_NONE:
-            #     if len(pa_table) > 1:
-            #         raise ValueError(f"Expected zero or one rows; got {len(pa_table)} rows")
-            #
-            # if self._finalizer in (Finalizer.ONE, Finalizer.ONE_OR_NONE, Finalizer.FIRST):
-            #     pa_table = pa_table.slice(0, 1)
-            yield pa_table
-
     def execute_to_result_handles(
         self,
         expected_features: Optional[Sequence[Feature]] = None,
         connection: Optional[Connection] = None,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
     ):
         if not hasattr(self.source, "execute_to_result_handles"):
             raise NotImplementedError
@@ -410,14 +518,15 @@ class FinalizedChalkQuery:
             col_name_mapping = self.get_col_name_mapping(tuple(column_names), expected_feature_root_fqns)
             return {k: Feature.from_root_fqn(v).converter for (k, v) in col_name_mapping.items()}
 
-        return self.source.execute_to_result_handles(self, converter_getters, connection)
+        return self.source.execute_to_result_handles(self, converter_getters, connection)  # pyright: ignore
 
     def execute_to_dataframe(
         self,
         expected_features: Optional[Sequence[Feature]] = None,
         connection: Optional[Connection] = None,
-        attempt_efficient_execution: bool = True,
+        query_execution_parameters: Optional[QueryExecutionParameters] = None,
     ):
+        # DEPRECATED -- will be removed soon. Only used internally
         """Actually execute the query, and return a DataFrame. Unlike :meth:`.execute`, this method will always keep the
         results as a DataFrame, even if the finalizer implies a singleton results (e.g. ONE, ONE_OR_NONE, or FIRST).
 
@@ -429,6 +538,8 @@ class FinalizedChalkQuery:
         connection
             Execute the query using the supplied connection. If None (the default), then a new connection will be acquired
             from the underlying source for the query
+        query_execution_parameters
+            Parameters to decide method of efficient/inefficient execution
 
         Returns
         -------
@@ -439,15 +550,15 @@ class FinalizedChalkQuery:
             import polars as pl
         except ImportError:
             raise missing_dependency_exception("chalkpy[runtime]")
-        pa_table = self.execute_to_pyarrow(expected_features, connection, attempt_efficient_execution)
-        df = pl.from_arrow(pa_table)
+        pa_table = self.execute_to_pyarrow(expected_features, connection, query_execution_parameters)
+        df = pa_table_to_pl_df(pa_table)
         assert isinstance(df, pl.DataFrame)
         return DataFrame(df)
 
     def get_col_name_mapping(
         self,
         result_columns: Tuple[str, ...],
-        expected_features: Optional[FrozenSet[str]],
+        expected_features: Optional[Collection[str]],
     ) -> Dict[str, str]:
         """Map the output columns to the expected feature names.
 
@@ -479,11 +590,8 @@ class FinalizedChalkQuery:
             ans[original_col_name] = v.root_fqn
         if expected_features is not None:
             unexpected_fields = [x for x in self._fields.values() if x.root_fqn not in expected_features]
-            if len(unexpected_fields) > 0:
-                raise ValueError(
-                    f"Fields {unexpected_fields} were in the field mapping but "
-                    "are not included in the resolver output signature."
-                )
+            for x in unexpected_fields:
+                del ans[normalized_to_original_col[normalize_string_for_matching(x.root_fqn)]]
             for normalized_col_name, original_col_name in normalized_to_original_col.items():
                 if original_col_name not in ans:
                     maybe_matching_root_fqn = _get_matching_root_fqn(normalized_col_name, expected_features)
