@@ -1,7 +1,4 @@
-from typing import Any, List, Literal, Optional, Union
-
-import pydantic
-from pydantic import BaseModel
+from typing import TYPE_CHECKING, Any, List, Literal, Optional, Union
 
 from chalk.integrations.named import load_integration_variable
 from chalk.sink._models import SinkIntegrationProtocol
@@ -9,6 +6,17 @@ from chalk.streams.base import StreamSource
 from chalk.utils.duration import Duration
 from chalk.utils.string import comma_whitespace_split
 
+if TYPE_CHECKING:
+    import pydantic
+    from pydantic import BaseModel
+else:
+    try:
+        import pydantic.v1 as pydantic
+        from pydantic.v1 import BaseModel
+    except ImportError:
+        import pydantic
+        from pydantic import BaseModel
+
 
 class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True):
     bootstrap_server: Optional[Union[str, List[str]]] = None
@@ -102,16 +110,16 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
             or load_integration_variable(name="KAFKA_SSL_KEYSTORE_LOCATION", integration_name=name),
             client_id_prefix=client_id_prefix
             or load_integration_variable(name="KAFKA_CLIENT_ID_PREFIX", integration_name=name)
-            or KafkaSource.__fields__.get("client_id_prefix").default,
+            or KafkaSource.__fields__["client_id_prefix"].default,
             group_id_prefix=group_id_prefix
             or load_integration_variable(name="KAFKA_GROUP_ID_PREFIX", integration_name=name)
-            or KafkaSource.__fields__.get("group_id_prefix").default,
+            or KafkaSource.__fields__["group_id_prefix"].default,
             security_protocol=security_protocol
             or load_integration_variable(name="KAFKA_SECURITY_PROTOCOL", integration_name=name)
-            or KafkaSource.__fields__.get("security_protocol").default,
+            or KafkaSource.__fields__["security_protocol"].default,
             sasl_mechanism=sasl_mechanism
             or load_integration_variable(name="KAFKA_SASL_MECHANISM", integration_name=name)
-            or KafkaSource.__fields__.get("sasl_mechanism").default,
+            or KafkaSource.__fields__["sasl_mechanism"].default,
             sasl_username=sasl_username or load_integration_variable(name="KAFKA_SASL_USERNAME", integration_name=name),
             sasl_password=sasl_password or load_integration_variable(name="KAFKA_SASL_PASSWORD", integration_name=name),
             name=name,
@@ -121,7 +129,7 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
         )
         self.registry.append(self)
 
-    def _config_to_json(self) -> Any:
+    def config_to_json(self) -> Any:
         return self.json()
 
     @property
@@ -133,5 +141,6 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
         return self.dead_letter_queue_topic
 
     @property
-    def stream_or_topic_name(self) -> str:
+    def stream_or_topic_name(self):
+        assert self.topic is not None
         return self.topic
