@@ -20,6 +20,7 @@ if TYPE_CHECKING:
 
 class WindowedInstance(Generic[TRich]):
     def __init__(self, values: Mapping[str, TRich]):
+        super().__init__()
         self.values = values
 
     def __call__(self, period: str):
@@ -96,9 +97,9 @@ class Windowed(Generic[TRich], metaclass=_WINDOWED_METACLASS):
     ...     failed_logins: Windowed[int] = windowed("10m", "24h")
     """
 
-    def __getitem__(self, item):
+    def __getitem__(self, item: Type[TRich]):
         # Here for editor support
-        super().__getitem__(item)
+        super().__getitem__(item)  # pyright: ignore
 
     @property
     def buckets_seconds(self) -> Set[int]:
@@ -193,6 +194,7 @@ class Windowed(Generic[TRich], metaclass=_WINDOWED_METACLASS):
         kind: Optional[Type[TRich]],
         offline_ttl: Optional[Union[Duration, ellipsis]],
     ):
+        super().__init__()
         self._kind = kind
         self._name: Optional[str] = None
         self._buckets = buckets
@@ -220,15 +222,16 @@ class Windowed(Generic[TRich], metaclass=_WINDOWED_METACLASS):
 
 class SelectedWindow:
     def __init__(self, kind: Windowed, selected: str):
+        super().__init__()
         self.windowed = kind
         self.selected = selected
 
 
 def windowed(
     *buckets: str,
-    days: Iterable[int, ...] = (),
-    hours: Iterable[int, ...] = (),
-    minutes: Iterable[int, ...] = (),
+    days: Iterable[int] = (),
+    hours: Iterable[int] = (),
+    minutes: Iterable[int] = (),
     description: Optional[str] = None,
     owner: Optional[str] = None,
     tags: Optional[Any] = None,
