@@ -1,5 +1,7 @@
+from __future__ import annotations
+
 import collections.abc
-from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, TypeVar, Union, cast, overload
+from typing import Any, Dict, Iterable, List, Sequence, Set, Tuple, TypeVar, Union, cast, overload
 
 from typing_extensions import Annotated, get_args, get_origin
 
@@ -21,14 +23,15 @@ def unwrap_annotated_if_needed(typ: Any):
     return typ
 
 
-def unwrap_optional_and_annotated_if_needed(typ: Any):
-    origin = get_origin(typ)
+def is_optional(typ: Any) -> bool:
     args = get_args(typ)
+    return get_origin(typ) in (Union, UnionType) and len(args) == 2 and any(d == _NoneType for d in args)
+
+
+def unwrap_optional_and_annotated_if_needed(typ: Any):
     typ = unwrap_annotated_if_needed(typ)
-
-    if origin in (Union, UnionType) and len(args) == 2 and any(d == _NoneType for d in args):
+    if is_optional(typ):
         return next(m for m in get_args(typ) if m != _NoneType)
-
     return typ
 
 
@@ -153,39 +156,27 @@ def ensure_tuple(x: Union[T, Sequence[T], Dict[Any, T], None]) -> Tuple[T, ...]:
     return (x,)
 
 
-def get_unique_item(collection: Iterable[Optional[T]], name: Optional[str] = None) -> T:
-    item = None
-    in_name = f" in `{name}`" if name is not None else ""
-    for x in collection:
-        if x is None:
-            raise ValueError(f"Item{in_name} None")
-        if item is not None:
-            if x != item:
-                raise ValueError(f"Multiple values{in_name} are not permitted. Found {x}, {item}")
-        item = x
-    if item is None:
+def get_unique_item(collection: Iterable[T], name: str | None = None) -> T:
+    """Get the unique item from the collection, ignoring equal duplicates. Raises a ValueError if there is not a single, unique element."""
+    unique_item_or_none = get_unique_item_if_exists(collection, name)
+    if unique_item_or_none is None:
+        in_name = f" in `{name}`" if name is not None else ""
         raise ValueError(f"There should be at least one item{in_name}")
-    return item
+    return unique_item_or_none
 
 
-def get_unique_item_or_none(iterable: Iterable[Optional[T]]) -> Optional[T]:
-    item = None
+def get_unique_item_if_exists(iterable: Iterable[T], name: str | None = None) -> T | None:
+    """Get the unique item from the collection, if any element exists inside the collection.
+
+    Raises a ValueError if there are multiple unique elements. Returns None if the collection is empty.
+    """
+    item = ...
+    in_name = f" in `{name}`" if name is not None else ""
     for x in iterable:
-        if x is None:
-            continue
-        if item is not None:
-            raise ValueError(f"Multiple values not permitted. Found {item}, {x}")
-        item = x
-    return item
-
-
-def get_unique_item_if_exists(collection: Iterable[Optional[T]]) -> Optional[T]:
-    item = None
-    for x in collection:
-        if x is None:
-            continue
-        if item is not None:
-            if x != item:
-                raise ValueError(f"Multiple values not permitted. Found {item}, {x}")
-        item = x
+        if item is ...:
+            item = x
+        if x != item:
+            raise ValueError(f"Multiple values{in_name} are not permitted. Found {x}, {item}")
+    if item is ...:
+        return None
     return item
