@@ -5,12 +5,10 @@ import collections.abc
 import contextlib
 import contextvars
 import logging
-from typing import Any, Dict, List, Mapping, Optional
+from typing import Any, Dict, Mapping, Optional
 from weakref import WeakKeyDictionary
 
-_LOGGING_CONTEXT_STACK: contextvars.ContextVar[List[Mapping[str, Any]]] = contextvars.ContextVar(
-    "_LOGGING_CONTEXT_STACK"
-)
+_LOGGING_CONTEXT: contextvars.ContextVar[Mapping[str, Any]] = contextvars.ContextVar("_LOGGING_CONTEXT")
 
 
 def _recursive_merge(a: Mapping[str, Any], b: Mapping[str, Any]):
@@ -28,41 +26,7 @@ def get_logging_context() -> Mapping[str, Any]:
     Retrieve the log context for the current python context.
     This initializes the thread-local variable if necessary.
     """
-    stack = _LOGGING_CONTEXT_STACK.get([{}])
-    return stack[-1]
-
-
-def set_logging_context(_merge: bool = True, **extra: Any):
-    """Sets the context-local log context to a new value.
-
-    Parameters
-    ----------
-    _merge
-        If True, the new context will be merged with the current context. Otherwise, the new context will overwrite the existing context.
-    """
-    log_context_stack = _LOGGING_CONTEXT_STACK.get([{}])
-    if _merge:
-        extra = _recursive_merge(
-            log_context_stack[-1],
-            extra,
-        )
-    log_context_stack.append(extra)
-    _LOGGING_CONTEXT_STACK.set(log_context_stack)
-
-
-def pop_logging_context():
-    """Pop the latest log context off of the log context stack, and reset it to its previous value.
-
-    Raises
-    ------
-    RuntimeError
-        If there is no current context from set_logging_context()
-    """
-    stack = _LOGGING_CONTEXT_STACK.get([])
-    if len(stack) == 0:
-        raise RuntimeError("Log context stack is empty")
-    stack.pop()
-    _LOGGING_CONTEXT_STACK.set(stack)
+    return _LOGGING_CONTEXT.get({})
 
 
 class LogWithContextFilter(logging.Filter):
@@ -92,7 +56,7 @@ Logger = get_logger
 
 
 @contextlib.contextmanager
-def add_logging_context(*, _merge: bool = True, **extra: Any):
+def add_logging_context(*, _merge: bool = True, **log_context: Any):
     """A context manager to push and pop `extra` dictionary keys.
 
     Parameters
@@ -102,8 +66,13 @@ def add_logging_context(*, _merge: bool = True, **extra: Any):
     extra
         Contextual information to add to the log record
     """
-    set_logging_context(_merge, **extra)
+    if _merge:
+        log_context = _recursive_merge(
+            _LOGGING_CONTEXT.get({}),
+            log_context,
+        )
+    token = _LOGGING_CONTEXT.set(log_context)
     try:
         yield
     finally:
-        pop_logging_context()
+        _LOGGING_CONTEXT.reset(token)
