@@ -4,14 +4,13 @@ import pathlib
 import typing
 import warnings
 from collections import defaultdict
-from typing import Dict, List, NamedTuple, Optional, Tuple, Type, Union
+from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Type, Union, cast
 
 from typing_extensions import Annotated, get_args, get_origin
 
-from chalk.features import Feature, Features, FeatureSetBase, FeatureWrapper, unwrap_feature
+from chalk.features import Feature, Features, FeatureSetBase, Vector, FeatureWrapper, unwrap_feature
 from chalk.importer import import_all_python_files_from_dir
 from chalk.utils.paths import get_directory_root
-from chalk.streams import Windowed
 
 try:
     from types import UnionType
@@ -32,12 +31,15 @@ class ParsedFeaturesClass(NamedTuple):
 
 class StubGenerator:
     def __init__(self) -> None:
+        super().__init__()
         self._parsed_feature_classes: Dict[str, ParsedFeaturesClass] = {}
         self._module_to_stubs: List[Type[Features]] = []
         self._imports: Dict[Tuple[str, ...], str] = {}  # Mapping of full module name to the name it is imported as
 
     def _add_import(
-        self, item: Union[str, type, typing._SpecialForm, Windowed], import_as: Optional[str] = None
+        self,
+        item: Union[str, type, typing._SpecialForm],  # pyright: ignore[reportPrivateUsage]
+        import_as: Optional[str] = None,
     ) -> str:
         """Add an import for item.
 
@@ -54,15 +56,13 @@ class StubGenerator:
         str
             The name that `item` will be imported as.
         """
-        if isinstance(item, typing._SpecialForm):
+        if isinstance(item, typing._SpecialForm):  # pyright: ignore[reportPrivateUsage]
             item = str(item)
         if isinstance(item, FeatureWrapper):
-            item = type(unwrap_feature(item))
-        if isinstance(item, Windowed):
-            item = type(item)
+            item = unwrap_feature(item).fqn
         if isinstance(item, type):
             item = f"{item.__module__}.{item.__name__}"
-        if not isinstance(item, str):
+        if not isinstance(item, str):  # pyright: ignore[reportUnnecessaryIsInstance]
             raise TypeError(f"Invalid type for {item}: {type(item)}")
         item_split = tuple(item.split("."))
         if item_split[0] == "builtins":
@@ -88,7 +88,7 @@ class StubGenerator:
         item = f"__stubgen_{'_'.join(item_split)}"
         return item
 
-    def _register_imports_for_annotation(self, annotation: Union[type, Windowed]) -> Tuple[str, str]:
+    def _register_imports_for_annotation(self, annotation: type) -> Tuple[str, str]:
         """Register all necessary imports, and return a tuple of (protocol annotation, init annotation)"""
         if annotation == type(None):
             return "None", "None"
@@ -97,15 +97,9 @@ class StubGenerator:
         origin, args = get_origin(annotation), get_args(annotation)
         if origin is None:
             assert args == ()
-            # Likely a "simple" annotation, like an enum class
-            if isinstance(annotation, Windowed):
-                origin_name = self._add_import(annotation, import_as="Windowed")
-                assert annotation._kind is not None
-                protocol_arg_name, init_arg_name = self._register_imports_for_annotation(annotation._kind)
-                # FIXME: I don't think this is entirely correct, b/c if a user manually constructs a windowed
-                # feature class, then it's unclear what should be passed in for the windowed arg in the init
-                # parameter
-                return f"{origin_name}[{protocol_arg_name}]", f"{origin_name}[{init_arg_name}]"
+            if isinstance(annotation, type) and issubclass(annotation, Vector):
+                origin_name = self._add_import(Vector, import_as="Vector")
+                return "Vector", "Vector"
             elif annotation.__module__ == "builtins":
                 assert isinstance(annotation, type)
                 return annotation.__name__, annotation.__name__
@@ -122,7 +116,7 @@ class StubGenerator:
                 protocol_arg_name, init_arg_name = self._register_imports_for_annotation(args[0])
                 return f"Primary[{protocol_arg_name}]", init_arg_name
             return self._register_imports_for_annotation(args[0])
-        origin_name = self._add_import(origin)
+        origin_name = self._add_import(cast(Any, origin))
         arg_names = [self._register_imports_for_annotation(arg) for arg in args]
         return (
             f"{origin_name}[" + ", ".join(x[0] for x in arg_names) + "]",
@@ -134,7 +128,7 @@ class StubGenerator:
         for f in features_cls.features:
             assert isinstance(f, Feature)
             assert f.typ is not None
-            if f.typ.is_dataframe:
+            if f.typ.as_dataframe() is not None:
                 self._add_import("chalk.features.DataFrame", "DataFrame")
                 annotations[f.attribute_name] = ParsedAnnotation(
                     "DataFrame",
@@ -146,14 +140,22 @@ class StubGenerator:
                 proto_annotation, init_annotation = self._register_imports_for_annotation(f.typ.parsed_annotation)
                 include_in_proto_cls = not f.is_windowed_pseudofeature and not f.is_autogenerated
                 if include_in_proto_cls:
-                    # TODO do not skip the annotations that do not appear in the protocol class
+                    protocol_annotation = proto_annotation
+                    if f.typ.as_features_cls() is not None:
+                        protocol_annotation = "Any"
+                    else:
+                        as_feature = f.typ.as_feature()
+                        if as_feature is not None:
+                            protocol_annotation, init_annotation = self._register_imports_for_annotation(
+                                as_feature.typ.parsed_annotation
+                            )
                     annotations[f.attribute_name] = ParsedAnnotation(
                         init_annotation,
                         include_in_protocol_cls=include_in_proto_cls,
                         # If it's a features class (i.e. a has-one), not specifying
                         # the type in the protocol class so the type checker can
                         # still match feature classes that have a circular reference
-                        protocol_annotation="Any" if f.typ.is_features_cls else proto_annotation,
+                        protocol_annotation="Any" if (f.typ.as_features_cls() is not None) else protocol_annotation,
                     )
         if features_cls.__name__ in self._parsed_feature_classes:
             raise ValueError(
@@ -403,6 +405,7 @@ class __stubgen__features_proto(Protocol):"""
 
 
 def configure_stubgen_argparse(parser: argparse.ArgumentParser):
+    parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
     parser.add_argument(
         "--root",
         type=str,
@@ -425,7 +428,10 @@ def configure_stubgen_argparse(parser: argparse.ArgumentParser):
     )
 
 
-def run_stubgen(args: argparse.Namespace, file_filter: Optional[str]):
+def run_stubgen(
+    args: argparse.Namespace,
+    file_filter: Optional[str],
+):
     root = args.root
     if root is None:
         root = get_directory_root() or pathlib.Path(os.getcwd())
